/*
 *
 * cni.c
 * Description:
 * DOCSIS datapath driver implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#define DRV_NAME    "Cable Modem Network Interface driver"
#define DRV_VERSION    "0.0.2"


#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/mii.h>
#include <linux/platform_device.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <linux/proc_fs.h>
#include <linux/inet_lro.h>
#include <linux/ethtool.h>
#include <linux/config.h>

#include "pal.h"
#include "pal_cppi41.h"
#include <linux/kernel.h>
#include <asm/arch/puma5/puma5_cppi.h>

/*
 * Related session router definitions
 * ----------------------------------
 */

#ifdef CONFIG_ARM_AVALANCHE_PPD
static int cni_pp_prepare_pid (struct net_device *dev);
static int cni_pp_set_pid_flags (struct net_device *dev, int flags);
#endif

/* Cable network interface driver */

extern struct net_device * wanDev;
extern struct net_device * mtaDev;

#define CNID_RXINT_NUM        (AVALANCHE_INTD_BASE_INT + CNI_ACC_RX_INTV)
#define CNID_TXINT_NUM        (AVALANCHE_INTD_BASE_INT + CNI_ACC_TXCMPL_INTV)
#define CNID_INTD_HOST_NUM    0
#define CPPI4_BD_LENGTH_FOR_CACHE   64

#define CIND_TX_SERVICE_MAX         32
#define CIND_RX_SERVICE_MAX         64

#define CNID_ACC_PAGE_NUM_ENTRY     (32)
#define CNID_ACC_NUM_PAGE           (2)
#define CNID_ACC_ENTRY_TYPE         (PAL_CPPI41_ACC_ENTRY_TYPE_D)
/* Byte size of page = number of entries per page * size of each entry */
#define CNID_ACC_PAGE_BYTE_SZ       (CNID_ACC_PAGE_NUM_ENTRY * ((CNID_ACC_ENTRY_TYPE + 1) * sizeof(unsigned int*)))
/* byte size of list = byte size of page * number of pages * */
#define CNID_ACC_LIST_BYTE_SZ       (CNID_ACC_PAGE_BYTE_SZ * CNID_ACC_NUM_PAGE)


#define GET_BD_PTR(base, num)       ((base) + ((num) * CNI_BD_SIZE))
#define CNID_QM_DESC_SIZE_CODE      10 /* ((sizeof(Cppi4HostDesc) - 24)/4) */

#define CNID_RX_HOST_BD_HI_OFFSET           0
#define CNID_RX_HOST_BD_MD_OFFSET           ((CNI_RX_HOST_BD_NUM_HIGH     * CNI_BD_SIZE) + CNID_RX_HOST_BD_HI_OFFSET)
#define CNID_RX_HOST_BD_LO_OFFSET           ((CNI_RX_HOST_BD_NUM_MED     * CNI_BD_SIZE) + CNID_RX_HOST_BD_MD_OFFSET)


/* define to enable copious debugging info */
/* #define CNID_DEBUG */

#ifdef CNID_DEBUG
/* note: prints function name for you */
#  define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)
#else
#  define DPRINTK(fmt, args...)
#endif


/* max supported frame size */
#define MAX_DOCSIS_FRAME_SIZE    1600
#define RX_BUF_SIZE    MAX_DOCSIS_FRAME_SIZE

MODULE_AUTHOR ("Texas Instruments, Inc");
MODULE_DESCRIPTION (DRV_NAME);
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);


/************************************************************************/
/*     CNI HOST Descriptor type definition                              */
/************************************************************************/
typedef struct cnid_desc_t
{
    Cppi4HostDesc       hw;         /* The Hardware Descriptor */
    int                 psi[3];     /* protocol specific information for fw  */
    struct sk_buff*     skb;        /* The data pointer virtual address */
}
cnid_desc_t;
/************************************************************************/


/************************************************************************/
/*     CNI driver private data                                          */
/************************************************************************/
typedef struct cnid_private_t {
#ifdef CONFIG_INET_LRO
#define CNI_MAX_LRO_ENTRIES  8

    struct net_lro_mgr      lro_mgr; /* This entry must be first */
    struct net_lro_desc     lro_arr[CNI_MAX_LRO_ENTRIES];
#endif
    PAL_Handle              pal_hnd;        /* The handle to PAL layer */
    PAL_Cppi4AccChHnd       tx_acc_hnd;     /* The Tx accumulator channel handle */
    Ptr                     tx_list_base;
    cnid_desc_t**           tx_list;
    struct tasklet_struct   tx_tasklet;     /* Tx completion processing tasklet */
    Ptr                     bdPoolBase;     /* The base address of entire descriptors region */

    PAL_Cppi4QueueHnd       tx_queue            [CNI_CPPI4x_TX_Q_COUNT];

    PAL_Cppi4QueueHnd       tx_qos_queue        [PPFW_CPPI4x_HOST_TO_PP_PRTY_Q_COUNT];
    PAL_Cppi4QueueHnd       tx_qos_free_queue   [PPFW_CPPI4x_HOST_TO_PP_PRTY_Q_COUNT];

    Ptr                     tx_bdpool;

    /*
     * CNI Rx Accumulator info
     */
    PAL_Cppi4AccChHnd       rxAcc_chan_hnd          [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* The Rx accumulator channel handle */
    unsigned int            rxAcc_chan              [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* The Rx accumulator channel numbers */
    cnid_desc_t**           rxAcc_chan_list         [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* Rx acc channel lists */
    Ptr                     rxAcc_chan_list_base    [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* Rx acc channel lists base*/
    PAL_Cppi4QueueHnd       rxAcc_queue_hnd         [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* Rx acc queue handles */

    /*
     * CNI Proxy channels info
     */
    PAL_Cppi4QueueHnd       rxProxy_free_queue_hnd  [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* Rx Proxy free queue handles */
    PAL_Cppi4QueueHnd       rxProxy_queue_hnd       [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* Rx Proxy queue handles */
    Ptr                     rxProxy_bdpool          [CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]; /* Rx Proxy bd pools */

    struct net_device_stats stats;
    spinlock_t              devlock;
    unsigned long           state;
}
cnid_private_t;
/************************************************************************/



typedef struct channel_info_t
{
    Cppi4Queue  outputQueue;
    Cppi4Queue  freeQueue;
    int         bdnum;
    Ptr         bdPoolOffset;
    int         dmaChannel;
}
channel_info_t;



/************************************************************************/
/*      PP -> HOST  proxy channels information                          */
/************************************************************************/
static  channel_info_t  rxProxy[CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT]=
{
    {
        .outputQueue.qMgr   = CNI_CPPI4x_RX_QMGR,
        .outputQueue.qNum   = CNI_CPPI4x_RX_QNUM( PAL_CPPI4x_PRTY_LOW ),
        .freeQueue.qMgr     = CNI_CPPI4x_FBD_QMGR,
        .freeQueue.qNum     = CNI_CPPI4x_FBD_QNUM( PAL_CPPI4x_PRTY_LOW ),
        .bdnum              = CNI_RX_HOST_BD_NUM_LOW,
        .bdPoolOffset       = (Ptr)CNID_RX_HOST_BD_LO_OFFSET,
        .dmaChannel         = CNI_CPPI4x_DOC2HOST_PROXY_CHNUM(0)
    },
    {
        .outputQueue.qMgr   = CNI_CPPI4x_RX_QMGR,
        .outputQueue.qNum   = CNI_CPPI4x_RX_QNUM( PAL_CPPI4x_PRTY_MED ),
        .freeQueue.qMgr     = CNI_CPPI4x_FBD_QMGR,
        .freeQueue.qNum     = CNI_CPPI4x_FBD_QNUM( PAL_CPPI4x_PRTY_MED ),
        .bdnum              = CNI_RX_HOST_BD_NUM_MED,
        .bdPoolOffset       = (Ptr)CNID_RX_HOST_BD_MD_OFFSET,
        .dmaChannel         = CNI_CPPI4x_DOC2HOST_PROXY_CHNUM(1)
    },
    {
        .outputQueue.qMgr   = CNI_CPPI4x_RX_QMGR,
        .outputQueue.qNum   = CNI_CPPI4x_RX_QNUM( PAL_CPPI4x_PRTY_HIGH ),
        .freeQueue.qMgr     = CNI_CPPI4x_FBD_QMGR,
        .freeQueue.qNum     = CNI_CPPI4x_FBD_QNUM( PAL_CPPI4x_PRTY_HIGH ),
        .bdnum              = CNI_RX_HOST_BD_NUM_HIGH,
        .bdPoolOffset       = (Ptr)CNID_RX_HOST_BD_HI_OFFSET,
        .dmaChannel         = CNI_CPPI4x_DOC2HOST_PROXY_CHNUM(2)
    }
};
/************************************************************************/



/************************************************************************/
/*      DOCSIS -> PP channels information                               */
/************************************************************************/
static  channel_info_t  rxIngress[CNI_CPPI4x_RX_DMA_CH_COUNT]=
{
    {
        .outputQueue.qMgr   = PPFW_CPPI4x_RX_INGRESS_QMGR,
        .outputQueue.qNum   = PPFW_CPPI4x_RX_INGRESS_QNUM( PAL_CPPI4x_PRTY_LOW ),
        .freeQueue.qMgr     = CNI_CPPI4x_FD_QMGR,
        .freeQueue.qNum     = CNI_CPPI4x_FD_QNUM( PAL_CPPI4x_PRTY_LOW ),
        .bdnum              = CNI_RX_EMBEDDED_BD_NUM_LOW,
        .dmaChannel         = CNI_CPPI4x_RX_DMA_CHNUM(0)
    },
    {
        .outputQueue.qMgr   = PPFW_CPPI4x_RX_INGRESS_QMGR,
        .outputQueue.qNum   = PPFW_CPPI4x_RX_INGRESS_QNUM( PAL_CPPI4x_PRTY_LOW ),
        .freeQueue.qMgr     = CNI_CPPI4x_FD_QMGR,
        .freeQueue.qNum     = CNI_CPPI4x_FD_QNUM( PAL_CPPI4x_PRTY_LOW ),
        .bdnum              = 0,
        .dmaChannel         = CNI_CPPI4x_RX_DMA_CHNUM(1)
    },
    {
        .outputQueue.qMgr   = PPFW_CPPI4x_RX_INGRESS_QMGR,
        .outputQueue.qNum   = PPFW_CPPI4x_RX_INGRESS_QNUM( PAL_CPPI4x_PRTY_HIGH ),
        .freeQueue.qMgr     = CNI_CPPI4x_FD_QMGR,
        .freeQueue.qNum     = CNI_CPPI4x_FD_QNUM( PAL_CPPI4x_PRTY_HIGH ),
        .bdnum              = CNI_RX_EMBEDDED_BD_NUM_HIGH,
        .dmaChannel         = CNI_CPPI4x_RX_DMA_CHNUM(2)
    }
};
/************************************************************************/


/* Use this MAC address if none is supplied on the command line */
static unsigned char defmac[] = {0x00, 0x50, 0xF1, 0x80, 0x00, 0x00};
/* Allocation for MAC string supplied on the command line */
static char *inpmac = NULL;

static int __devinit    cnid_probe              (struct device *        dev);
static int __devexit    cnid_remove             (struct device *        dev);
static int              cnid_open               (struct net_device *    dev);
#if 0
static void             cnid_tx_timeout         (struct net_device *    dev);
#endif
static int              cnid_start_xmit         (struct sk_buff *skb, struct net_device *dev);
static int              cnid_poll               (struct net_device *dev, int *budget);
static int              cnid_close              (struct net_device *dev);
static int              cnid_ioctl              (struct net_device *dev, struct ifreq *rq, int cmd);
static struct net_device_stats *cnid_get_stats  (struct net_device *dev);
static void             cnid_set_multicast      (struct net_device *dev);
static irqreturn_t      cnid_tx_interrupt       (int irq, void *dev, struct pt_regs *regs);
static irqreturn_t      cnid_rx_interrupt       (int irq, void *dev, struct pt_regs *regs);
static void             cnid_do_tx_complete     (unsigned long data);

static int              cnid_close_rx_prep      (struct net_device* dev);
static int              cnid_close_rx_finish    (struct net_device* dev);
#ifdef CONFIG_INET_LRO
static void             cnid_set_ethtool_ops(struct net_device *dev);
#endif

/*
 * TODO: This should not be here. It should be in a board/platform
 * specific file which is *actually* aware of what devices are
 * present on board.
 */
static ssize_t cnid_show_version(struct device_driver *drv, char *buf)
{
    return sprintf(buf, "%s, version: %s", DRV_NAME, DRV_VERSION);
}


static DRIVER_ATTR(version, S_IRUGO, cnid_show_version, NULL);

/**************************************************************************/
/*! \fn         cnid_tx_bd_link_skb
 **************************************************************************
 *
 *  \brief      Links an skb to a Tx Descriptor
 *
 *  \param[in]  Net Device
 *  \param[in]  Descriptor
 *  \param[in]  SK buff
 *  \return     length of final packet
 **************************************************************************/
static unsigned int cnid_tx_bd_link_skb(struct net_device* dev, cnid_desc_t* bd, struct sk_buff *skb)
{
    unsigned int len = ((skb->len < ETH_ZLEN)?ETH_ZLEN:skb->len);
    unsigned int   queueNum;
    unsigned int   sfIndex = (skb->ti_meta_info >> 24);

    bd->hw.descInfo = (PAL_CPPI4_HOSTDESC_DESC_TYPE_HOST << PAL_CPPI4_HOSTDESC_DESC_TYPE_SHIFT)
                    | (1                                 << PAL_CPPI4_HOSTDESC_PROT_WORD_CNT_SHIFT)
                    | len;
    bd->hw.bufPtr = PAL_CPPI4_VIRT_2_PHYS((unsigned int)skb->data);
    bd->hw.buffLen = len;
    bd->skb = skb;
    bd->psi[0] = skb->ti_meta_info;
    bd->psi[1] = 0;

    if (8 <= sfIndex)
    {
        /* UGS case - no QoS applied (queue 181) */
        queueNum = CNI_CPPI4x_TX_QNUM(0);
    }
    else
    {
        if (0 != skb->dev->vpid_block.qos_clusters_count)
        {
            if ((skb->input_dev == wanDev)||(skb->input_dev == mtaDev))
            {
                /* Send packets from wan to high priority queues */
                /* Send packets to best effort QoS queues (64, 66, 68...) */
                queueNum = CNI_CPPI4x_QoS_HIGH_TX_QNUM(sfIndex);
            }
            else
            {
                /* Send packets to best effort QoS queues (65, 67, 69...) */
                queueNum = CNI_CPPI4x_QoS_LOW_TX_QNUM(sfIndex);
            }
        }
        else
        {
            queueNum = CNI_CPPI4x_TX_QNUM( 1 + sfIndex );
        }
    }


    /* Set SYNC Q PTID info in egress descriptor */
    if(skb->pp_packet_info.ti_pp_flags == TI_PPM_SESSION_INGRESS_RECORDED)
    {
        memcpy(&(bd->hw.netInfoWord0), skb->pp_packet_info.ti_epi_header, 8);
        bd->hw.netInfoWord1 &= ~(0xFFFF);
        bd->hw.netInfoWord1 |= queueNum; /* after QPDSP, push to QoS/CNI Queues */
    }
    else
    {
        bd->hw.netInfoWord1 = queueNum;
    }

    PAL_CPPI4_CACHE_WRITEBACK((unsigned long)skb->data, skb->len);
    PAL_CPPI4_CACHE_WRITEBACK((unsigned long)bd, CPPI4_BD_LENGTH_FOR_CACHE);

    return len;
}


/**************************************************************************/
/*! \fn         cnid_rx_bd_init
 **************************************************************************
 *
 *  \brief      INIT RX Descriptor
 *
 *  \param[in]  Pointer to Descriptor
 *  \param[in]  Queue Manager
 *  \param[in]  Queue Number
 *  \return     None
 **************************************************************************/
static void cnid_rx_bd_init(cnid_desc_t* bd, int qmgr, int qnum)
{
    bd->hw.descInfo     =  PAL_CPPI4_HOSTDESC_DESC_TYPE_HOST << PAL_CPPI4_HOSTDESC_DESC_TYPE_SHIFT;
    bd->hw.pktInfo      =
              (PAL_CPPI4_HOSTDESC_PKT_TYPE_ETH       << PAL_CPPI4_HOSTDESC_PKT_TYPE_SHIFT)
            | (PAL_CPPI4_HOSTDESC_PKT_RETPLCY_LINKED << PAL_CPPI4_HOSTDESC_PKT_RETPLCY_SHIFT)
            | (PAL_CPPI4_HOSTDESC_DESC_LOC_OFFCHIP   << PAL_CPPI4_HOSTDESC_DESC_LOC_SHIFT)
            | (qmgr                                  << PAL_CPPI4_HOSTDESC_PKT_RETQMGR_SHIFT)
            | (qnum                                  << PAL_CPPI4_HOSTDESC_PKT_RETQNUM_SHIFT);
}


/**************************************************************************/
/*! \fn         cnid_rx_bd_link_skb
 **************************************************************************
 *
 *  \brief      Links an skb to an Rx Descriptor
 *
 *  \param[in]  Net Device
 *  \param[in]  Descriptor
 *  \param[in]  SK buff
 *  \return     length of final packet
 **************************************************************************/
static void cnid_rx_bd_link_skb(struct net_device* dev, cnid_desc_t* bd, struct sk_buff *skb)
{
    /* Invalidate the skb data cache: Do it before the data is presented to the DMA */
    PAL_CPPI4_CACHE_INVALIDATE(skb->data, RX_BUF_SIZE);
    skb->dev            = dev;
    skb_reserve (skb, NET_IP_ALIGN);    /* 16 byte align the IP fields. */
    bd->hw.orgBuffLen  = RX_BUF_SIZE - NET_IP_ALIGN;
    bd->hw.orgBufPtr   = PAL_CPPI4_VIRT_2_PHYS(skb->data);
    bd->skb            = skb;

    bd->psi[0] = 0;
    bd->psi[1] = 0;
    bd->psi[2] = 0;

    /* Write the BD to the RAM */
    PAL_CPPI4_CACHE_WRITEBACK(bd, CPPI4_BD_LENGTH_FOR_CACHE);

}

/** Tx processing functions start here **/

/**************************************************************************/
/*! \fn         cnid_do_tx_complete
 **************************************************************************
 *
 *  \brief      Links an skb to a Rx Descriptor
 *
 *  \param[in]  Net Device
 *  \return     None
 **************************************************************************/
static void cnid_do_tx_complete(unsigned long data)
{
    struct net_device*      dev = (struct net_device*) data;
    cnid_private_t*         priv = netdev_priv(dev);
    cnid_desc_t*            bd;
    int packets_processed = 0;
    int done = 1;
    int txPriority;

    DPRINTK(KERN_DEBUG " Enter %s \n", __FUNCTION__);
    while(avalanche_intd_get_interrupt_count(CNID_INTD_HOST_NUM, CNI_ACC_TXCMPL_CHNUM))
    {
        while((bd = (cnid_desc_t*)((unsigned long)*priv->tx_list & QMGR_QUEUE_N_REG_D_DESC_ADDR_MASK)))
        {
            bd = PAL_CPPI4_PHYS_2_VIRT(bd);
            PAL_CPPI4_CACHE_INVALIDATE(bd, CPPI4_BD_LENGTH_FOR_CACHE);

            priv->stats.tx_packets++;
            priv->stats.tx_bytes += bd->skb->len;

            dev_kfree_skb_any(bd->skb);
            bd->skb = NULL;

            /* Queue back the BD to free pool */
            txPriority = bd->psi[2];
            PAL_cppi4QueuePush(priv->tx_qos_free_queue[txPriority] , (Ptr)PAL_CPPI4_VIRT_2_PHYS(bd), CNID_QM_DESC_SIZE_CODE, 0);

            packets_processed++;
            priv->tx_list++;

            /* did enough work, move now.. */
            if(!test_bit(0, &priv->state) && (packets_processed == CIND_TX_SERVICE_MAX))
            {
                done = 0;
                goto out;
            }
        }

        /* Update the list entry for next time */
        priv->tx_list = PAL_cppi4AccChGetNextList(priv->tx_acc_hnd);

        avalanche_intd_set_interrupt_count(CNID_INTD_HOST_NUM, CNI_ACC_TXCMPL_CHNUM, 1);
    }

out:

    if(done)
    {
        avalanche_intd_write_eoi(CNI_ACC_TXCMPL_INTV);
    }
    else
    {
        tasklet_schedule(&priv->tx_tasklet);
    }
}

/**************************************************************************/
/*! \fn         cnid_start_xmit
 **************************************************************************
 *
 *  \brief      Transmit Function
 *
 *  \param[in]  SK buff
 *  \param[in]  Net Device
 *  \return     OK or error
 **************************************************************************/
static int cnid_start_xmit (struct sk_buff *skb, struct net_device *dev)
{
    cnid_private_t* priv = netdev_priv(dev);
    cnid_desc_t*    bd;
    unsigned int len;
    unsigned int prio = 0;

    DPRINTK( " Enter \n");

    if ((skb->input_dev == wanDev)||(skb->input_dev == mtaDev))
    {
        prio = PAL_CPPI4x_PRTY_HIGH;
    }

    /* get a free Tx descriptor */
    if(!(bd = (cnid_desc_t *)PAL_cppi4QueuePop(priv->tx_qos_free_queue[prio]) ))
    {
        /* This should not occur in this driver
         * (because of what is done later in this function) */
        priv->stats.tx_dropped++;
        dev_kfree_skb_any(skb);
        return NETDEV_TX_OK;
    }

    bd = PAL_CPPI4_PHYS_2_VIRT(bd);
    /* We save the queue for TX complete */
    bd->psi[2] = prio;
    bd->hw.pktInfo &= ~(PAL_CPPI4_HOSTDESC_PKT_RETQMGR_MASK | PAL_CPPI4_HOSTDESC_PKT_RETQNUM_MASK);
    bd->hw.pktInfo |= (CNI_CPPI4x_TX_COMP_QMGR    << PAL_CPPI4_HOSTDESC_PKT_RETQMGR_SHIFT)
                   |  (CNI_CPPI4x_TX_COMP_QNUM(0) << PAL_CPPI4_HOSTDESC_PKT_RETQNUM_SHIFT);

    len = cnid_tx_bd_link_skb(dev, bd, skb);

    dev->trans_start = jiffies;

    /* Push to QPDSP Q - then after handle PTID it will be send to the CNI Tx queues */
    PAL_cppi4QueuePush(priv->tx_qos_queue[prio], (Uint32 *) PAL_CPPI4_VIRT_2_PHYS(bd), CNID_QM_DESC_SIZE_CODE, len);

    return NETDEV_TX_OK;
}

/**************************************************************************/
/*! \fn         cnid_tx_interrupt
 **************************************************************************
 *
 *  \brief      Transmit ISR
 *
 *  \param[in]  IRQ
 *  \param[in]  Device
 *  \param[in]  regs
 *  \return     OK or error
 **************************************************************************/
static irqreturn_t cnid_tx_interrupt (int irq, void *dev, struct pt_regs *regs)
{
    cnid_private_t* priv = netdev_priv((struct net_device*) dev);

    tasklet_schedule(&priv->tx_tasklet);

    return IRQ_RETVAL(1);
}

/** Rx processing functions go here **/

/**************************************************************************/
/*! \fn         cnid_do_rx_complete
 **************************************************************************
 *
 *  \brief      Rx Complete handler
 *
 *  \param[in]  Net Device
 *  \param[in]  Processed packets budget
 *  \return     Number of processed packets
 **************************************************************************/
static int cnid_do_rx_complete(struct net_device* dev, int budget)
{
    cnid_private_t* priv = netdev_priv(dev);
    cnid_desc_t* bd;
    int packets_processed = 0, i;

    /* process high priority packets first */
    for(i = 0; i < CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT; i++)
    {
        while(avalanche_intd_get_interrupt_count(CNID_INTD_HOST_NUM, priv->rxAcc_chan[i]))
        {
            while((bd = (cnid_desc_t*)((unsigned long)*priv->rxAcc_chan_list[i]
                                            & QMGR_QUEUE_N_REG_D_DESC_ADDR_MASK)))
            {
                struct sk_buff *newskb;

                bd = PAL_CPPI4_PHYS_2_VIRT(bd);

                /* get a new skb for this bd */

                if(!test_bit(0, &priv->state))
                {   /* if not cleaning up .. */

                    if((newskb = dev_alloc_skb(RX_BUF_SIZE)))
                    { /* .. and able to get a new buffer */

                        struct sk_buff* rxskb;

                        PAL_CPPI4_CACHE_INVALIDATE(bd, CPPI4_BD_LENGTH_FOR_CACHE);
                        rxskb = bd->skb;
                        skb_put(rxskb, bd->hw.buffLen - 4); /* remove CRC from length */
                        dev->last_rx = jiffies;
                        priv->stats.rx_packets++;
                        priv->stats.rx_bytes += bd->hw.buffLen;

                        /* Keep SYNC Q PTID info in skb for egress */
                        if(bd->hw.netInfoWord1)
                        {
                            memcpy(rxskb->pp_packet_info.ti_epi_header, &(bd->hw.netInfoWord0), 8);
                            rxskb->pp_packet_info.ti_pp_flags = TI_PPM_SESSION_INGRESS_RECORDED;
                        }

                        /* ... then send the packet up */
                        {
                            extern int DBridge_receive (struct sk_buff *skb);

                            /* This is the DOCSIS Interface; pass the packet to the DOCSIS Bridge; initialize the
                            * MAC RAW Pointer before doing so. */
                            rxskb->mac.raw = rxskb->data;
                            rxskb->ti_meta_info = bd->psi[0];
                            rxskb->ti_meta_info2 = bd->psi[2];
                            DBridge_receive (rxskb);
                            DPRINTK( ": %d packets received\n",(int)(priv->stats.rx_packets));
                        }

                        /* Prepare to return to free Proxy queue */
                        cnid_rx_bd_link_skb(dev, bd, newskb);
                    }
                }

                packets_processed++;
                priv->rxAcc_chan_list[i]++;
                /* Return to free queue */
                PAL_cppi4QueuePush(priv->rxProxy_free_queue_hnd[i], (Uint32 *) PAL_CPPI4_VIRT_2_PHYS(bd),
                                CNID_QM_DESC_SIZE_CODE, 0);
                /* thats it, we did enough. Jump out now! */
                if(budget == packets_processed) goto out;
            }

            /* Update the list entry for next time */
            priv->rxAcc_chan_list[i] = PAL_cppi4AccChGetNextList(priv->rxAcc_chan_hnd[i]);
            avalanche_intd_set_interrupt_count (CNID_INTD_HOST_NUM, priv->rxAcc_chan[i], 1);
        }
    }

out:
    return packets_processed;
}


/**************************************************************************/
/*! \fn         cnid_poll
 **************************************************************************
 *
 *  \brief      Polling function
 *
 *  \param[in]  Net Device
 *  \param[in]  Processed packets budget
 *  \return     Number of processed packets
 **************************************************************************/
static int cnid_poll(struct net_device *dev, int *budget)
{
    int orig_budget = min(*budget, dev->quota);
    int done = 1;
    int work_done;

    DPRINTK( " Enter \n");
    work_done = cnid_do_rx_complete(dev, orig_budget);

    if (likely(work_done > 0))
    {
        *budget -= work_done;
        dev->quota -= work_done;
        done = (work_done < orig_budget); /* "done" would be zero if we hit budget */
    }

    /* order is important here. If we do EOI before calling netif_rx_complete, an interrupt
     * can occur just before we take ourselves out of the poll list; we will not
     * schedule NAPI thread on that interrupt, no further Rx interrupts and
     * Rx will stall forever. Scary...
     * */
    if (done)
    {
        netif_rx_complete(dev);
        avalanche_intd_write_eoi (CNI_ACC_RX_INTV);
    }
#ifdef CONFIG_INET_LRO
    if( dev->features & NETIF_F_LRO ) {
        cnid_private_t* priv = netdev_priv(dev);

        lro_flush_all(&priv->lro_mgr);
    }
#endif
    return !done; /* poll should return 0 if finished, 1 if not */
}



/**************************************************************************/
/*! \fn         cnid_rx_interrupt
 **************************************************************************
 *
 *  \brief      Receive ISR
 *
 *  \param[in]  IRQ
 *  \param[in]  Device
 *  \param[in]  regs
 *  \return     OK or error
 **************************************************************************/
static irqreturn_t cnid_rx_interrupt (int irq, void *dev_instance, struct pt_regs *regs)
{
    struct net_device *dev = (struct net_device *) dev_instance;

    DPRINTK( " Enter \n");

    /* if poll routine is not running, start it now. */
    netif_rx_schedule(dev);

    return IRQ_RETVAL(1);
}

/** Driver init/open/close functions go here **/

/**************************************************************************/
/*! \fn         cnid_free_rxpool
 **************************************************************************
 *
 *  \brief      Rx Pool FREE function
 *              Note this will not rewind the queue pushes
 *              (not needed as queues are reset at start)
 *
 *  \param[in]  Net Device
 *  \param[in]  Pool Base Address
 *  \param[in]  SK buffers number
 *  \return     None
 **************************************************************************/
static void cnid_free_rxpool(struct net_device *dev, Ptr pool_base, int num_skb)
{
    cnid_desc_t *desc;
    int j;

    if(pool_base)
    {
        for(j = 0; j < num_skb; j++)
        {
            desc = GET_BD_PTR(pool_base, j);
            dev_kfree_skb_any(desc->skb);
        }
    }
}


/**************************************************************************/
/*! \fn         cnid_init_rxpool
 **************************************************************************
 *
 *  \brief      Allocate and initialize Rx BD pool
 *
 *  \param[in]  Net Device
 *  \param[in]  Queue Handle
 *  \param[in]  Queue Manager
 *  \param[in]  Descriptors number
 *  \param[in]  Queue Number
 *  \return     Pool Address
 **************************************************************************/
static int cnid_init_rxpool(struct net_device *dev, int idx)
{
    cnid_private_t* priv = netdev_priv(dev);
    Ptr ptr;
    int i=0;

    /* Allocate BD pools.. */
    ptr = (Ptr)((unsigned int)rxProxy[idx].bdPoolOffset + (unsigned int)priv->bdPoolBase) ;

    /* .. and prepare the BDs... */
    for (i = 0; i < rxProxy[idx].bdnum; i++)
    {
        struct sk_buff* skb = dev_alloc_skb(RX_BUF_SIZE);
        cnid_desc_t* bd = GET_BD_PTR(ptr, i);

        if(!skb)
        {
            DPRINTK( " Unable to allocate %dth Rx buffers\n", i);
            goto rewind;
        }

        PAL_osMemSet(bd, 0, CNI_BD_SIZE);

        cnid_rx_bd_init(bd, rxProxy[idx].freeQueue.qMgr , rxProxy[idx].freeQueue.qNum);
        cnid_rx_bd_link_skb(dev, bd, skb);

        PAL_cppi4QueuePush (priv->rxProxy_free_queue_hnd[idx] , (Ptr) PAL_CPPI4_VIRT_2_PHYS(bd), CNID_QM_DESC_SIZE_CODE, 0);
    }

    priv->rxProxy_bdpool[idx] = ptr;
    return 0;

rewind:
    cnid_free_rxpool(dev, ptr, i);
    return -1;
}


/************************************************************************/
/*                                                                      */
/*  Interrupt Accumulator Channels INIT routine                         */
/*                                                                      */
/************************************************************************/
static int cnid_init_acc_chan(PAL_Handle pal_hnd, int chan_num, Cppi4Queue queue, PAL_Cppi4AccChHnd* acc_hnd)
{
    Cppi4AccumulatorCfg cfg;
    *acc_hnd = NULL;

    cfg.accChanNum             = chan_num;
    cfg.list.maxPageEntry      = CNID_ACC_PAGE_NUM_ENTRY;        /* This is entries per page (and we have 2 pages) */
    cfg.list.listEntrySize     = CNID_ACC_ENTRY_TYPE;            /* Only interested in register 'D' which has the desc pointer */
    cfg.list.listCountMode     = 0;                              /* Zero indicates null terminated list. */
    cfg.list.pacingMode        = 1;                              /* Wait for time since last interrupt */
    cfg.pacingTickCnt          = 40;                             /* Wait for 1000uS == 1ms */
    cfg.list.maxPageCnt        = CNID_ACC_NUM_PAGE;              /* Use two pages */
    cfg.list.stallAvoidance    = 1;                              /* Use the stall avoidance feature */
    cfg.queue                  = queue;
    cfg.mode                   = 0;

    /* kmalloc returns cache line aligned memory unless you are debugging the slab allocator (2.6.18) */
    if(!(cfg.list.listBase = kzalloc(CNID_ACC_LIST_BYTE_SZ, GFP_KERNEL)))
    {
        DPRINTK(" Unable to allocate list page\n");
        return -1;
    }

    PAL_CPPI4_CACHE_WRITEBACK((unsigned long)cfg.list.listBase, CNID_ACC_LIST_BYTE_SZ);

    cfg.list.listBase = (Ptr) PAL_CPPI4_VIRT_2_PHYS((Ptr)cfg.list.listBase);

    if(!(*acc_hnd = PAL_cppi4AccChOpen(pal_hnd, &cfg)))
    {
        DPRINTK( " Unable to open accumulator channel #%d\n", chan_num);
        kfree(cfg.list.listBase);
        return -1;
    }

    return 0;
}


/**************************************************************************/
/*! \fn         cnid_open_tx
 **************************************************************************
 *
 *  \brief      Open Tx routine
 *
 *  \param[in]  Net Device
 *  \return     OK or error
 **************************************************************************/
static int cnid_open_tx(struct net_device* dev)
{
    cnid_private_t* priv = netdev_priv(dev);
    Cppi4Queue tx_cmpl  = {CNI_CPPI4x_TX_COMP_QMGR, CNI_CPPI4x_TX_COMP_QNUM(0)};
    Cppi4Queue queue;   /* used generically */
    int i = 0, j;

    /************************************************/
    /*                                              */
    /* Initialize US DOCSIS Tx queues               */
    /*                                              */
    /************************************************/
    queue.qMgr = CNI_CPPI4x_TX_QMGR;
    for(i = 0; i < CNI_CPPI4x_TX_Q_COUNT; i++)
    {
        queue.qNum = CNI_CPPI4x_TX_QNUM(i);

        if(!(priv->tx_queue[i] = PAL_cppi4QueueOpen(priv->pal_hnd, queue)))
        {
            DPRINTK( " unable to open Tx Queue #%d!\n", i);
            goto err;
        }
    }
    /************************************************/

    for (i = 0; PPFW_CPPI4x_HOST_TO_PP_PRTY_Q_COUNT > i; i++)
    {
        /************************************************/
        /*                                              */
        /* Initialize System Generic Priority Tx queues */
        /*                                              */
        /************************************************/
        queue.qMgr = PPFW_CPPI4x_HOST_TO_PP_PRTY_QMGR;
        queue.qNum = PPFW_CPPI4x_HOST_TO_PP_PRTY_QNUM(i);

        if(!(priv->tx_qos_queue[i] = PAL_cppi4QueueOpen(priv->pal_hnd, queue)))
        {
            DPRINTK(" unable to open Tx QOS Queue #%d!\n", i);
            goto err;
        }
        /************************************************/

        /************************************************/
        /*                                              */
        /* Initialize Free Host Descriptors Tx queues   */
        /*                                              */
        /************************************************/
        queue.qMgr = PPFW_CPPI4x_FDB_QMGR;
        queue.qNum = PPFW_CPPI4x_FDB_QNUM(i);

        if(!(priv->tx_qos_free_queue[i] = PAL_cppi4QueueOpen(priv->pal_hnd, queue)))
        {
            DPRINTK(" unable to open Tx Queue #%d!\n", queue.qNum);
            goto err;
        }
        /************************************************/
    }

    /* reset Tx completion queue */
    PAL_cppi4QueueClose(priv->pal_hnd, PAL_cppi4QueueOpen(priv->pal_hnd, tx_cmpl));

    tasklet_init(&priv->tx_tasklet, cnid_do_tx_complete, (unsigned long) dev);

    /* Init the Tx complete accumulator channel */
    if(cnid_init_acc_chan(priv->pal_hnd, CNI_ACC_TXCMPL_CHNUM, tx_cmpl, &priv->tx_acc_hnd))
    {
        DPRINTK(KERN_ERR "%s: unable to open accumulator channel!\n", __FUNCTION__);
        goto err;
    }

    priv->tx_list_base = priv->tx_list = PAL_cppi4AccChGetNextList(priv->tx_acc_hnd);

    /************************************************/
    /*                                              */
    /*  Open DOCSIS Upstream MAC/PHY DMA channels   */
    /*                                              */
    /************************************************/
    for (i = 0; i < CNI_CPPI4x_TX_DMA_CH_COUNT; i++)
    {
        volatile Cppi4TxChInitCfg  txchInfo;
        PAL_Cppi4TxChHnd           txchHdl;

        txchInfo.chNum          = CNI_CPPI4x_TX_DMA_CHNUM(i);
        txchInfo.dmaNum         = PAL_CPPI41_DMA_BLOCK1;
        txchInfo.tdQueue.qMgr   = DMA1_CPPI4x_FTD_QMGR;
        txchInfo.tdQueue.qNum   = DMA1_CPPI4x_FTD_QNUM;

        if (NULL == (txchHdl = PAL_cppi4TxChOpen( priv->pal_hnd, (Cppi4TxChInitCfg *)(&txchInfo), NULL )))
        {
            DPRINTK(" Unable to open %d channel \n",txchInfo.chNum);
            goto err;
        }

        PAL_cppi4EnableTxChannel (txchHdl, NULL);
    }
    /************************************************/

#ifdef INIT_EXT_DMA
        {
             volatile Cppi4RxChInitCfg  rxchInfo;
             PAL_Cppi4RxChHnd           rxchHdl;
             volatile Cppi4TxChInitCfg  txchInfo;
             PAL_Cppi4TxChHnd           txchHdl;
             Cppi4Queue                 fdQueue = {1,60};   //cable_pp: QMGR = 1 !!!!!!!

        /* init co-proc TX channel */
            txchInfo.chNum          = CNI_CPPI4x_TX_CoP_DMA_CHNUM(0);
            txchInfo.dmaNum         = PAL_CPPI41_DMA_BLOCK1;
            txchInfo.tdQueue.qMgr   = DMA1_CPPI4x_FTD_QMGR;
            txchInfo.tdQueue.qNum   = DMA1_CPPI4x_FTD_QNUM;

            if (NULL == (txchHdl = PAL_cppi4TxChOpen( priv->pal_hnd, (Cppi4TxChInitCfg *)(&txchInfo), NULL )))
            {
                DPRINTK(" Unable to open %d channel \n",txchInfo.chNum);
                goto err;
            }

            PAL_cppi4EnableTxChannel (txchHdl, NULL);

        /* init co-proc RX channel qman 1 queue */
            rxchInfo.chNum          = CNI_CPPI4x_RX_CoP_DMA_CHNUM(0);
            rxchInfo.dmaNum         = PAL_CPPI41_DMA_BLOCK1;
            rxchInfo.defDescType    = CPPI41_DESC_TYPE_EMBEDDED;
            rxchInfo.sopOffset=0;
            rxchInfo.retryOnStarvation = 0;

            rxchInfo.u.embeddedPktCfg.fdQueue = fdQueue;
            rxchInfo.u.embeddedPktCfg.numBufSlot = 2;
            rxchInfo.u.embeddedPktCfg.sopSlotNum = 0;
            rxchInfo.u.embeddedPktCfg.fBufPool[0].bMgr  = BUF_POOL_MGR0;
            rxchInfo.u.embeddedPktCfg.fBufPool[0].bPool = BMGR0_POOL12;
            rxchInfo.u.embeddedPktCfg.fBufPool[1].bMgr  = BUF_POOL_MGR0;
            rxchInfo.u.embeddedPktCfg.fBufPool[1].bPool = BMGR0_POOL12;
            rxchInfo.u.embeddedPktCfg.fBufPool[2].bMgr  = BUF_POOL_MGR0;
            rxchInfo.u.embeddedPktCfg.fBufPool[2].bPool = BMGR0_POOL12;
            rxchInfo.u.embeddedPktCfg.fBufPool[3].bMgr  = BUF_POOL_MGR0;
            rxchInfo.u.embeddedPktCfg.fBufPool[3].bPool = BMGR0_POOL12;

            if (NULL == (rxchHdl = PAL_cppi4RxChOpen (priv->pal_hnd, (Cppi4RxChInitCfg *)(&rxchInfo), NULL)))
            {
                DPRINTK(" Unable to open %d channel \n",rxchInfo.chNum);
                goto err;
            }

            PAL_cppi4EnableRxChannel (rxchHdl, NULL);
        }
#endif

    /* request the Tx IRQs */
    if(request_irq (CNID_TXINT_NUM, cnid_tx_interrupt, IRQF_DISABLED, dev->name, dev))
    {
        DPRINTK(KERN_ERR "%s: unable to get IRQ #%d!\n", __FUNCTION__, CNID_TXINT_NUM);
        goto err;
    }

    return 0;

err:
    for(j = 0; j < i ; j++)
    {
        queue.qNum = CNI_CPPI4x_TX_QNUM(j);
        PAL_cppi4QueueClose(priv->pal_hnd, priv->tx_queue[j]);
    }

    return -1;
}


/**************************************************************************/
/*! \fn         cnid_open_rx
 **************************************************************************
 *
 *  \brief      Open Rx routine
 *
 *  \param[in]  Net Device
 *  \return     Ok or error
 **************************************************************************/
static int cnid_open_rx(struct net_device* dev)
{
    cnid_private_t* priv = netdev_priv(dev);

    /*
     * Prepare DOCSIS proxy channels
     * =============================
     */
    {
        int iProxyChan;

        /*
         * Open Docsis Proxy Tx channels
         * -----------------------------
         * Input queues are predefined from Q[222]..
         * TxCompl queues are from Q[122+2]
         */
        for (iProxyChan=0; iProxyChan<CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT; iProxyChan++)
        {
            volatile Cppi4TxChInitCfg ProxyTxChInfo;
            volatile Cppi4RxChInitCfg ProxyRxChInfo;
            PAL_Cppi4TxChHnd ProxyTxChHdl;
            PAL_Cppi4RxChHnd ProxyRxChHdl;

            ProxyTxChInfo.chNum         = CNI_CPPI4x_DOC2HOST_PROXY_CHNUM(iProxyChan);
            ProxyTxChInfo.dmaNum        = PAL_CPPI41_DMA_BLOCK1;
            ProxyTxChInfo.tdQueue.qMgr  = DMA1_CPPI4x_FTD_QMGR;
            ProxyTxChInfo.tdQueue.qNum  = DMA1_CPPI4x_FTD_QNUM;
            ProxyTxChInfo.defDescType   = CPPI41_DESC_TYPE_EMBEDDED;

            DPRINTK(" Call PAL_cppi4TxChOpen channel=%d\n", ProxyTxChInfo.chNum);

            ProxyTxChHdl = PAL_cppi4TxChOpen (priv->pal_hnd, (Cppi4TxChInitCfg *)(&ProxyTxChInfo), NULL);
            if(ProxyTxChHdl == NULL)
            {
                DPRINTK(" Unable to open %d channel \n", ProxyTxChInfo.chNum);
                goto err;
            }
            PAL_cppi4EnableTxChannel (ProxyTxChHdl, NULL);

        /*
         * Prepare Proxy rx queues
         * -----------------------
         * Free queues are from   Q[128+8]
         * Rx complete queues are Q[100+5]
         */
            if(!(priv->rxProxy_queue_hnd[iProxyChan] = PAL_cppi4QueueOpen(priv->pal_hnd, rxProxy[iProxyChan].outputQueue)))
            {
                DPRINTK(" Unable to open queue %d \n", rxProxy[iProxyChan].outputQueue.qNum);
                goto err;
            }

            if(!(priv->rxProxy_free_queue_hnd[iProxyChan] = PAL_cppi4QueueOpen(priv->pal_hnd, rxProxy[iProxyChan].freeQueue)))
            {
                DPRINTK(" Unable to open free desc queue %d\n", rxProxy[iProxyChan].freeQueue.qNum);
                goto err;
            }

            if (cnid_init_rxpool(dev, iProxyChan))
            {
                DPRINTK(" Unable to init BD pool for %d priority channel\n", iProxyChan);
                goto err;
            }

        /*
         * Open Docsis Proxy Rx channels
         * -----------------------------
         * Free queues are from Q[128+8]
         * Output queues are    Q[100+5]
         */

            ProxyRxChInfo.chNum     = CNI_CPPI4x_DOC2HOST_PROXY_CHNUM(iProxyChan);
            ProxyRxChInfo.dmaNum    = PAL_CPPI41_DMA_BLOCK1;
            ProxyRxChInfo.rxCompQueue.qMgr = rxProxy[iProxyChan].outputQueue.qMgr;
            ProxyRxChInfo.rxCompQueue.qNum = rxProxy[iProxyChan].outputQueue.qNum;
            ProxyRxChInfo.sopOffset = 0; // SOF skip=0
            ProxyRxChInfo.defDescType = CPPI41_DESC_TYPE_HOST;
            ProxyRxChInfo.retryOnStarvation = 0;
            ProxyRxChInfo.u.hostPktCfg.fdbQueue[0] = rxProxy[iProxyChan].freeQueue;
            ProxyRxChInfo.u.hostPktCfg.fdbQueue[1] = rxProxy[iProxyChan].freeQueue;
            ProxyRxChInfo.u.hostPktCfg.fdbQueue[2] = rxProxy[iProxyChan].freeQueue;
            ProxyRxChInfo.u.hostPktCfg.fdbQueue[3] = rxProxy[iProxyChan].freeQueue;

            DPRINTK(" Call PAL_cppi4RxChOpen channel=%d\n", ProxyRxChInfo.chNum);

            ProxyRxChHdl = PAL_cppi4RxChOpen (priv->pal_hnd, (Cppi4RxChInitCfg *)(&ProxyRxChInfo), NULL);
            if(ProxyRxChHdl == NULL)
            {
                DPRINTK(" Unable to open %d channel \n", ProxyRxChInfo.chNum);
                goto err;
            }
            PAL_cppi4EnableRxChannel (ProxyRxChHdl, NULL);
        }
    }

    /***************************************************************
     * Prepare Accumulator channels
     * ============================
     */
    {
        int iAccChan;

        for(iAccChan=0; iAccChan < CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT; iAccChan++)
        {
            /*
             * Open Docsis Accumulator channels
             * --------------------------------
             * Acc channels are 28..30
             * !!!! queues are the same as rxProxy queues, Q[100+5]
             */
            priv->rxAcc_chan_hnd[iAccChan] = NULL;
            priv->rxAcc_chan    [iAccChan] = CNI_ACC_RX_CHNUM(iAccChan);

            if( cnid_init_acc_chan( priv->pal_hnd,
                                    priv->rxAcc_chan[iAccChan],
                                    rxProxy[iAccChan].outputQueue,
                                   &priv->rxAcc_chan_hnd[iAccChan] ) )
            {
                DPRINTK(" Unable to open accumulator channel for %d priority channel\n", iAccChan);
                goto err;
            }

            priv->rxAcc_chan_list_base[iAccChan] = priv->rxAcc_chan_list[iAccChan] = PAL_cppi4AccChGetNextList(priv->rxAcc_chan_hnd[iAccChan]);
        }
    }

    /***************************************************************
     * Prepare CNI Rx channels
     * ====================
     */
    {
        int iCniChan;

        /*
         * Prepare Cni Rx queues
         * ---------------------
         * Free queues are from Q[144+7]..
         * Rx complete queues are from Q[56].. - Prefetcher queues
         */
        for (iCniChan=0; iCniChan<CNI_CPPI4x_RX_DMA_CH_COUNT; iCniChan++)
        {
        /*
         * Open Docsis Cni Rx channels
         * -----------------------------
         * Free queues are from   Q[128+0]
         * Output queues are from Q[100+5]
         */
            {
                volatile Cppi4RxChInitCfg CniRxChInfo;
                PAL_Cppi4RxChHnd CniRxChHdl;

                CniRxChInfo.chNum               = rxIngress[iCniChan].dmaChannel;
                CniRxChInfo.dmaNum              = PAL_CPPI41_DMA_BLOCK1;
                CniRxChInfo.rxCompQueue.qMgr    = rxIngress[iCniChan].outputQueue.qMgr;
                CniRxChInfo.rxCompQueue.qNum    = rxIngress[iCniChan].outputQueue.qNum;
                CniRxChInfo.sopOffset = 0; // SOF skip=0
                CniRxChInfo.retryOnStarvation = 0;

                CniRxChInfo.defDescType = CPPI41_DESC_TYPE_EMBEDDED;
                CniRxChInfo.u.embeddedPktCfg.fdQueue.qMgr = rxIngress[iCniChan].freeQueue.qMgr;
                CniRxChInfo.u.embeddedPktCfg.fdQueue.qNum = rxIngress[iCniChan].freeQueue.qNum;
                CniRxChInfo.u.embeddedPktCfg.numBufSlot = (EMSLOTCNT-1);
                CniRxChInfo.u.embeddedPktCfg.sopSlotNum = 1;

                CniRxChInfo.u.embeddedPktCfg.fBufPool[0].bMgr   = BUF_POOL_MGR0;
                CniRxChInfo.u.embeddedPktCfg.fBufPool[1].bMgr   = BUF_POOL_MGR0;
                CniRxChInfo.u.embeddedPktCfg.fBufPool[2].bMgr   = BUF_POOL_MGR0;
                CniRxChInfo.u.embeddedPktCfg.fBufPool[3].bMgr   = BUF_POOL_MGR0;
                switch (iCniChan)
                {
                    case 0:
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[0].bPool  = BMGR0_POOL00;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[1].bPool  = BMGR0_POOL00;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[2].bPool  = BMGR0_POOL04;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[3].bPool  = BMGR0_POOL00;
                    break;

                    case 1:
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[0].bPool  = BMGR0_POOL01;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[1].bPool  = BMGR0_POOL01;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[2].bPool  = BMGR0_POOL04;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[3].bPool  = BMGR0_POOL00;
                    break;

                    case 2:
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[0].bPool  = BMGR0_POOL05;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[1].bPool  = BMGR0_POOL05;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[2].bPool  = BMGR0_POOL05;
                        CniRxChInfo.u.embeddedPktCfg.fBufPool[3].bPool  = BMGR0_POOL00;
                    break;
                }

                DPRINTK(" Call PAL_cppi4RxChOpen channel=%d\n", CniRxChInfo.chNum);

                CniRxChHdl = PAL_cppi4RxChOpen (priv->pal_hnd, (Cppi4RxChInitCfg *)(&CniRxChInfo), NULL);
                if(CniRxChHdl == NULL)
                {
                    DPRINTK(" Unable to open %d channel \n", CniRxChInfo.chNum);
                    goto err;
                }

                PAL_cppi4EnableRxChannel (CniRxChHdl, NULL);

                DPRINTK(" Call PAL_cppi4EnableRxChannel channel=%d Done\n", CniRxChInfo.chNum);
            }
        }
    }

    /* request the Rx IRQs */
    if(request_irq (CNID_RXINT_NUM, cnid_rx_interrupt, IRQF_DISABLED, dev->name, dev))
    {
        DPRINTK(" unable to get IRQ #%d !\n", CNID_RXINT_NUM);
        goto err;
    }

    return 0;

err:
    cnid_close_rx_prep(dev);
    cnid_close_rx_finish(dev);

    return -1;
}


/**************************************************************************/
/*! \fn         cnid_open
 **************************************************************************
 *
 *  \brief      CNI Device Open API
 *
 *  \param[in]  Net Device
 *  \return     Ok
 **************************************************************************/
static int cnid_open (struct net_device *dev)
{
    DPRINTK( " Enter \n");

#ifdef CONFIG_ARM_AVALANCHE_PPD
    cni_pp_set_pid_flags(dev, 0);
#endif

    if(cnid_open_tx(dev)) return -1;
    if(cnid_open_rx(dev)) return -1;

    netif_start_queue (dev);

    return 0;
}


/************************************************************************/
/*                                                                      */
/*                                                                      */
/*                                                                      */
/************************************************************************/
static int cnid_close_tx_prep(struct net_device* dev)
{
    struct cnid_private_t* priv = netdev_priv(dev);

    /* de-init the Tx complete accumulator channel */
    PAL_cppi4AccChClose(priv->tx_acc_hnd, NULL);

    /* disable the Tx IRQs */
    disable_irq (CNID_TXINT_NUM);

    return 0;
}
/*----------------------------------------------------------------------*/
static int cnid_close_tx_finish(struct net_device* dev)
{
    cnid_private_t* priv = netdev_priv(dev);
    int i;

    /* de-init the Tx queues */
    for(i = 0; i < CNI_CPPI4x_TX_Q_COUNT; i++)
    {
        PAL_cppi4QueueClose(priv->pal_hnd, priv->tx_queue[i]);
    }

    tasklet_kill(&priv->tx_tasklet);

    kfree(priv->tx_list_base);

    free_irq (CNID_TXINT_NUM, dev);

    return 0;
}
/************************************************************************/


/************************************************************************/
/*                                                                      */
/*                                                                      */
/*                                                                      */
/************************************************************************/
static int cnid_close_rx_prep(struct net_device* dev)
{
    cnid_private_t* priv = netdev_priv(dev);
    int i;

    /* disable the Rx IRQs */
    disable_irq (CNID_RXINT_NUM);

    for(i = 0; i < CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT; i++)
    {
        PAL_cppi4AccChClose(priv->rxAcc_chan_hnd[i], NULL);
    }

    return 0;
}
/*----------------------------------------------------------------------*/
static int cnid_close_rx_finish(struct net_device* dev)
{
    int i;
    cnid_private_t* priv = netdev_priv(dev);

    for(i=0; i < CNI_CPPI4x_DOC2HOST_PROXY_CH_COUNT; i++)
    {
        if(priv->rxProxy_free_queue_hnd[i])
        {
            PAL_cppi4QueueClose(priv->pal_hnd, priv->rxProxy_free_queue_hnd[i]);
        }

        if(priv->rxProxy_bdpool[i])
        {
            cnid_free_rxpool( dev,
                              priv->rxProxy_bdpool[i],
                              rxProxy[i].bdnum );
        }

        if(priv->rxAcc_chan_hnd[i])
        {
            kfree(priv->rxAcc_chan_list_base[i]);
        }
    }

    /* free the Rx IRQs */
    free_irq (CNID_RXINT_NUM, dev);

    return 0;
}
/************************************************************************/


static int cnid_close (struct net_device *dev)
{
    cnid_private_t* priv = netdev_priv(dev);
    unsigned long flags;

    netif_stop_queue(dev);

#ifdef CONFIG_ARM_AVALANCHE_PPD
    cni_pp_set_pid_flags(dev, TI_PP_PID_DISCARD_ALL_RX);
#endif

    /*
     * Do the urgent stuff first
     */
    cnid_close_tx_prep(dev);
    cnid_close_rx_prep(dev);

    spin_lock_irqsave(&priv->devlock, flags);
    set_bit(0, &priv->state);

    if(cnid_do_rx_complete( dev,
                            rxProxy[0].bdnum +
                            rxProxy[1].bdnum +
                            rxProxy[2].bdnum ))
    {
        avalanche_intd_write_eoi(CNI_ACC_RX_INTV);
    }

    cnid_do_tx_complete((unsigned long)dev);

    clear_bit(0, &priv->state);
    spin_unlock_irqrestore(&priv->devlock, flags);

    /*
     * finish rest of janitorial stuff now
     */
    cnid_close_tx_finish(dev);
    cnid_close_rx_finish(dev);

    return 0;
}


static int cnid_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
    return 0;
}


static struct net_device_stats *cnid_get_stats (struct net_device *dev)
{
    cnid_private_t* priv = netdev_priv((struct net_device*) dev);

    return &priv->stats;
}


static void cnid_set_multicast (struct net_device *dev)
{

}

#if 0
static void cnid_tx_timeout (struct net_device *dev)
{
    cnid_private_t* priv = netdev_priv(dev);
    unsigned long flags;

    netif_stop_queue(dev);

    spin_lock_irqsave(&priv->devlock, flags);
    set_bit(0, &priv->state);

    /* reset Tx processing */
    cnid_close_tx_prep(dev);

    cnid_close_tx_finish(dev);

    cnid_open_tx(dev);

    clear_bit(0, &priv->state);
    spin_unlock_irqrestore(&priv->devlock, flags);

}
#endif

/* TODO: Hack this function to setup device address according to taste */
static void cnid_hard_addr_setup(struct net_device *dev)
{
    memcpy(dev->dev_addr, defmac, dev->addr_len);
    memcpy(dev->perm_addr, dev->dev_addr, dev->addr_len);
}


static void cnid_netdev_setup(struct net_device *dev)
{
    dev->open               =   cnid_open;
    dev->hard_start_xmit    =   cnid_start_xmit;
    dev->poll               =   cnid_poll;
    dev->weight             =   CIND_RX_SERVICE_MAX;
    dev->stop               =   cnid_close;
    dev->get_stats          =   cnid_get_stats;
    dev->set_multicast_list =   cnid_set_multicast;
    dev->do_ioctl           =   cnid_ioctl;
   // dev->tx_timeout         =   cnid_tx_timeout;

    /* TODO: study the effects of these features */
    /* dev->features |= NETIF_F_SG | NETIF_F_HW_CSUM | NETIF_F_HIGHDMA */

    ether_setup(dev);
    cnid_hard_addr_setup(dev);
}


/* structure describing the CNID driver */
static struct device_driver cnid_driver = {
    .name       = "cni",
    .bus        = NULL,
    .probe      = cnid_probe,
    .remove     = cnid_remove,
    .suspend    = NULL,
    .resume     = NULL,
};


static int
cnid_read_proxy(char *buf, char **start, off_t offset, int count,
                  int *eof, void *data)
{
    struct net_device * dev = (struct net_device *) data;
    cnid_desc_t*        currBD = NULL;
    cnid_private_t*     priv = netdev_priv(dev);
    int len = 0;

    len += sprintf(buf + len, "cni proxy \n");

    currBD = (cnid_desc_t*) PAL_cppi4QueuePop(priv->rxProxy_free_queue_hnd[0]);
    len += sprintf(buf + len, "read %x from %x \n",(unsigned int)currBD, (unsigned int)(priv->rxProxy_free_queue_hnd[0]));
    if(currBD != NULL)
    {

        currBD->hw.bufPtr = currBD->hw.orgBufPtr;
        currBD->hw.buffLen = 20;

        PAL_cppi4QueuePush (priv->rxProxy_queue_hnd[0],(Ptr) PAL_CPPI4_VIRT_2_PHYS(currBD),
                                  CNID_QM_DESC_SIZE_CODE, 0);

    }

    return(len);

}



#ifdef CONFIG_ARM_AVALANCHE_PPD

/*
 * CNI PID handles
 * ---------------
 */

/* PID/VPID definitions */
#define PP_CNI_SR_DELAY     200
#include <asm-arm/arch-avalanche/puma5/puma5_pp.h> /* For PID base config */
#ifndef CONFIG_TI_PACKET_PROCESSOR
static TI_PP_PID cni_pid_list[PP_CNI_PID_COUNT];
#endif

/*
 * Set PID Flags
 * -------------
 */
static int cni_pp_set_pid_flags(struct net_device *dev, int flags)
{
#ifdef CONFIG_TI_PACKET_PROCESSOR
    ti_ppm_set_pid_flags (dev->pid_handle, flags);
#else
    ti_ppd_set_pid_flags (&cni_pid_list[0], flags);
#endif
    /* this delay is to make sure all the packets with the PID successfully egress throgh the respective ports.*/
    mdelay(PP_CNI_SR_DELAY);
    return 0;
}

static int cni_select_qos(struct sk_buff *skb)
{
    unsigned int sfIndex = (skb->ti_meta_info >> 24);

    if (8 > sfIndex)
    {
        skb->pp_packet_info.ti_session.cluster = sfIndex;
        skb->pp_packet_info.ti_session.priority = 1; /* Low Priority */
    }
    else
    {
        skb->pp_packet_info.ti_session.cluster = 0xFF;
        skb->pp_packet_info.ti_session.priority = 0; /* Default */
    }

    return 0;
}

static TI_PP_QOS_CLST_CFG  cni_qos_cluster_db[CNI_CPPI4x_QoS_CLUSTER_COUNT];

#define MAX_IP_PACKET_SIZE  1514
#define MIN_IP_PACKET_SIZE  64
static int cni_setup_qos (struct net_device *dev)
{
    int rc;
    unsigned int        clustIndex;
    TI_PP_QOS_QUEUE     *qcfg;

    for (clustIndex=CNI_CPPI4x_QoS_CLUSTER_IDX; clustIndex < CNI_CPPI4x_QoS_CLUSTER_COUNT; clustIndex++)
    {
        //
        // Setup 2 QOS queues, one that gets linespeed and then trickles down to the other.
        //
        cni_qos_cluster_db[clustIndex].qos_q_cnt  = 2;

        // Queue 0
        qcfg = &cni_qos_cluster_db[clustIndex].qos_q_cfg[0];
        qcfg->q_num         = CNI_CPPI4x_QoS_HIGH_TX_QNUM(clustIndex) - PAL_CPPI41_QPDSP_FW_Q_BASE;
        qcfg->flags         = 0;
        qcfg->egr_q         = CNI_CPPI4x_TX_QNUM(1 + clustIndex);
        qcfg->it_credit     = (120*1024*1024)/40000/8; /* <link speed> / <PP ticks per sec> / <8 bits in byte> */
        qcfg->max_credit    = MIN_IP_PACKET_SIZE * 378;
        qcfg->congst_thrsh  = MIN_IP_PACKET_SIZE * 378;
        qcfg->congst_thrsh_packets = 0;

        // Queue 1
        qcfg = &cni_qos_cluster_db[clustIndex].qos_q_cfg[1];
        qcfg->q_num         = CNI_CPPI4x_QoS_LOW_TX_QNUM(clustIndex) - PAL_CPPI41_QPDSP_FW_Q_BASE;
        qcfg->flags         = 0;
        qcfg->egr_q         = CNI_CPPI4x_TX_QNUM(1 + clustIndex);
        qcfg->it_credit     = 0;
        qcfg->max_credit    = MAX_IP_PACKET_SIZE * 16;
        qcfg->congst_thrsh  = MAX_IP_PACKET_SIZE * 20;
        qcfg->congst_thrsh_packets = 0;

        // Cluster 0
        cni_qos_cluster_db[clustIndex].global_credit      = 0;
        cni_qos_cluster_db[clustIndex].max_global_credit  = MAX_IP_PACKET_SIZE * 108;
        cni_qos_cluster_db[clustIndex].egr_congst_thrsh1  = MAX_IP_PACKET_SIZE / 2;
        cni_qos_cluster_db[clustIndex].egr_congst_thrsh2  = MAX_IP_PACKET_SIZE * 3.5;
        cni_qos_cluster_db[clustIndex].egr_congst_thrsh3  = MAX_IP_PACKET_SIZE * 4;
        cni_qos_cluster_db[clustIndex].egr_congst_thrsh4  = MAX_IP_PACKET_SIZE * 6;
        cni_qos_cluster_db[clustIndex].egr_congst_thrsh_packets  = 0;

        rc = ti_ppm_qos_cluster_disable (clustIndex);
        rc = ti_ppm_qos_cluster_setup   (clustIndex, &cni_qos_cluster_db[clustIndex]);
        rc = ti_ppm_qos_cluster_enable  (clustIndex);

        dev->vpid_block.qos_cluster[clustIndex]  = &cni_qos_cluster_db[clustIndex];
    }

    dev->vpid_block.qos_clusters_count = clustIndex;

    return rc;
}


static int cni_shutdown_qos (struct net_device *dev)
{
    int rc;
    unsigned int        clustIndex;

    for (clustIndex=CNI_CPPI4x_QoS_CLUSTER_IDX; clustIndex < CNI_CPPI4x_QoS_CLUSTER_COUNT; clustIndex++)
    {
        rc = ti_ppm_qos_cluster_disable (clustIndex);
    }

    dev->vpid_block.qos_clusters_count = 0;

    return rc;
}


/*
 * Create CNI PID range, PID, VPID
 * -------------------------------
 * Defualt Q when no match - 222 (proxy ch queue)
 * Defualt Q when match - 182 (DOS US)
 */
static int cni_pp_prepare_pid(struct net_device *dev)
{
    TI_PP_PID_RANGE  pid_range_cni;
#ifdef CONFIG_TI_PACKET_PROCESSOR
    TI_PP_PID        cni_pid_list[PP_CNI_PID_COUNT];
#endif
    int ret_val;
    int iPid;

    /*
     * Config CNI PID range
     * --------------------
     */
    pid_range_cni.type        = TI_PP_PID_TYPE_DOCSIS;
    pid_range_cni.port_num    = CPPI41_SRCPORT_DOCSISMACPHY;
    pid_range_cni.count       = PP_CNI_PID_COUNT;
    pid_range_cni.base_index  = PP_CNI_PID_BASE;

#ifdef CONFIG_TI_PACKET_PROCESSOR
    if (ti_ppm_config_pid_range (&pid_range_cni))
#else
    if (ti_ppd_config_pid_range (&pid_range_cni))
#endif
        DPRINTK (" config_pid_range failed\n");

    /*
     * Create CNI PIDs
     * ---------------
     */
    // for (iPid=0; iPid<PP_CNI_PID_COUNT; iPid++) /* prepare for couple of CNI PIDs */
    iPid = 0;
    {
        cni_pid_list[iPid].type            = TI_PP_PID_TYPE_DOCSIS;
        cni_pid_list[iPid].ingress_framing = TI_PP_PID_INGRESS_ETHERNET
                                           | TI_PP_PID_INGRESS_IPV6
                                           | TI_PP_PID_INGRESS_IPV4;
        cni_pid_list[iPid].pri_mapping     = 1;    /* Num prio Qs for fwd */
        cni_pid_list[iPid].dflt_pri_drp    = 0;
        cni_pid_list[iPid].dflt_dst_tag    = 0x3FFF;
        cni_pid_list[iPid].dflt_fwd_q      = CNI_CPPI4x_DOCSIS_TO_HOST_PROXY_QNUM(0);
        cni_pid_list[iPid].tx_pri_q_map[0] = CNI_CPPI4x_TX_QNUM(1); /* default Q for Egress */
        cni_pid_list[iPid].tx_hw_data_len  = 0;
        cni_pid_list[iPid].pid_handle      = PP_CNI_PID_BASE+iPid;

#ifdef CONFIG_TI_PACKET_PROCESSOR
        if ((ret_val = ti_ppm_create_pid (&cni_pid_list[iPid])) < 0)
#else
        if ((ret_val = ti_ppd_create_pid (&cni_pid_list[iPid])))
#endif
        {
            DPRINTK (" create_pid failed with error code %d.\n", ret_val);
            cni_pid_list[iPid].pid_handle = -1;
        }

        dev->pid_handle = cni_pid_list[iPid].pid_handle;
    }

    /*
     * Create CNI VPIDs
     * ----------------
     */
#ifdef CONFIG_TI_PACKET_PROCESSOR
    dev->vpid_block.type               = TI_PP_ETHERNET;
    dev->vpid_block.parent_pid_handle  = dev->pid_handle;
    dev->vpid_block.egress_mtu         = 0;
    dev->vpid_block.priv_tx_data_len   = 0;
    dev->qos_setup_hook                = cni_setup_qos;
    dev->qos_shutdown_hook             = cni_shutdown_qos;
    dev->qos_select_hook               = cni_select_qos;
#endif
    return 0;
}

#endif




unsigned int    cppiQueuePop(int qnum)
{
    volatile unsigned int* qPopAddr = (volatile unsigned int *)0xD906000C;

    qPopAddr += (qnum * 4);

    return *qPopAddr;
}



/**************************************************************************
 * FUNCTION NAME : cppi_write_cmds
 **************************************************************************
 * DESCRIPTION   :
 *  This is used to debug and display various
 *  CPPI entity information from the console.
 *
 * RETURNS       :
 *  -1              - Error.
 *  Non-Zero        - Success.
 ***************************************************************************/
static int cppi_write_cmds (struct file *file, const char *buffer, unsigned long count, void *data)
{
    char    proc_cmd[100];
    char*   argv[10];
    int     argc = 0;
    char*   ptr_cmd;
    char*   delimitters = " \n\t";
    char*   ptr_next_tok;

    /* Validate the length of data passed. */
    if (count > 100)
        count = 100;

    /* Initialize the buffer before using it. */
    memset ((void *)&proc_cmd[0], 0, sizeof(proc_cmd));
    memset ((void *)&argv[0], 0, sizeof(argv));

    /* Copy from user space. */
    if (copy_from_user (&proc_cmd, buffer, count))
        return -EFAULT;

    ptr_next_tok = &proc_cmd[0];

    /* Tokenize the command. Check if there was a NULL entry. If so be the case the
     * user did not know how to use the entry. Print the help screen. */
    ptr_cmd = strsep(&ptr_next_tok, delimitters);
    if (ptr_cmd == NULL)
        return -1;

    /* Parse all the commands typed. */
    do
    {
        /* Extract the first command. */
        argv[argc++] = ptr_cmd;

        /* Validate if the user entered more commands.*/
        if (argc >=10)
        {
            printk ("ERROR: Incorrect too many parameters dropping the command\n");
            return -EFAULT;
        }

        /* Get the next valid command. */
        ptr_cmd = strsep(&ptr_next_tok, delimitters);
    } while (ptr_cmd != NULL);

    /* We have an extra argument when strsep is used instead of strtok */
    argc--;

    /******************************* Command Handlers *******************************/

    /* Display Command Handlers */
    if (strncmp(argv[0], "show", strlen("show")) == 0)
    {
        if (strncmp(argv[1], "mem", strlen("mem")) == 0)
        {
            unsigned int dmem = simple_strtol(argv[2], NULL, 16);
            int i;

            for (i=0; i<16; i++)
            {
                printk("[0x%08X]: 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n", dmem+i*16,
                    *((unsigned long *)(dmem+i*16)),
                    *((unsigned long *)(dmem+i*16+4)),
                    *((unsigned long *)(dmem+i*16+8)),
                    *((unsigned long *)(dmem+i*16+12))
                    );
            }
        }
    }

    if (strncmp(argv[0], "pop", strlen("pop")) == 0)
    {
        unsigned int desc;
        int queue = (int) simple_strtol(argv[2], NULL, 0);
        int entries = 1;
        int i;

        if (argc >= 4)
        {
            entries = (int) simple_strtol(argv[3], NULL, 0);
        }

        do
        {
            desc = cppiQueuePop(queue);

            printk(" POP Descriptor Q[%3d] = [0x%08x] \n", queue, (unsigned int)(desc));

            if (desc >= 0x80000000)
            {
                desc &= 0xFFFFFFe0;
                desc = (unsigned int)PAL_CPPI4_PHYS_2_VIRT(desc);

                for (i=0; i<4; i++)
                {
                    printk("\t0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",
                        *((unsigned long *)(desc+i*16)),
                        *((unsigned long *)(desc+i*16+4)),
                        *((unsigned long *)(desc+i*16+8)),
                        *((unsigned long *)(desc+i*16+12))
                        );
                }
            }
            entries--;
        } while(desc && entries);
    }

    if (strncmp(argv[0], "repeat", strlen("repeat")) == 0)
    {
        /* repeat <iterations> <copy> <len> <src> <dst> */
        int iterations = (int) simple_strtol(argv[1], NULL, 0);

        if (strncmp(argv[2], "copy", strlen("copy")) == 0)
        {
            unsigned int * src = (unsigned int *)simple_strtol(argv[4], NULL, 16);
            unsigned int * dst = (unsigned int *)simple_strtol(argv[5], NULL, 16);
            unsigned int * tmp_1;
            unsigned int * tmp_2;
                     int len = (int) simple_strtol(argv[3], NULL, 0);
            printk (" repeat: %d copy from 0x%08X to 0x%08X (len=%d)\n", iterations, (unsigned int)src, (unsigned int)dst, len);

            do
            {
                printk("#");
                tmp_1 = (src + len - 1);
                tmp_2 = (dst + len - 1);

                while (tmp_1 >= src)
                {
                    *tmp_2-- = *tmp_1--;
                }

            } while (iterations--);


        }

        printk("\n repeat DONE !!\n");
    }

    return count;
}

#ifdef CONFIG_INET_LRO

static int cnid_get_skb_hdr(struct sk_buff *skb, void **iphdr,
               void **tcph, u64 *hdr_flags, void *data)
{
    /* Check that this is an ethernet packet */
    if(skb->protocol != ntohs(ETH_P_IP) )
    {
        return -1;
    }

    if( !skb->nh.iph )
    {
        if( !skb->mac.raw )
        {
            return -1;
        }
        /* In case the pointers are not initialized */
        skb->nh.iph = (struct iphdr *)(skb->mac.raw + ETH_HLEN);
        skb->h.th = (struct tcphdr *)(( (unsigned char *)(skb->nh.iph)+skb->nh.iph->ihl*4));
    }

    /* Continue only if its TCP */
    if( skb->nh.iph->protocol != IPPROTO_TCP )
    {
        return -1;
    }

    if(skb->nh.iph->version == 4)
        *hdr_flags = LRO_IPV4;

    *tcph = (void *)skb->h.th;
    *iphdr = (void *)skb->nh.iph;
    *hdr_flags |= LRO_TCP;

    return 0;
}
#endif

/* TODO: in a fully linux framework compliant driver,
 * we will receive the resources: io, mem and intr
 * in the device structure.
 */
static int __devinit cnid_probe(struct device *dev)
{
    struct net_device *netdev = NULL;
    cnid_private_t *priv = NULL;
    int ret;

    driver_create_file(&cnid_driver, &driver_attr_version);

    /* dev and priv zeroed in alloc_netdev. Thank God for small mercies... */
    netdev = alloc_netdev (sizeof(cnid_private_t), "cni%d", cnid_netdev_setup);
    if (netdev == NULL)
    {
        printk("cni0: Unable to alloc new net device\n");
        return -ENOMEM;
    }

    SET_MODULE_OWNER(netdev);
    SET_NETDEV_DEV(netdev, dev);
    platform_set_drvdata(to_platform_device(dev), netdev);

    priv = netdev_priv(netdev);

    ret = register_netdev (netdev);
    if(ret) {
        DPRINTK( "Unable to register device named %s (%p)...\n", netdev->name, netdev);
        return ret;
    }

    DPRINTK( "Registered device named %s (%p)...\n", netdev->name, netdev);

    /* get a PAL handle */
    priv->pal_hnd = PAL_cppi4Init(NULL, NULL);

    /* Allocate BD pool.. */
    if(!(priv->bdPoolBase =
                PAL_cppi4AllocDesc( priv->pal_hnd,
                                    PAL_CPPI41_QUEUE_MGR0,
                                    CNI_BD_NUM,
                                    CNI_BD_SIZE)) )
    {
        DPRINTK(": Unable to allocate Rx BD pool!\n");
    }
    else
    {
        DPRINTK(": Desc Pool has been allocated at 0x%08X \n", priv->bdPoolBase);
    }



#ifdef CONFIG_ARM_AVALANCHE_PPD
    /*
     * Config CNI PID/VPID
     */
    cni_pp_prepare_pid (netdev);
#endif

    spin_lock_init(&priv->devlock);

#ifdef CONFIG_INET_LRO
    /* LRO Setup */
    priv->lro_mgr.dev = netdev;
    memset(&priv->lro_mgr.stats, 0, sizeof(priv->lro_mgr.stats));
    priv->lro_mgr.features = LRO_F_NAPI;
    priv->lro_mgr.ip_summed = CHECKSUM_UNNECESSARY;
    priv->lro_mgr.ip_summed_aggr = CHECKSUM_UNNECESSARY; //CHECKSUM_NONE;
    priv->lro_mgr.max_desc = ARRAY_SIZE(priv->lro_arr);
    priv->lro_mgr.max_aggr = 32;
    priv->lro_mgr.frag_align_pad = 0;
    priv->lro_mgr.lro_arr = priv->lro_arr;
    priv->lro_mgr.get_skb_header = cnid_get_skb_hdr;
    memset(&priv->lro_arr, 0, sizeof(priv->lro_arr));

    /* Disable LRO by default */
    netdev->features &= ~NETIF_F_LRO;

    /* Init ethtool options */
    cnid_set_ethtool_ops( netdev );
#endif

    create_proc_read_entry("cnid", 0, NULL, cnid_read_proxy, netdev);

    {
        struct proc_dir_entry *res=create_proc_entry("cppi",0,NULL);
        if (res)
        {
            res->write_proc = cppi_write_cmds;
        }
    }
    return 0;
}


static int __devexit cnid_remove (struct device *dev)
{
    struct net_device *netdev = platform_get_drvdata (to_platform_device(dev));
    cnid_private_t* priv = netdev_priv(netdev);

    PAL_cppi4DeallocDesc( priv->pal_hnd, PAL_CPPI41_QUEUE_MGR0, priv->bdPoolBase );
    PAL_cppi4Exit(priv->pal_hnd, NULL);

    unregister_netdev (netdev);

    platform_device_unregister(to_platform_device(dev));

    return 0;
}


static int __init cnid_init_module (void)
{
    static struct platform_device *cnid_dev;

    DPRINTK (KERN_INFO DRV_NAME "\n");

    if(sizeof(cnid_desc_t) > CNI_BD_SIZE) {
       DPRINTK( "%s fundamentally broken. Contact maintainer!\n", DRV_NAME);
       return -1;
    }

    /* Initialize MAC address */
    if (inpmac && strlen(inpmac) == 17)
    {
        int i;
        int m[6];

        /* Translate MAC address from ASCII to binary */
        sscanf(inpmac, "%x:%x:%x:%x:%x:%x", &(m[0]), &(m[1]), &(m[2]), &(m[3]), &(m[4]), &(m[5]));
        for (i = 0; i < 6; i++)
        {
            defmac[i] = (Uint8)m[i];
        }
    }
    DPRINTK(KERN_INFO"cnid : Using MAC %x:%x:%x:%x:%x:%x", defmac[0], defmac[1], defmac[2], defmac[3], defmac[4], defmac[5]);

    /* TODO: this should be in a board file not here.
     * No fun registering driver and device in the same file. */
    cnid_dev = platform_device_register_simple("cni", -1, NULL, 0);

    cnid_driver.bus = platform_bus_type_ptr;
    if (driver_register(&cnid_driver)) {
        platform_device_unregister(cnid_dev);
        return -1;
    }

    return 0;
}


static void __exit cnid_cleanup_module (void)
{
    driver_remove_file(&cnid_driver, &driver_attr_version);
    driver_unregister(&cnid_driver);
}

#ifdef CONFIG_INET_LRO

/* Very minimal ethtool support for LRO */
static const struct ethtool_ops cnid_ethtool_ops = {
        .get_flags              = ethtool_op_get_flags,
        .set_flags              = ethtool_op_set_flags,
};

static void cnid_set_ethtool_ops(struct net_device *dev)
{
        SET_ETHTOOL_OPS(dev, (struct ethtool_ops *)&cnid_ethtool_ops);
}
#endif

module_param(inpmac, charp, 0);
module_init(cnid_init_module);
module_exit(cnid_cleanup_module);

