/*
 *
 * hal_interrupt_cause.c
 * Description:
 * Implementation functions that retrieve the interrupt cause for DOCSIS Interrupts
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#define _HAL_INTERRUPT_CAUSE_C_

/*! \file hal_interrupt_handlers_u.c
    \brief Implementation functions that retrieve the interrupt cause for DOCSIS
    \Interrupts. 
    \IMPORTANT:
    \   A. This code will run as part of a loadable kernel module, and should 
    \      be used in kernel-space only.
    \   B. These functions assume that they are protected externally, meaning 
    \      that their code is executed in "atomic" form.
*/

/**************************************************************************/
/*      INCLUDES:                                                         */
/**************************************************************************/
#include <linux/module.h>	/* module */
#include <linux/interrupt.h>
#include <linux/irq.h>
#include "hal_interrupt_cause.h"
#include "hal_mac_interrupt_regs.h"
#include "hal_phy_basic_regs.h"
#include "hal_phy_interrupt_regs.h"
#include <asm/arch/puma5/puma5_hardware.h>

/*#define DEBUG_CAUSE_PRINTS*/

/**************************************************************************/
/*      EXTERNS Declaration:                                              */
/**************************************************************************/
extern unsigned int pgaGrtIntNum;
extern void secondaryPumaWriteDocsisPhyReg(regOffset, value);
extern void secondaryPumaReadDocsisPhyReg(regOffset, value);
extern Uint32 secondaryPumaGetDocsisAddr (void);
extern Uint32 secondaryPumaIsInitialized(void);
extern Uint32 secondaryPumaIsProxyLoaded(void);
extern void babyPumaWriteDocsisPhyReg (Uint32 regOffset, Uint16 value);
extern void babyPumaReadDocsisPhyReg  (Uint32 regOffset, Uint16 *value);
extern int VLYNQ_clear_phy_interrupt( void );

/**************************************************************************/
/*      DEFINES:                                                          */
/**************************************************************************/

#define PUMA5_HW_REV_1_X   0
#define PUMA5_HW_REV_2_0   1

#define PHY_REMOTE_REG_BASE_OFFSET                 (0x400)

#define FEC_LOCK_VAL                                1
#define FEC_LOST_VAL                                0
#define NUM_OF_DS                                   8

#define PHY_RX1                                     0
#define PHY_RX2                                     1
#define PHY_RX3                                     2
#define PHY_RX4                                     3
#define PHY_RX5                                     4
#define PHY_RX6                                     5
#define PHY_RX7                                     6
#define PHY_RX8                                     7

typedef struct {
    Bool    isFecLockMasked; /* Indicates if FecLock interrupt is masked or enabled */
    Bool    isFecLostMasked; /* Indicates if FecLost interrupt is masked or enabled */
    Uint16  isrEnRegAddress; /* Interrupt enable register address (0xc2/0xc4/0x4c2/0x4c4) */
    Uint8   currState;       /* Last known FEC lost/lock status */
    Uint8   prevState;       /* Previous FEC lost/lock status */
    Uint8   fecLockBit;      /* FecLock bit in Interrupt enable register */
    Uint8   fecLostBit;      /* FecLost bit in Interrupt enable register */
    Uint8   fecLockCause;    /* Offset of FecLock interrupt cause in the interrupt group */
    Uint8   fecLostCause;    /* Offset of FecLost interrupt cause in the interrupt group */
}PhyFecLockLostDb_t;

static PhyFecLockLostDb_t gPhyFecLockLostDb[NUM_OF_DS];

#define READ_REMOTE_PHY_REG(regOffset, pvalue)    ((secondaryPumaIsInitialized()) ? (secondaryPumaReadDocsisPhyReg(regOffset, pValue)) : (babyPumaReadDocsisPhyReg(regOffset, pValue)) )
#define WRITE_REMOTE_PHY_REG(regOffset, value)    ((secondaryPumaIsInitialized()) ? (secondaryPumaWriteDocsisPhyReg(regOffset, value)) : (babyPumaWriteDocsisPhyReg(regOffset, value)) )
        
/**************************************************************************/
/*      LOCAL DECLARATIONS:                                               */
/**************************************************************************/
static Int32 HAL_IsrMaskPhyInterrupt(Uint32 cause, Bool maskOp, Bool isRemote);

static Int32 HAL_GetPhyReg16K(Uint32 phyRegAddr, Uint16 *pValue);
static Int32 HAL_SetPhyReg16K(Uint32 phyRegAddr, Uint16 value);


/**************************************************************************/
/*      LOCAL VARIABLES:                                                  */
/**************************************************************************/


/**************************************************************************/
/*      INTERFACE FUNCTIONS Implementation:                               */
/**************************************************************************/


/**********************************/               
/*  MAC Errors Interrupt Group    */               
/**********************************/               

/**************************************************************************/
/*! \fn Int32 HAL_IsrGetCauseMacErrorGroup(IntCauseErrorGroup_t *pCause)
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC-error group interrupt-line.
 *  \param[out] pCause pointer to the structure that will hold the interrupt cause(s).
 *  \return OK if successful, reject or fail otherwise                                     
 **************************************************************************/
Int32 HAL_IsrGetCauseMacErrorGroup(IntCauseErrorGroup_t *pCause)
{
    *(volatile Uint32*)(pCause) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_1_STATUS_REG);

    *(((volatile Uint32*)pCause) + 1) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_2_STATUS_REG);

    *(((volatile Uint32*)pCause) + 2) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_3_STATUS_REG);

    *(((volatile Uint32*)pCause) + 3) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_4_STATUS_REG);

    return 0;
}


/****************************************/               
/*    MAC UCD-Change Interrupt group    */               
/****************************************/               

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCauseUcdChange()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC UCD Change interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCauseUcdChange()
{
    return *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                  MAC_UCD_CHANGE_INTERRUPT_STATUS_REG);
}

/***********************************/               
/*    MAC US FW Interrupt group    */               
/***********************************/               

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCauseUsFw()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC US FW interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCauseUsFw()
{
    return *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                  MAC_US_FW_INTERRUPT_STATUS_REG);
}

/***********************************/               
/*    MAC DS FW Interrupt group    */               
/***********************************/               

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCauseDsFw()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC DS FW interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCauseDsFw()
{
    return *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                  MAC_DS_FW_INTERRUPT_STATUS_REG);
}



/***********************************/
/*         PHY  Interrupts         */
/***********************************/

/**************************************************************************/
/*! \fn Int32 HAL_IsrResetPhyFecDB()
 **************************************************************************
 *  \brief Reset Phy-Rx database.
 *  \param[in] none.
 *  \return OK if successful, reject or fail otherwise.
 **************************************************************************/
Int32 HAL_IsrResetPhyFecDB()
{
    Uint32 i;

    for (i = 0; i < NUM_OF_DS; i++)
    {
        gPhyFecLockLostDb[i].isFecLockMasked = True; /* Indicates if FecLock interrupt is masked or enabled */
        gPhyFecLockLostDb[i].isFecLostMasked = True; /*Indicates if FecLost interrupt is masked or enabled */
        gPhyFecLockLostDb[i].currState = 0;          /* Last known FEC lost/lock status */
        gPhyFecLockLostDb[i].prevState = 0;          /* Previous FEC lost/lock status */
    }

    /* Interrupt enable register address (0xc2/0xc4/0x4c2/0x4c4) */
    gPhyFecLockLostDb[0].isrEnRegAddress = PHY_RX1_RX2_INTERRUPT_ENABLE;
    gPhyFecLockLostDb[1].isrEnRegAddress = PHY_RX1_RX2_INTERRUPT_ENABLE;
    gPhyFecLockLostDb[2].isrEnRegAddress = PHY_RX3_RX4_INTERRUPT_ENABLE;
    gPhyFecLockLostDb[3].isrEnRegAddress = PHY_RX3_RX4_INTERRUPT_ENABLE;
    gPhyFecLockLostDb[4].isrEnRegAddress = PHY_RX5_RX6_INTERRUPT_ENABLE;
    gPhyFecLockLostDb[5].isrEnRegAddress = PHY_RX5_RX6_INTERRUPT_ENABLE;
    gPhyFecLockLostDb[6].isrEnRegAddress = PHY_RX7_RX8_INTERRUPT_ENABLE;
    gPhyFecLockLostDb[7].isrEnRegAddress = PHY_RX7_RX8_INTERRUPT_ENABLE;

    /* Offset of FecLock interrupt cause in the interrupt group */
    gPhyFecLockLostDb[0].fecLockCause = PHY_RX1_INT_FEC_LOCK;
    gPhyFecLockLostDb[1].fecLockCause = PHY_RX2_INT_FEC_LOCK;
    gPhyFecLockLostDb[2].fecLockCause = PHY_RX3_INT_FEC_LOCK;
    gPhyFecLockLostDb[3].fecLockCause = PHY_RX4_INT_FEC_LOCK;
    gPhyFecLockLostDb[4].fecLockCause = PHY_RX5_INT_FEC_LOCK;
    gPhyFecLockLostDb[5].fecLockCause = PHY_RX6_INT_FEC_LOCK;
    gPhyFecLockLostDb[6].fecLockCause = PHY_RX7_INT_FEC_LOCK;
    gPhyFecLockLostDb[7].fecLockCause = PHY_RX8_INT_FEC_LOCK;

    /* Offset of FecLost interrupt cause in the interrupt group */
    gPhyFecLockLostDb[0].fecLostCause = PHY_RX1_INT_FEC_LOST;
    gPhyFecLockLostDb[1].fecLostCause = PHY_RX2_INT_FEC_LOST;
    gPhyFecLockLostDb[2].fecLostCause = PHY_RX3_INT_FEC_LOST;
    gPhyFecLockLostDb[3].fecLostCause = PHY_RX4_INT_FEC_LOST;
    gPhyFecLockLostDb[4].fecLostCause = PHY_RX5_INT_FEC_LOST;
    gPhyFecLockLostDb[5].fecLostCause = PHY_RX6_INT_FEC_LOST;
    gPhyFecLockLostDb[6].fecLostCause = PHY_RX7_INT_FEC_LOST;
    gPhyFecLockLostDb[7].fecLostCause = PHY_RX8_INT_FEC_LOST;

    if (system_rev == PUMA5_HW_REV_2_0)
    {
        /* FecLock bit in Interrupt enable register */
        gPhyFecLockLostDb[0].fecLockBit = PHY_RX1_FEC_LOCKED_HW_REV_2_0;
        gPhyFecLockLostDb[1].fecLockBit = PHY_RX2_FEC_LOCKED_HW_REV_2_0;
        gPhyFecLockLostDb[2].fecLockBit = PHY_RX3_FEC_LOCKED_HW_REV_2_0;
        gPhyFecLockLostDb[3].fecLockBit = PHY_RX4_FEC_LOCKED_HW_REV_2_0;
        gPhyFecLockLostDb[4].fecLockBit = PHY_RX5_FEC_LOCKED_HW_REV_2_0;
        gPhyFecLockLostDb[5].fecLockBit = PHY_RX6_FEC_LOCKED_HW_REV_2_0;
        gPhyFecLockLostDb[6].fecLockBit = PHY_RX7_FEC_LOCKED_HW_REV_2_0;
        gPhyFecLockLostDb[7].fecLockBit = PHY_RX8_FEC_LOCKED_HW_REV_2_0;

        /* FecLost bit in Interrupt enable register */
        gPhyFecLockLostDb[0].fecLostBit = PHY_RX1_FEC_LOST_HW_REV_2_0;
        gPhyFecLockLostDb[1].fecLostBit = PHY_RX2_FEC_LOST_HW_REV_2_0;
        gPhyFecLockLostDb[2].fecLostBit = PHY_RX3_FEC_LOST_HW_REV_2_0;
        gPhyFecLockLostDb[3].fecLostBit = PHY_RX4_FEC_LOST_HW_REV_2_0;
        gPhyFecLockLostDb[4].fecLostBit = PHY_RX5_FEC_LOST_HW_REV_2_0;
        gPhyFecLockLostDb[5].fecLostBit = PHY_RX6_FEC_LOST_HW_REV_2_0;
        gPhyFecLockLostDb[6].fecLostBit = PHY_RX7_FEC_LOST_HW_REV_2_0;
        gPhyFecLockLostDb[7].fecLostBit = PHY_RX8_FEC_LOST_HW_REV_2_0;
    }
    else
    {
        /* FecLock bit in Interrupt enable register */
        gPhyFecLockLostDb[0].fecLockBit = PHY_RX1_FEC_LOCKED_HW_REV_1_X;
        gPhyFecLockLostDb[1].fecLockBit = PHY_RX2_FEC_LOCKED_HW_REV_1_X;
        gPhyFecLockLostDb[2].fecLockBit = PHY_RX3_FEC_LOCKED_HW_REV_1_X;
        gPhyFecLockLostDb[3].fecLockBit = PHY_RX4_FEC_LOCKED_HW_REV_1_X;
        gPhyFecLockLostDb[4].fecLockBit = PHY_RX5_FEC_LOCKED_HW_REV_1_X;
        gPhyFecLockLostDb[5].fecLockBit = PHY_RX6_FEC_LOCKED_HW_REV_1_X;
        gPhyFecLockLostDb[6].fecLockBit = PHY_RX7_FEC_LOCKED_HW_REV_1_X;
        gPhyFecLockLostDb[7].fecLockBit = PHY_RX8_FEC_LOCKED_HW_REV_1_X;

        /* FecLost bit in Interrupt enable register */
        gPhyFecLockLostDb[0].fecLostBit = PHY_RX1_FEC_LOST_HW_REV_1_X;
        gPhyFecLockLostDb[1].fecLostBit = PHY_RX2_FEC_LOST_HW_REV_1_X;
        gPhyFecLockLostDb[2].fecLostBit = PHY_RX3_FEC_LOST_HW_REV_1_X;
        gPhyFecLockLostDb[3].fecLostBit = PHY_RX4_FEC_LOST_HW_REV_1_X;
        gPhyFecLockLostDb[4].fecLostBit = PHY_RX5_FEC_LOST_HW_REV_1_X;
        gPhyFecLockLostDb[5].fecLostBit = PHY_RX6_FEC_LOST_HW_REV_1_X;
        gPhyFecLockLostDb[6].fecLostBit = PHY_RX7_FEC_LOST_HW_REV_1_X;
        gPhyFecLockLostDb[7].fecLostBit = PHY_RX8_FEC_LOST_HW_REV_1_X;
    }

    return 0;
}

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetFecStatusBitVal(Uint32 ch)
 **************************************************************************
 *  \brief retrieves current FEC status of the channel (FEC_LOCK_VAL/FEC_LOST_VAL)
 *  \param[in] Uint32 ch - channel to check for FEC status
 *  \return The ch FEC status.
 **************************************************************************/
Uint32 HAL_IsrGetFecStatusBitVal(Uint32 ch)
{
    Uint16 valFec;
    Uint16 annexVal;
    Uint32 bitVal;
    
    /*check FEC status register for FEC lost/lock status */
    HAL_GetPhyReg16K(PHY_RX_FEC_STATUS(ch + 1), &valFec);
    
    bitVal = (valFec >> PHY_RX_FEC_STATUS_TR_SYNC) & 1;
    
    /* Euro patch: FEC A Status Register is inverted to FEC B Status Register */
    HAL_GetPhyReg16K(PHY_RX_CH_REG_OFFSET(ch, PHY_QAM_MODE_0_REGISTER), &annexVal);
    annexVal = annexVal & (1 << PHY_QAM_ANNEX_TYPE);
    if(annexVal) /*0 - Annex B, 1 - Annex A*/
    {
        bitVal = 1 - bitVal;
    }

    return bitVal;
}

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCausePhy(Bool isFromRemote)
 **************************************************************************
 *  \brief retrieves the specific causes that asserted an interrupt on the
 *  \PHY interrupt-line.
 *  \param[in] Bool isFromRemote - interrupt received from master puma phy (0) or remote baby puma (1)
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCausePhy(Bool isFromRemote)
{
    Uint32 pCause;
    Uint32 bitVal;
    Uint32 i;
    Uint32 startCh, endCh;
    Uint32 remoteOffset;
    
#ifdef DEBUG_CAUSE_PRINTS
    printk( "--> %s: %s\n", __FUNCTION__, isFromRemote ? "From Remote" : "From Master");
#endif
    
    if (isFromRemote)
    {
        startCh = NUM_OF_DS / 2;
        endCh = NUM_OF_DS;
        remoteOffset = PHY_REMOTE_REG_BASE_OFFSET;
    }
    else
    {
        startCh = 0;
        endCh = NUM_OF_DS / 2;
        remoteOffset = 0;
    }

    /* Clear all interrupts */
    if (system_rev == PUMA5_HW_REV_2_0)
    {
        Uint16 clearRx1Rx2, clearRx3Rx4;

        HAL_GetPhyReg16K(remoteOffset + PHY_RX1_RX2_INTERRUPT_STATUS, &clearRx1Rx2);
        HAL_GetPhyReg16K(remoteOffset + PHY_RX3_RX4_INTERRUPT_STATUS, &clearRx3Rx4);
        HAL_SetPhyReg16K(remoteOffset + PHY_RX1_RX2_INTERRUPT_STATUS, clearRx1Rx2);
        HAL_SetPhyReg16K(remoteOffset + PHY_RX3_RX4_INTERRUPT_STATUS, clearRx3Rx4);

        if (isFromRemote)
        {
            VLYNQ_clear_phy_interrupt();
        }
    }

    pCause = 0;

    for(i = startCh; i < endCh; i++)
    {
        bitVal = HAL_IsrGetFecStatusBitVal(i);

#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: port=%d: bitVal = %d, currState  = %d \n",__FUNCTION__, i+1, bitVal, gPhyFecLockLostDb[i].currState);
#endif

        if (bitVal != gPhyFecLockLostDb[i].currState)
        {
            /* updating database */
            gPhyFecLockLostDb[i].prevState = gPhyFecLockLostDb[i].currState;
            gPhyFecLockLostDb[i].currState = bitVal;

            if(FEC_LOST_VAL == bitVal)
            {
#ifdef DEBUG_CAUSE_PRINTS
                printk( "-->%s: bitVal = FEC_LOST_VAL\n", __FUNCTION__);
#endif

                if(!(gPhyFecLockLostDb[i].isFecLostMasked))
                {
                    pCause |= (1 << gPhyFecLockLostDb[i].fecLostCause);
                }
            }

            else
            {
#ifdef DEBUG_CAUSE_PRINTS
                printk( "-->%s: bitVal = FEC_LOCK_VAL\n",__FUNCTION__);
#endif

                if(!(gPhyFecLockLostDb[i].isFecLockMasked))
                {
                    pCause |= (1 << gPhyFecLockLostDb[i].fecLockCause);
                }
            }
        }
    }

    

#ifdef DEBUG_CAUSE_PRINTS
    printk( "-->%s: pCause = 0x%x \n\n",__FUNCTION__, pCause);
#endif

    /*returning bitmask*/
    return pCause;

}

/***********************************/               
/*    Remote MAC interrupt group   */               
/***********************************/               

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCauseRemoteMac()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \Remote MAC interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCauseRemoteMac()
{
    unsigned int remoteDocsisBase, regVal;

    if (!secondaryPumaIsInitialized())
        return 0;

    if (!secondaryPumaIsProxyLoaded())
    {
        remoteDocsisBase = secondaryPumaGetDocsisAddr();
        regVal = *(volatile Uint32*)(remoteDocsisBase + MAC_DS_FW_INTERRUPT_STATUS_REG);
    }
    else
    {
        secondaryPumaReadDocsisMacReg( MAC_DS_FW_INTERRUPT_STATUS_REG, &regVal );
    }

    VLYNQ_clear_mac_interrupt();

    return regVal;
}



/**************************************************************************/
/*! \fn Int32 HAL_IsrMaskInterrupt(Uint32 cause, Bool maskOp)
 **************************************************************************
 *  \brief Masks or unmasks a speific interrupt.
 *  \param[in] cause defines the specific interrupt that needs to be 
 *  \                masked/unmasked.
 *  \param[in] maskOp defines whether to mask or unmask. 
 *  \                 True = mask, False= unmask. 
 *  \return OK if successful, reject or fail otherwise                     
 **************************************************************************/
Int32 HAL_IsrMaskInterrupt(Uint32 cause, Bool maskOp)
{
    Uint32 address = MAC_KERNEL_BASE_ADDRESS;
    Uint32 group = HAL_GET_INTERRUPT_GROUP(cause);
    Uint8  bit = HAL_GET_INTERRUPT_BIT(cause);

    switch (group) {
	case MAC_ERR1_INT_GROUP:
		if (bit > MAC_ERR1_INT_FLTR_UCD_US4)
			return -1;
		address += MAC_ERROR_INTERRUPT_1_MASK_REG;
		break;

	case MAC_ERR2_INT_GROUP:
		address += MAC_ERROR_INTERRUPT_2_MASK_REG;
		break;

	case MAC_ERR3_INT_GROUP:
		if (bit > MAC_ERR3_INT_MP_ORD_FIFO_FULL)
			return -1;
		address += MAC_ERROR_INTERRUPT_3_MASK_REG;
		break;

	case MAC_ERR4_INT_GROUP:
		address += MAC_ERROR_INTERRUPT_4_MASK_REG;
		break;

	case MAC_UCD_INT_GROUP:
		if (bit > MAC_UCD_INT_NEW_UCD_US4)
			return -1;
		address += MAC_UCD_CHANGE_INTERRUPT_MASK_REG;
		break;

	case MAC_US_FW_INT_GROUP:
		address += MAC_US_FW_INTERRUPT_MASK_REG;
		break;

	case MAC_DS_FW_INT_GROUP:
		address += MAC_DS_FW_INTERRUPT_MASK_REG;
		break;

	case PHY_INT_GROUP:
        return HAL_IsrMaskPhyInterrupt(cause, maskOp, False);
        break;

    case REMOTE_PHY_INT_GROUP:
        return HAL_IsrMaskPhyInterrupt(cause, maskOp, True);
        break;
			

    case PHY_PGA_GRT_INT_GROUP:
        if(maskOp == True)
        {
            printk(KERN_INFO "HAL_IsrMaskInterrupt: calling disable_irq(PHY_PGA_GRT_INT_NUM)\n");
            disable_irq(PHY_PGA_GRT_INT_NUM);
        }
        else
        {
            printk(KERN_INFO "HAL_IsrMaskInterrupt: calling enable_irq(PHY_PGA_GRT_INT_NUM)\n");
            enable_irq(PHY_PGA_GRT_INT_NUM);
        }
        return 0;
        break;


    case PHY_MPT_INT_GROUP:
        
        if (system_rev == PUMA5_HW_REV_2_0)
        {
            if(maskOp == True)
            {
                printk(KERN_INFO "HAL_IsrMaskInterrupt: calling disable_irq(PHY_MPT_INT_NUM)\n");
                disable_irq(PHY_MPT_INT_NUM);
            }
            else
            {
                printk(KERN_INFO "HAL_IsrMaskInterrupt: calling enable_irq(PHY_MPT_INT_NUM)\n");
                enable_irq(PHY_MPT_INT_NUM);
            }
        }
        return 0;

    case REMOTE_MAC_INT_GROUP:
        {
            Uint32 maskRegVal;

            if (!secondaryPumaIsInitialized())
                return 0;

            if (!secondaryPumaIsProxyLoaded())
            {
                address = secondaryPumaGetDocsisAddr() + MAC_DS_FW_INTERRUPT_MASK_REG;
                break;
            }
            else
            {
                
                secondaryPumaReadDocsisMacReg( MAC_DS_FW_INTERRUPT_MASK_REG, &maskRegVal );
                if (maskOp)
                {
                    maskRegVal |= (1<<bit);        
                }
                else
                {
                    maskRegVal &= ~(1<<bit);
                }
                secondaryPumaWriteDocsisMacReg( MAC_DS_FW_INTERRUPT_MASK_REG, maskRegVal );
                
                return 0;
            }
        }


	default:
		return -1;
	}

    if(maskOp)
        /* mask the interrupt*/
        *((volatile Uint32 *)address) |= (1<<bit);
    else
        /* unmask the interrupt*/
        *((volatile Uint32 *)address) &= ~(1<<bit);
    return 0;
}


/**************************************************************************/
/*      LOCAL FUNCTIONS:                                                  */
/**************************************************************************/
/**************************************************************************/
/*! \fn static Int32 HAL_IsrMaskPhyInterrupt(Uint32 cause, Bool maskOp, Bool isRemote)
 **************************************************************************
 *  \brief Perform mask or unmask operation.
 *  \param[in] Uint32 cause - interrupts bitmask.
 *  \param[in] Bool maskOp - Mask operation (maskOp = 1) or Unmask operation
 *  \param[in] Bool isRemote - PHY interrupt was from remote (Baby Puma) or Master
 *   (maskOp = 0).
 *  \return OK if successful, reject or fail otherwise.
 **************************************************************************/
static Int32 HAL_IsrMaskPhyInterrupt(Uint32 cause, Bool maskOp, Bool isRemote)
{
    Uint32 address;
    Uint8  intrBit;
    Uint16 regVal;
    Uint32 bitVal;
    Int32  i;
    Uint32 arrIndex;
    Bool   isFecLock;

#ifdef DEBUG_CAUSE_PRINTS
    printk( "-->%s:\n", __FUNCTION__);
#endif

    intrBit   = HAL_GET_INTERRUPT_BIT(cause);
    arrIndex  = HAL_PHY_GET_FEC_STATUS_DB_INDX(intrBit);
    if (isRemote)
    {
        arrIndex += (NUM_OF_DS / 2);
    }
    
    address = gPhyFecLockLostDb[arrIndex].isrEnRegAddress;
    isFecLock = HAL_PHY_IS_FEC_LOCK(intrBit);
    if (isFecLock)
    {
        intrBit = gPhyFecLockLostDb[arrIndex].fecLockBit;
        gPhyFecLockLostDb[arrIndex].isFecLockMasked = maskOp;
    }
    else
    {
        intrBit = gPhyFecLockLostDb[arrIndex].fecLostBit;
        gPhyFecLockLostDb[arrIndex].isFecLostMasked = maskOp;
    }

#ifdef DEBUG_CAUSE_PRINTS
    printk( "-->%s: RX=%d, lock=%d, intrBit=%d, maskOp=%d\n", __FUNCTION__, arrIndex+1, isFecLock, intrBit, maskOp);
#endif
    

    if (system_rev == PUMA5_HW_REV_2_0)
    {
        /* in PHY core rev2.0, the logic of Mask/Unmask is:          */
        /* intrBit=1 & bit12=1 ==> Interrupt Unmasked (enabled)      */
        /* intrBit=1 & bit12=0 ==> Interrupt Masked (disabled)       */
        /* intrBit=0 ==> No change                                   */
        regVal = 1 << intrBit;
        if (!maskOp)
        {
            regVal |= (1 << PHY_BITWISE_ENABLE_ISR_HW_REV_2_0);
        }
    }
	else
	{
        /* in PHY core rev1.x, the logic of Mask/Unmask is:     */
        /* intrBit=1 ==> Interrupt Unmasked (enabled)           */
        /* intrBit=0 ==> Interrupt Masked (disabled)            */
		HAL_GetPhyReg16K(address, &regVal);
		if (maskOp)
        {
            /* mask (disable) the interrupt */
			regVal &= (~(1 << intrBit));
        }
		else
        {
            /* unmask (enable) the interrupt*/
			regVal |= (1 << intrBit);
        }
	}
    HAL_SetPhyReg16K(address, regVal);

#ifdef DEBUG_CAUSE_PRINTS
    printk( "-->%s: address = 0x%x, regVal = 0x%x\n", __FUNCTION__, address, regVal);
    printk(KERN_INFO "-->%s: updating database...\n", __FUNCTION__);
#endif

    /* Check the relevant registers for FEC loss/lock */
    bitVal = HAL_IsrGetFecStatusBitVal(arrIndex);
    
    if (bitVal != gPhyFecLockLostDb[arrIndex].currState)
    {
        /*updating database*/
        gPhyFecLockLostDb[arrIndex].prevState = gPhyFecLockLostDb[arrIndex].currState;
        gPhyFecLockLostDb[arrIndex].currState = bitVal;
    }

    return 0;
}


static Int32 HAL_GetPhyReg16K(Uint32 phyRegAddr, Uint16 *pValue)
{
    Uint32 regOffset;

    if (phyRegAddr >= PHY_REMOTE_REG_BASE_OFFSET)
    {
        phyRegAddr = PHY_REMOTE_CH_REG_OFFSET(phyRegAddr);
        regOffset = phyRegAddr*2;
        if(regOffset >= HAL_PHY_ADDRESS_SPACE_SIZE)
            return -1;
        regOffset += HAL_PHY_BASE_OFFSET;

        READ_REMOTE_PHY_REG(regOffset, pValue);
    }
    else
    {
        regOffset = phyRegAddr*2;/*2-bytes for each reg*/
        if(regOffset >= HAL_PHY_ADDRESS_SPACE_SIZE)
            return -1;
        
        *pValue = *(volatile Uint16*)(PHY_KERNEL_BASE_ADDRESS + regOffset);
    }

#ifdef DEBUG_CAUSE_PRINTS
    printk(KERN_INFO "%s: phyRegAddr = 0x%x, regOffset = %d, value = (0x%x)\n", __FUNCTION__, phyRegAddr, regOffset, *pValue);
#endif

    return 0;
}

static Int32 HAL_SetPhyReg16K(Uint32 phyRegAddr, Uint16 value)
{
    Uint32 regOffset;

    if (phyRegAddr >= PHY_REMOTE_REG_BASE_OFFSET)
    {
        phyRegAddr = PHY_REMOTE_CH_REG_OFFSET(phyRegAddr);
        regOffset = phyRegAddr*2;
        if(regOffset >= HAL_PHY_ADDRESS_SPACE_SIZE)
            return -1;
        regOffset += HAL_PHY_BASE_OFFSET;

        WRITE_REMOTE_PHY_REG(regOffset, value);
    }

    else
    {
        regOffset = phyRegAddr*2;/*2-bytes for each reg*/
        if(regOffset >= HAL_PHY_ADDRESS_SPACE_SIZE)
            return -1;
        
        *(volatile Uint16*)(PHY_KERNEL_BASE_ADDRESS + regOffset) = value;
    }

#ifdef DEBUG_CAUSE_PRINTS
    printk(KERN_INFO "%s: phyRegAddr =  0x%x, regOffset = %d, value = (0x%x)\n", __FUNCTION__, phyRegAddr, regOffset, value);
#endif

    return 0;
}

