/*
 *
 * baby_puma_driver.c
 * Description:
 * Baby-Puma Driver implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

/*! \file baby_puma_driver.c
    \brief Implementation of Baby-Puma Driver. 
    \code.
*/

#include <linux/module.h>	/* module */
#include <asm/uaccess.h>
#include "pal.h"
#include "soc_interface_driver.h"
#include "baby_puma_driver.h"
#include "hal_vlynq.h"
#include "soc_docsis_global_driver.h"
#include "mpeg_out_driver.h"

/*****************************************/
/* Definitions for Baby-Puma Driver      */
/*****************************************/

/**********************/
/* VLYNQ definitions  */
/**********************/
#define BABY_PUMA_VLYNQ_NUM_WIN           (4)
#define BABY_PUMA_VLYNQ_LOCAL_WIN_ADDR    (0x0C000000)
/* win index 0 */
#define BABY_PUMA_VLYNQ_SOC_INX           (0)
#define BABY_PUMA_VLYNQ_SOC_ADDR          (IO_VIRT2PHY(PUMA5_SOC_BASE_ADDRESS))
#define BABY_PUMA_VLYNQ_SOC_SIZE          (0x1000000)     /*16777216 bytes*/
/* win index 1 */
#define BABY_PUMA_VLYNQ_MPEG_OUT_INX      (1)
#define BABY_PUMA_VLYNQ_MPEG_OUT_ADDR     (IO_VIRT2PHY(PUMA5_MPEG_OUT_BASE_ADDR))
#define BABY_PUMA_VLYNQ_MPEG_OUT_SIZE     (0x1000000)   /*16777216 bytes*/
/* win index 2 */
#define BABY_PUMA_VLYNQ_DOCSIS_INX        (2)
#define BABY_PUMA_VLYNQ_DOCSIS_ADDR       (0x09000000)
#define BABY_PUMA_VLYNQ_DOCSIS_SIZE       (0x1000000)   /*16777216 bytes*/

/**********************/
/* PLL definitions    */
/**********************/
#define PUMA5_PLL_CONTROL_OFFSET          (0x20100)
#define PUMA5_PLL_OSCILLATOR_DIV_OFFSET   (0x20124)
#define PUMA5_PLL_WAKEUP_OFFSET           (0x20130)
#define PUMA5_PLL_CONTROL_CMD_OFFSET      (0x20138)
#define PUMA5_GOSET_1_VAL                 (0x1)

/**********************/
/* I2C definitions    */
/**********************/
#define PUMA5_I2C_PSC_OFFSET              (0x21a30)
#define PUMA5_PSC_I2C_MODULE_EN_VAL       (0x103)

/**********************/
/* LED definitions    */
/**********************/
#define REMOTE_LED_REG_OFSET              (0x1090C)


/********************/
/* PROXY definitions*/
/********************/
#define PUMA5_PROXY_PSC_OFFSET                (0x21A8C)
#define PUMA5_PROXY_PSC_VALUE                 (0x103)

#define PUMA5_PROXY_BASE_OFFSET               (0x18000)
                                                 
#define PUMA5_PROXY_CONFIG_OFFSET             (PUMA5_PROXY_BASE_OFFSET + 0x0)
#define PUMA5_PROXY_CONFIG_DIR_SHIFT          (0x0)
#define PUMA5_PROXY_CONFIG_DIR_RD             (0x1)
#define PUMA5_PROXY_CONFIG_DIR_WR             (0x0)
#define PUMA5_PROXY_CONFIG_BYTE_COUNT_SHIFT   (0x1)

#define PUMA5_PROXY_ADDR_OFFSET               (PUMA5_PROXY_BASE_OFFSET + 0x4)
#define PUMA5_PROXY_WR_DATA_OFFSET            (PUMA5_PROXY_BASE_OFFSET + 0x10)
#define PUMA5_PROXY_RD_DATA_OFFSET            (PUMA5_PROXY_BASE_OFFSET + 0x20)
#define PUMA5_PROXY_MODE_OFFSET               (PUMA5_PROXY_BASE_OFFSET + 0x34)
#define PUMA5_PROXY_MODE_STALL                (0x1)
#define PUMA5_PROXY_REQ_OFFSET                (PUMA5_PROXY_BASE_OFFSET + 0x38)

/*****************************************/
/*      LOCAL DECLARATIONS:              */
/*****************************************/
static void babyPumaVlynqGlobalConfig   (void);
static int  babyPumaVlynqWindowsConfig  (void);
static void babyPumaDocsisConfig        (Uint32 docsisConfParam);
static void babyPumaMpegOutConfig       (void);
static void babyPumaI2CConfig           (void);
static void babyPumaProxyInit           (void);
void babyPumaReadDocsisPhyReg    (Uint32 regOffset, Uint16 *value);
void babyPumaWriteDocsisPhyReg   (Uint32 regOffset, Uint16 value);
static void babyPumaDriverInit          (Uint32 docsisConfParam);
static int  babyPumaDriverIoctl         (struct inode *inode,struct file *file,
                                         unsigned int ioctl_num,unsigned long ioctl_param, struct semaphore *soc_sem);
/* Vlynq Win struct - Holds all information regarding the Vlynq Windows */
typedef struct BabyPumaVlynqWin {
   Uint32   localAddr;
   Uint32   remoteAddr;
   Uint32   winSize;
}BabyPumaVlynqWin_t;

/*****************************************/
/*      LOCAL VARIABLES:                 */
/*****************************************/
static BabyPumaVlynqWin_t  baby_puma_vlynq_win[BABY_PUMA_VLYNQ_NUM_WIN];

static SoCDriverOperations_t baby_puma_operations =  
{
   .socModuleID   = SOC_BABY_PUMA_MODULE_ID,
   .soc_ioctl     = &babyPumaDriverIoctl, 
   .soc_release   = NULL,
   .soc_open      = NULL,
   .soc_cleanup   = NULL,
   .soc_init      = NULL,
};

static Uint32     babyPuma_debug_enable = 0;
static Uint32     babyPuma_driver_init  = 0;

#define  IS_BABY_DEBUG_ON     ((babyPuma_debug_enable)? True:False)

/*****************************************/
/* INTERFACE FUNCTIONS Implementation:   */
/*****************************************/

/**************************************************************************/
/*! \fn void babyPumaRegisterSoCModule (SoCDriverOperations_t **mpeg_out_operations)
 **************************************************************************
 *  \brief Register the Baby-Puma driver function pointers object.
 *  \param[in] **baby_puma_operations_out - pointer to Baby-Puma operations.
 *  \return none.
 **************************************************************************/
void babyPumaRegisterSoCModule (SoCDriverOperations_t **baby_puma_operations_out)
{
   *baby_puma_operations_out  =  &baby_puma_operations;
}

/**************************************************************************/
/*! \fn int babyPumaGetSoCAddr (Uint32  *soc_addr)
 **************************************************************************
 *  \brief Get the start address of the local VLYNQ window that represents 
 *  \brief the remote (Baby-Puma) SoC Addr.
 *  \param[out] *soc_addr - Local VLYNQ window that represents the remote SoC Addr.
 *  \return 0 on success of -1 on fail.
 **************************************************************************/
Uint32 babyPumaGetSoCAddr (void)
{
   return (baby_puma_vlynq_win[BABY_PUMA_VLYNQ_SOC_INX].localAddr);
}

/**************************************************************************/
/*! \fn int babyPumaGetMpegOutAddr (Uint32  *mpeg_out_addr)
 **************************************************************************
 *  \brief Get the start address of the local VLYNQ window that represents 
 *  \brief the remote (Baby-Puma) MPEG-Out Addr.
 *  \param[out] *soc_addr - Local VLYNQ window that represents the remote MPEG-Out Addr.
 *  \return 0 on success of -1 on fail.
 **************************************************************************/
Uint32 babyPumaGetMpegOutAddr (void)
{
  return (baby_puma_vlynq_win[BABY_PUMA_VLYNQ_MPEG_OUT_INX].localAddr);
}

/**************************************************************************/
/*! \fn int babyPumaGetDocsisAddr (Uint32  *docsis_addr)
 **************************************************************************
 *  \brief Get the start address of the local VLYNQ window that represents 
 *  \brief the remote (Baby-Puma) DOCSIS Addr.
 *  \param[out] *soc_addr - Local VLYNQ window that represents the remote DOCSIS Addr.
 *  \return 0 on success of -1 on fail.
 **************************************************************************/
Uint32 babyPumaGetDocsisAddr (void)
{
   return (baby_puma_vlynq_win[BABY_PUMA_VLYNQ_DOCSIS_INX].localAddr);
}


/**************************************************************************/
/*! \fn int babyPumaIsInitialized ()
 **************************************************************************
 *  \brief Gets the babyPuma_driver_init variable. 
 **************************************************************************/
Uint32 babyPumaIsInitialized()
{
    return babyPuma_driver_init;
}



/**************************************************************************/
/*      STATIC  Functions                                                 */
/**************************************************************************/

/**************************************************************************/
/*! \fn static int   babyPumaDriverIoctl(struct inode *inode, struct file *file,
                                        unsigned int ioctl_num, unsigned long ioctl_param)
 **************************************************************************
 *  \brief This function is called whenever a process tries to do an ioctl to our
 * SoC driver file. We get two extra parameters (additional to the inode and file
 * structures, which all device functions get): the number of the ioctl called
 * and the parameter given to the ioctl function.
 *
 * If the ioctl is write or read/write (meaning output is returned to the
 * calling process), the ioctl call returns the output of this function.
 * _IOC macros are used to decode ioctl numbers:
 *    _IOC_DIR(nr)  - get direction number 
 *    _IOC_TYPE(nr) - get type number 
 *    _IOC_NR(nr)   - get ordinal (sequential) number 
 *    _IOC_SIZE(nr) - get size number 
 *  \return 0 on success otherwise ioctl error code.
 **************************************************************************/
static int   babyPumaDriverIoctl(struct inode *inode,
                                struct file *file,
                                unsigned int ioctl_num,
                                unsigned long ioctl_param,
                                struct semaphore *soc_sem)
{
   RemoteRegInfo_t reg_info;
   Uint32 docsisConfParam;
   Uint32 remote_addr;

   /*
    * extract the type and number bitfields, and don't decode
    * wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok(  )
    */
   if (_IOC_TYPE(ioctl_num) != SOC_BABY_PUMA_MODULE_ID) 
      return -ENOTTY;
   if (_IOC_NR(ioctl_num) > BABY_PUMA_IOCTL_MAXNR) 
      return -ENOTTY;

   down(soc_sem);
   /* 
	 * Switch according to the ioctl called 
	 */
	switch (ioctl_num) {
   case BABY_PUMA_WRITE_PHY_REG:
      /* get the value from the user and write it to remote baby puma reg */
      if ( copy_from_user((void *)(&reg_info),
                          (void *)ioctl_param,
                          sizeof(RemoteRegInfo_t)) )
         {
            printk(KERN_ERR "babyPumaDriverIoctl:: copy_from_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
         }
      babyPumaWriteDocsisPhyReg(reg_info.reg_addr,reg_info.reg_value);
      break;
   case BABY_PUMA_READ_PHY_REG:
      /* read from remote baby puma reg and return the value to the user */
      if ( copy_from_user((void *)(&reg_info),
                          (void *)ioctl_param,
                          sizeof(RemoteRegInfo_t)) )
         {
            printk(KERN_ERR "babyPumaDriverIoctl:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
         }
      babyPumaReadDocsisPhyReg(reg_info.reg_addr,&(reg_info.reg_value));
      if ( copy_to_user((void *)ioctl_param,(void *)(&reg_info),sizeof(RemoteRegInfo_t)) )
         {
            printk(KERN_ERR "babyPumaDriverIoctl:: copy_to_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
         }
      break;
   case BABY_PUMA_DRIVER_INIT:
      if ( copy_from_user((void *)(&docsisConfParam),
                          (void *)ioctl_param,
                          sizeof(Uint32)) )
         {
            printk(KERN_ERR "babyPumaDriverIoctl:: Invalid argument for docsisConfParam.\n");
            up(soc_sem);
            return -EINVAL;
         }
      babyPumaDriverInit(docsisConfParam);
      break;
   case BABY_PUMA_GET_SOC_ADDR:
      if (VLYNQ_get_link_status() == 0)
         return -EPERM;
      
      remote_addr = babyPumaGetSoCAddr();

      if ( copy_to_user((void *)(ioctl_param),
                        (void *)&remote_addr,
                        sizeof(Uint32)) )
         {
            printk(KERN_ERR "babyPumaDriverIoctl:: Invalid argument for remote_addr.\n");
            up(soc_sem);
            return -EINVAL;
         }
      break;
   default:  /* redundant, as cmd was checked against MAXNR */
      up(soc_sem);
      return -ENOTTY;
	}

   up(soc_sem);
	return 0;
}


/**************************************************************************/
/*! \fn static void babyPumaDriverInit(Uint32 docsisConfParam)
 **************************************************************************
 *  \brief Perform the required initialization of the Baby-Puma chip.
 *  \      This function is envoked at the installation of the driver.
 *  \param[in]  *soc_sem - pointer to the driver's semaphore.
 *  \return none.
 **************************************************************************/
static void babyPumaDriverInit(Uint32 docsisConfParam)
{
    if (babyPuma_driver_init)
    {
       printk(KERN_INFO "babyPumaDriverInit you try to reinit me !!! no can do\n");
       return;
    }

    babyPuma_driver_init = 1;
    
    printk(KERN_INFO "babyPumaDriverInit\n");

    /* Reset the Baby-Puma (To be sure the Baby-Puma Endians boot pin is correct) */
    /* This is a MUST operation after we configure the Mpeg-In unit */
    VLYNQ_reset_slave();

    /* Configure the VLYNQ interface between the Puma-5 and the Baby-Puma */
    babyPumaVlynqGlobalConfig();
    babyPumaVlynqWindowsConfig();
   
    /* Configure the Baby-Puma's Docsis SoC */
    babyPumaDocsisConfig(docsisConfParam);

    /* Configure the Baby-Puma's MPEG-Out unit (Opens all 4 ports of the Baby) */
    babyPumaMpegOutConfig();

    /* Configure I2C over VLYNQ module */
    babyPumaI2CConfig();

    /* Configure Baby-Puma's Proxy module (access to the remote DOCSIS PHY) */
    babyPumaProxyInit();

    /* [AmihayT] Foo Foo Baby's leds. if baby's leds are ON this init was OK */
    socWriteRegister(babyPumaGetSoCAddr() + REMOTE_LED_REG_OFSET,0xFFFFFFFF,True);
    socWriteRegister(babyPumaGetSoCAddr() + REMOTE_LED_REG_OFSET,0x0,True);

    return;
}

/*****************************************/
/* LOCAL FUNCTIONS Implementation:       */
/*****************************************/

/**************************************************************************/
/*! \fn static void babyPumaVlynqGlobalConfig   (void)
 **************************************************************************
 *  \brief Perform the required Global Initialization of the VLYNQ interface.
 *  \return none.
 **************************************************************************/
static void babyPumaVlynqGlobalConfig   ()
{
   printk(KERN_INFO "babyPumaVlynqGlobalConfig\n");

   VLYNQ_setup_clock_direction(VLYNQ_SOURCE_CLOCK, VLYNQ_LOCAL);
   /*VLYNQ_setup_clock_direction(VLYNQ_SINK_CLOCK, VLYNQ_REMOTE);*/
   VLYNQ_setup_clock_divider(7, VLYNQ_LOCAL);
   VLYNQ_setup_tx_region(BABY_PUMA_VLYNQ_LOCAL_WIN_ADDR ,VLYNQ_LOCAL);
   VLYNQ_setup_interrupt_configuration( True, True, VLYNQ_LOCAL );
   VLYNQ_setup_interrupt_configuration( False, False, VLYNQ_REMOTE );
   VLYNQ_enable_remote_phy_interrupt();

   {
      /* TBD : do we need to restart link ? VLYNQ_reset_link( 0 or 1 ); */
      VLYNQ_reset_link(1);
      msleep(100);
      VLYNQ_reset_link(0);
      msleep(100);
   }

   printk(KERN_INFO "babyPumaVlynqGlobalConfig:: VLYNQ Link is %s \n",(VLYNQ_get_link_status()==1)?"UP":"DOWN");
}

/**************************************************************************/
/*! \fn static int babyPumaVlynqWindowsConfig  (void)
 **************************************************************************
 *  \brief Perform the required VLYNQ Windows Initialization.
 *  \return none.
 **************************************************************************/
static int babyPumaVlynqWindowsConfig  (void)
{
   Uint32 index;

   printk(KERN_INFO "babyPumaVlynqWindowsConfig\n");

   /* Init Global VLYNQ Windows data */
   memset(baby_puma_vlynq_win,0,sizeof(baby_puma_vlynq_win));
   baby_puma_vlynq_win[BABY_PUMA_VLYNQ_SOC_INX].remoteAddr        = BABY_PUMA_VLYNQ_SOC_ADDR;
   baby_puma_vlynq_win[BABY_PUMA_VLYNQ_SOC_INX].winSize           = BABY_PUMA_VLYNQ_SOC_SIZE;

   baby_puma_vlynq_win[BABY_PUMA_VLYNQ_MPEG_OUT_INX].remoteAddr   = BABY_PUMA_VLYNQ_MPEG_OUT_ADDR;
   baby_puma_vlynq_win[BABY_PUMA_VLYNQ_MPEG_OUT_INX].winSize      = BABY_PUMA_VLYNQ_MPEG_OUT_SIZE;

   baby_puma_vlynq_win[BABY_PUMA_VLYNQ_DOCSIS_INX].remoteAddr     = BABY_PUMA_VLYNQ_DOCSIS_ADDR;
   baby_puma_vlynq_win[BABY_PUMA_VLYNQ_DOCSIS_INX].winSize        = BABY_PUMA_VLYNQ_DOCSIS_SIZE;


   for (index=0; index<BABY_PUMA_VLYNQ_NUM_WIN;   index++)
   {
      if (baby_puma_vlynq_win[index].winSize)
      {
         /* Allocates local addr (VLYNQ local win start addr)  */
         if (index == 0)
         {
            baby_puma_vlynq_win[index].localAddr = IO_PHY2VIRT(BABY_PUMA_VLYNQ_LOCAL_WIN_ADDR); 
         }
         else
         {
            baby_puma_vlynq_win[index].localAddr = (baby_puma_vlynq_win[index-1].localAddr + baby_puma_vlynq_win[index-1].winSize);
         }
         /* Configure the VLYNQ windows */
         if (VLYNQ_setup_rx_region(baby_puma_vlynq_win[index].remoteAddr,
                                   baby_puma_vlynq_win[index].winSize,
                                   index,
                                   VLYNQ_REMOTE))
         {
            printk(KERN_INFO "babyPumaVlynqWindowsConfig:: VLYNQ_setup_rx_region failed !!!\n");
            return -1;
         }
      }
   } /* for */
 
   return 0;
}

/**************************************************************************/
/*! \fn static void babyPumaDocsisConfig      (Uint32 docsisConfParam)
 **************************************************************************
 *  \brief Perform the required initialization of the Baby-Puma Docis section.
 *  \return none.
 **************************************************************************/
static void babyPumaDocsisConfig        (Uint32 docsisConfParam)
{
   Uint32 remote_soc_addr;
   printk(KERN_INFO "babyPumaDocsisConfig\n");

   remote_soc_addr = babyPumaGetSoCAddr();

   /* Configure Baby-Puma PLL Control */
   printk(KERN_INFO "babyPumaDocsisConfig:: Configure Baby-Puma PLL Control\n");
   socWriteRegister((remote_soc_addr + PUMA5_PLL_CONTROL_OFFSET),0x0,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_OSCILLATOR_DIV_OFFSET),0x1,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_WAKEUP_OFFSET),0xFFFF,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_CONTROL_CMD_OFFSET),PUMA5_GOSET_1_VAL,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_CONTROL_OFFSET),0x9,True);

   printk(KERN_INFO "babyPumaDocsisConfig:: initialize Baby-Puma Docsis Global\n");
   socDocsisGlobalInit(SOC_DEVICE_REMOTE);

   printk(KERN_INFO "babyPumaDocsisConfig:: Configure Baby-Puma Docsis Global\n");
   socDocsisGlobalConfig(docsisConfParam,SOC_DEVICE_REMOTE);
}  

/**************************************************************************/
/*! \fn static void babyPumaMpegOutConfig       (void)
 **************************************************************************
 *  \brief Perform the required initialization of the Baby-Puma MPEG-Out section.
 *  \return none.
 **************************************************************************/
static void babyPumaMpegOutConfig       (void)
{
   SoCMPEGOutGlobalConfig_t mpegOutConfiguration;

   /* Configure the remote MPEG-Out unit */
   memset(&mpegOutConfiguration,0,sizeof(SoCMPEGOutGlobalConfig_t));
   mpegOutConfiguration.clockInv        = 1;
   mpegOutConfiguration.msbFirstTxOrder = 1;

   printk(KERN_INFO "babyPumaMpegOutConfig\n");
   /* TBD: trigger the MPEG-Out init + open all 4 ports */
   mpegOutUnitEnable( SOC_DEVICE_REMOTE );
   mpegOutUnitGlobalConfig( &mpegOutConfiguration, SOC_DEVICE_REMOTE);
   mpegOutPortSet( 1, True, SOC_DEVICE_REMOTE );
   mpegOutPortSet( 2, True, SOC_DEVICE_REMOTE );
   mpegOutPortSet( 3, True, SOC_DEVICE_REMOTE );
   mpegOutPortSet( 4, True, SOC_DEVICE_REMOTE );
}

/**************************************************************************/
/*! \fn static void babyPumaMpegOutConfig       (void)
 **************************************************************************
 *  \brief Perform the required initialization of the Baby-Puma I2C section.
 *  \return none.
 **************************************************************************/
static void babyPumaI2CConfig           (void)
{
   printk(KERN_INFO "babyPumaI2CConfig\n");

   /* Configure Baby-Puma I2C LPSC (out from reset) */
   socWriteRegister((babyPumaGetSoCAddr() + PUMA5_I2C_PSC_OFFSET),PUMA5_PSC_I2C_MODULE_EN_VAL,True);
   socWriteRegister((PUMA5_PSC_GO_REG( SOC_DEVICE_REMOTE )), PUMA5_PSC_GO_VAL  ,True); 
}

/**************************************************************************/
/*! \fn static void babyPumaProxyInit    (void)
 **************************************************************************
 *  \brief Initializes the proxy mechanism in the Baby-Puma, which allows 
 *  \access to the DOCSIS PHY.
 *  \return none.
 **************************************************************************/
static void babyPumaProxyInit()
{

    Uint32   remote_soc_addr;

    printk(KERN_INFO "babyPumaProxyInit\n");

    remote_soc_addr = babyPumaGetSoCAddr();

    /* PROXY PSC */
    socWriteRegister((remote_soc_addr + PUMA5_PROXY_PSC_OFFSET),PUMA5_PROXY_PSC_VALUE,True);
    socWriteRegister((PUMA5_PSC_GO_REG( SOC_DEVICE_REMOTE )), PUMA5_PSC_GO_VAL  ,True);

    /* Set PROXY mode to "STALL" */
    socWriteRegister((remote_soc_addr + PUMA5_PROXY_PSC_OFFSET),PUMA5_PROXY_MODE_STALL,True);

}

void babyPumaReadDocsisPhyReg (Uint32 regOffset, Uint16 *value)
{
   Uint32 remote_soc_addr;
   Uint32 regVal=0;

   remote_soc_addr  = babyPumaGetSoCAddr();

   /* Set proxy Config */
   regVal |= ( (PUMA5_PROXY_CONFIG_DIR_RD         << PUMA5_PROXY_CONFIG_DIR_SHIFT) |
               (2 /* 2-bytes for DOCSIS PHY access */ << PUMA5_PROXY_CONFIG_BYTE_COUNT_SHIFT));
   
   socWriteRegister((remote_soc_addr + PUMA5_PROXY_CONFIG_OFFSET),regVal,IS_BABY_DEBUG_ON);

   /* Set proxy address */
   socWriteRegister((remote_soc_addr + PUMA5_PROXY_ADDR_OFFSET),BABY_PUMA_VLYNQ_DOCSIS_ADDR + regOffset,IS_BABY_DEBUG_ON);

   /* proxy read-request */
   socReadRegister((remote_soc_addr + PUMA5_PROXY_REQ_OFFSET),&regVal,IS_BABY_DEBUG_ON);

   /* get value from proxy */
   socReadRegister((remote_soc_addr + PUMA5_PROXY_RD_DATA_OFFSET),&regVal,IS_BABY_DEBUG_ON);

   if((regOffset%4) == 2)
   {
       *value = (Uint16)(regVal&0xFFFF);
   }
   else
   {
       *value = (Uint16)((regVal>>16)&0xFFFF);
   }
}

void babyPumaWriteDocsisPhyReg (Uint32 regOffset, Uint16 value)
{
    Uint32 remote_soc_addr;
    Uint32 regVal=0;

    remote_soc_addr  = babyPumaGetSoCAddr();

    /* Set proxy Config */
    regVal |= ( (PUMA5_PROXY_CONFIG_DIR_WR         << PUMA5_PROXY_CONFIG_DIR_SHIFT) |
                (2 /* 2-bytes for DOCSIS PHY access */ << PUMA5_PROXY_CONFIG_BYTE_COUNT_SHIFT));

    socWriteRegister((remote_soc_addr + PUMA5_PROXY_CONFIG_OFFSET),regVal,True);

    /* Set proxy address */
    socWriteRegister((remote_soc_addr + PUMA5_PROXY_ADDR_OFFSET),BABY_PUMA_VLYNQ_DOCSIS_ADDR + regOffset,IS_BABY_DEBUG_ON);


    /* set value from proxy */
    if((regOffset%4) == 2)
    {
        regVal = (Uint32)(value&0xFFFF);
    }
    else
    {
        regVal = (Uint32)(value);
        regVal = ((regVal<<16) & 0xFFFF0000);
    }

    socWriteRegister((remote_soc_addr + PUMA5_PROXY_WR_DATA_OFFSET),regVal,IS_BABY_DEBUG_ON);


    /* proxy write-request */
    socWriteRegister((remote_soc_addr + PUMA5_PROXY_REQ_OFFSET),0,IS_BABY_DEBUG_ON);
}

/* brief: Export the follwing functions to those who needs to access the Baby-Puma registers. */
EXPORT_SYMBOL( babyPumaGetSoCAddr );
EXPORT_SYMBOL( babyPumaGetMpegOutAddr );
EXPORT_SYMBOL( babyPumaGetDocsisAddr );
EXPORT_SYMBOL( babyPumaWriteDocsisPhyReg );
EXPORT_SYMBOL( babyPumaReadDocsisPhyReg );
EXPORT_SYMBOL( babyPumaIsInitialized );

