/*
 *
 * secondary_puma_driver.c
 * Description:
 * Secondary-Puma Driver implementation (for 16x4 applications)
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

/*! \file secondary_puma_driver.c
    \brief Implementation of Secondary-Puma Driver. 
    \code.
*/

#include <linux/module.h>	/* module */
#include <asm/uaccess.h>
#include "pal.h"
#include "soc_interface_driver.h"
#include "secondary_puma_driver.h"
#include "hal_vlynq.h"
#include "soc_docsis_global_driver.h"
#include "mpeg_out_driver.h"
#include "secondary_puma_cppi.h"


/**********************************************/
/* Definitions for Secondary-Puma Driver      */
/**********************************************/

#define     SECONDARY_PUMA_APP_MODE

/**********************/
/* VLYNQ definitions  */
/**********************/

#define VLYNQ_NUM_STATIC_REMOTE_RX_WIN         (2)
#define VLYNQ_NUM_REMOTE_RX_WIN                (4)
#define VLYNQ_NUM_STATIC_LOCAL_RX_WIN          (2)

#define SECONDARY_PUMA_VLYNQ_LOCAL_WIN_ADDR    (0x0C000000)
/* win index 0 - Static window */
#define SECONDARY_PUMA_VLYNQ_SOC_INX           (0)
#define SECONDARY_PUMA_VLYNQ_SOC_ADDR          (IO_VIRT2PHY(PUMA5_SOC_BASE_ADDRESS))
#define SECONDARY_PUMA_VLYNQ_SOC_SIZE          (0x1000000)     /*16 MB*/
/* win index 1 - Static window */
#define SECONDARY_PUMA_VLYNQ_DOCSIS_INX        (1)
#define SECONDARY_PUMA_VLYNQ_DOCSIS_ADDR       (0x09000000)
#define SECONDARY_PUMA_VLYNQ_DOCSIS_SIZE       (0x1000000)   /*16 MB*/

/* win index 2 - Dynamic window 1 */
#define SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX       (2)
#define SECONDARY_PUMA_VLYNQ_DYNAMIC1_SIZE      (0x1000000)   /*16 MB*/

/* win index 3 - Dynamic window 2 */
#define SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX       (3)
#define SECONDARY_PUMA_VLYNQ_DYNAMIC2_SIZE      (0x1000000)   /*16 MB*/

#define SECONDARY_PUMA_VLYNQ_CPPI_ADDR         (0x03000000)
#define SECONDARY_PUMA_VLYNQ_CPPI_SIZE         (0x1000000)   /*16 MB*/

#define SECONDARY_PUMA_VLYNQ_RAM_LO_ADDR       (0x80000000)
#define SECONDARY_PUMA_VLYNQ_RAM_HI_ADDR       (0x81000000)

#define SECONDARY_PUMA_PRIMARY_RAM_ADDR        (0x85000000)  /* This describes the first address in the PRIMARY RAM which is mapped to the secondary. */
#define SECONDARY_PUMA_VLYNQ_RAM_SIZE          (0x1000000)   /*16 MB*/
#define SECONDARY_PUMA_VLYNQ_REMOTE_RAM_SIZE   (0x3000000)   /*48 MB*/



#define SECONDARY_PUMA_VLYNQ_SRAM_ADDR         (0x00000000)

/* LOCAL RX Windows */
#define SECONDARY_PUMA_VLYNQ_CPPI_LOCAL_IDX     (0)
#define SECONDARY_PUMA_VLYNQ_RAM_LOCAL_IDX      (1)


/* ILANZ TBD - following defines are duplicated, tbd. */

/**********************/
/* PLL definitions    */
/**********************/
#define PUMA5_PLL_CONTROL_OFFSET          (0x20100)
#define PUMA5_PLL_OSCILLATOR_DIV_OFFSET   (0x20124)
#define PUMA5_PLL_WAKEUP_OFFSET           (0x20130)
#define PUMA5_PLL_CONTROL_CMD_OFFSET      (0x20138)
#define PUMA5_GOSET_1_VAL                 (0x1)

/**********************/
/* I2C definitions    */
/**********************/
#define PUMA5_I2C_PSC_OFFSET              (0x21a30)
#define PUMA5_PSC_I2C_MODULE_EN_VAL       (0x103)

/**********************/
/* LED definitions    */
/**********************/
#define REMOTE_LED_REG_OFSET              (0x1090C)


/********************/
/* PROXY definitions*/
/********************/
#define PUMA5_PROXY_PSC_OFFSET                (0x21A8C)
#define PUMA5_PROXY_PSC_VALUE                 (0x103)

#define PUMA5_PROXY_BASE_OFFSET               (0x18000)
                                                 
#define PUMA5_PROXY_CONFIG_OFFSET             (PUMA5_PROXY_BASE_OFFSET + 0x0)
#define PUMA5_PROXY_CONFIG_DIR_SHIFT          (0x0)
#define PUMA5_PROXY_CONFIG_DIR_RD             (0x1)
#define PUMA5_PROXY_CONFIG_DIR_WR             (0x0)
#define PUMA5_PROXY_CONFIG_BYTE_COUNT_SHIFT   (0x1)

#define PUMA5_PROXY_ADDR_OFFSET               (PUMA5_PROXY_BASE_OFFSET + 0x4)
#define PUMA5_PROXY_WR_DATA_OFFSET            (PUMA5_PROXY_BASE_OFFSET + 0x10)
#define PUMA5_PROXY_RD_DATA_OFFSET            (PUMA5_PROXY_BASE_OFFSET + 0x20)
#define PUMA5_PROXY_MODE_OFFSET               (PUMA5_PROXY_BASE_OFFSET + 0x34)
#define PUMA5_PROXY_MODE_STALL                (0x1)
#define PUMA5_PROXY_REQ_OFFSET                (PUMA5_PROXY_BASE_OFFSET + 0x38)


/********************/
/* GPIO definitions */
/********************/

#define PUMA5_BUFFER_CONTROL_GPIO           (13)
#define PUMA5_RF_SPLITTER_AUX_GPIO_REV1     (27+AVALANCHE_MAX_PRIMARY_GPIOS)
#define PUMA5_RF_SPLITTER_AUX_GPIO_REV2     (29+AVALANCHE_MAX_PRIMARY_GPIOS)

/* END TBD*/


/*****************************************/
/*      LOCAL DECLARATIONS:              */
/*****************************************/

static void secondaryPumaOutOfResetSequence         (void);
static int  secondaryPumaVlynqGlobalConfig          (void);
static int  secondaryPumaVlynqLinkReset             (void);
static int  secondaryPumaVlynqConfigPhase1          (void);
static int  secondaryPumaVlynqConfigPhase2          (void);
static int  secondaryPumaVlynqDynamic1WindowConfig  (Uint32 remoteAddr);
static int  secondaryPumaVlynqDynamic2WindowConfig  (Uint32 remoteAddr);
static void secondaryPumaDocsisConfig               (Uint32 docsisConfParam);
static void secondaryPumaI2CConfig           (void);
static void secondaryPumaProxyInit           (void);
       void secondaryPumaReadDocsisPhyReg    (Uint32 regOffset, Uint16 *value, Bool secondaryPumaProxyLoaded);
       void secondaryPumaWriteDocsisPhyReg   (Uint32 regOffset, Uint16 value, Bool secondaryPumaProxyLoaded);
       void secondaryPumaReadDocsisMacReg    (Uint32 regOffset, Uint32 *value);
       void secondaryPumaWriteDocsisMacReg   (Uint32 regOffset, Uint32 value);
       void secondaryPumaWriteDocsisMacBuff  (Uint32 regOffset, Uint32 len, Uint8 *buff);
       void secondaryPumaReadDocsisMacBuff   (Uint32 regOffset, Uint32 len, Uint8 *buff);
       void secondaryPumaGetReadyToTransfer  (unsigned int *lockKey);
       void secondaryPumaFinishTransfer      (unsigned int lockKey);
static int  secondaryPumaDriverInit          (Uint32 docsisConfParam);
static int  secondaryPumaDriverIoctl         (struct inode *inode,struct file *file,
                                         unsigned int ioctl_num,unsigned long ioctl_param, struct semaphore *soc_sem);
static void secondaryMpegInEnable(void);
static void secondaryRFSplitterConfig(unsigned int splitterEn);
static void socDocsisBufferControlConfig(void);

/* Vlynq Win struct - Holds all information regarding the Vlynq Windows */
typedef struct SecondaryPumaVlynqWin {
   Uint32   localAddr;
   Uint32   remoteAddr;
   Uint32   winSize;
}SecondaryPumaVlynqWin_t;

/*****************************************/
/*      LOCAL VARIABLES:                 */
/*****************************************/
static SecondaryPumaVlynqWin_t  secondary_puma_remote_rx_vlynq_win[VLYNQ_NUM_REMOTE_RX_WIN];
static SecondaryPumaVlynqWin_t  secondary_puma_local_rx_vlynq_win[VLYNQ_NUM_STATIC_LOCAL_RX_WIN];

static SoCDriverOperations_t secondary_puma_operations =  
{
   .socModuleID   = SOC_SECONDARY_PUMA_MODULE_ID,
   .soc_ioctl     = &secondaryPumaDriverIoctl, 
   .soc_release   = NULL,
   .soc_open      = NULL,
   .soc_cleanup   = NULL,
   .soc_init      = NULL,
};

static Uint32     secondaryPuma_debug_enable = 0;
static Uint32     secondaryPuma_driver_init  = 0;
static Uint32     secondaryPuma_proxy_init   = 0;

static RemoteMacCopyInfo_t copyInfo;

#define  IS_SECONDARY_DEBUG_ON     ((secondaryPuma_debug_enable)? True:False)

/*****************************************/
/* INTERFACE FUNCTIONS Implementation:   */
/*****************************************/

/**************************************************************************/
/*! \fn void secondaryPumaRegisterSoCModule (SoCDriverOperations_t **mpeg_out_operations)
 **************************************************************************
 *  \brief Register the Secondary-Puma driver function pointers object.
 *  \param[in] **secondary_puma_operations_out - pointer to Secondary-Puma operations.
 *  \return none.
 **************************************************************************/
void secondaryPumaRegisterSoCModule (SoCDriverOperations_t **secondary_puma_operations_out)
{
   *secondary_puma_operations_out  =  &secondary_puma_operations;
}

/**************************************************************************/
/*! \fn int secondaryPumaGetSoCAddr ()
 **************************************************************************
 *  \brief Get the start address of the local VLYNQ window that represents 
 *  \brief the remote (Secondary-Puma) SoC Addr.
 *  \param[out] *soc_addr - Local VLYNQ window that represents the remote SoC Addr.
 *  \return 0 on success of -1 on fail.
 **************************************************************************/
Uint32 secondaryPumaGetSoCAddr (void)
{
   return (secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_SOC_INX].localAddr);
}

/**************************************************************************/
/*! \fn int secondaryPumaGetDocsisAddr ()
 **************************************************************************
 *  \brief Get the start address of the local VLYNQ window that represents 
 *  \brief the remote (Secondary-Puma) DOCSIS Addr.
 *  \param[out] *soc_addr - Local VLYNQ window that represents the remote DOCSIS Addr.
 *  \return 0 on success of -1 on fail.
 **************************************************************************/
Uint32 secondaryPumaGetDocsisAddr (void)
{
   return (secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DOCSIS_INX].localAddr);
}

/**************************************************************************/
/*! \fn int secondaryPumaGetDynamic1Addr ()
 **************************************************************************
 *  \brief Get the start address of the local VLYNQ window that represents 
 *  \brief the first dynamic memory mapping on the secondary puma.
 *  \param[out] *soc_addr - Local VLYNQ window.
 *  \return 0 on success of -1 on fail.
 **************************************************************************/
Uint32 secondaryPumaGetDynamic1Addr (void)
{
  return (secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX].localAddr);
}

/**************************************************************************/
/*! \fn int secondaryPumaGetDynamic2Addr ()
 **************************************************************************
 *  \brief Get the start address of the local VLYNQ window that represents 
 *  \brief the second dynamic memory mapping on the secondary puma.
 *  \param[out] *soc_addr - Local VLYNQ window.
 *  \return 0 on success of -1 on fail.
 **************************************************************************/
Uint32 secondaryPumaGetDynamic2Addr (void)
{
    return (secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX].localAddr);
}

/**************************************************************************/
/*! \fn int secondaryPumaIsInitialized ()
 **************************************************************************
 *  \brief Gets the secondaryPuma_driver_init variable. 
 **************************************************************************/
Uint32 secondaryPumaIsInitialized()
{
    return secondaryPuma_driver_init;
}


/**************************************************************************/
/*! \fn int secondaryPumaIsProxyLoaded ()
 **************************************************************************
 *  \brief Gets the secondaryPuma_proxy_init variable. 
 **************************************************************************/
Uint32 secondaryPumaIsProxyLoaded()
{
    return secondaryPuma_proxy_init;
}


/**************************************************************************/
/*      STATIC  Functions                                                 */
/**************************************************************************/

/**************************************************************************/
/*! \fn static int   secondaryPumaDriverIoctl(struct inode *inode, struct file *file,
                                        unsigned int ioctl_num, unsigned long ioctl_param)
 **************************************************************************
 *  \brief This function is called whenever a process tries to do an ioctl to our
 * SoC driver file. We get two extra parameters (additional to the inode and file
 * structures, which all device functions get): the number of the ioctl called
 * and the parameter given to the ioctl function.
 *
 * If the ioctl is write or read/write (meaning output is returned to the
 * calling process), the ioctl call returns the output of this function.
 * _IOC macros are used to decode ioctl numbers:
 *    _IOC_DIR(nr)  - get direction number 
 *    _IOC_TYPE(nr) - get type number 
 *    _IOC_NR(nr)   - get ordinal (sequential) number 
 *    _IOC_SIZE(nr) - get size number 
 *  \return 0 on success otherwise ioctl error code.
 **************************************************************************/
static int   secondaryPumaDriverIoctl(struct inode *inode,
                                struct file *file,
                                unsigned int ioctl_num,
                                unsigned long ioctl_param,
                                struct semaphore *soc_sem)
{
   RemoteRegInfo_t reg_info;
   RemoteMacRegInfo_t macRegInfo;
   Uint32 docsisConfParam;
   Uint32 remote_addr;
   Uint32 retVal;

   /*
    * extract the type and number bitfields, and don't decode
    * wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok(  )
    */
   if (_IOC_TYPE(ioctl_num) != SOC_SECONDARY_PUMA_MODULE_ID) 
      return -ENOTTY;
   if (_IOC_NR(ioctl_num) > SECONDARY_PUMA_IOCTL_MAXNR) 
      return -ENOTTY;

   down(soc_sem);
   /* 
	 * Switch according to the ioctl called 
	 */
	switch (ioctl_num) {
   case SECONDARY_PUMA_WRITE_PHY_REG:
      /* get the value from the user and write it to remote secondary puma reg */
      if ( copy_from_user((void *)(&reg_info),
                          (void *)ioctl_param,
                          sizeof(RemoteRegInfo_t)) )
         {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_from_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
         }
      secondaryPumaWriteDocsisPhyReg(reg_info.reg_addr,reg_info.reg_value, TRUE);
      break;

    case SECONDARY_PUMA_READ_PHY_REG:
      /* read from remote secondary puma reg and return the value to the user */
      if ( copy_from_user((void *)(&reg_info),
                          (void *)ioctl_param,
                          sizeof(RemoteRegInfo_t)) )
         {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
         }
      secondaryPumaReadDocsisPhyReg(reg_info.reg_addr,&(reg_info.reg_value), TRUE);
      if ( copy_to_user((void *)ioctl_param,(void *)(&reg_info),sizeof(RemoteRegInfo_t)) )
         {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_to_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
         }
      break;

    case SECONDARY_PUMA_WRITE_PHY_REG_IMMEDIATE:
        /* get the value from the user and write it to remote secondary puma reg */
        if ( copy_from_user((void *)(&reg_info),
            (void *)ioctl_param,
            sizeof(RemoteRegInfo_t)) )
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_from_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        secondaryPumaWriteDocsisPhyReg(reg_info.reg_addr,reg_info.reg_value, FALSE);
        break;
        
    case SECONDARY_PUMA_READ_PHY_REG_IMMEDIATE:
        /* read from remote secondary puma reg and return the value to the user */
        if ( copy_from_user((void *)(&reg_info),
            (void *)ioctl_param,
            sizeof(RemoteRegInfo_t)) )
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        secondaryPumaReadDocsisPhyReg(reg_info.reg_addr,&(reg_info.reg_value), FALSE);
        if ( copy_to_user((void *)ioctl_param,(void *)(&reg_info),sizeof(RemoteRegInfo_t)) )
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_to_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
      break;

    case SECONDARY_PUMA_WRITE_MAC_REG:
        /* get the value from the user and write it to remote secondary puma reg */
        if (copy_from_user((void *)(&macRegInfo), (void *)ioctl_param, sizeof(RemoteMacRegInfo_t)))
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_from_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        secondaryPumaWriteDocsisMacReg(macRegInfo.reg_addr,macRegInfo.reg_value);
        break;

    case SECONDARY_PUMA_READ_MAC_REG:
        /* get the value from the user and write it to remote secondary puma reg */
        if (copy_from_user((void *)(&macRegInfo), (void *)ioctl_param, sizeof(RemoteMacRegInfo_t)))
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_from_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        secondaryPumaReadDocsisMacReg(macRegInfo.reg_addr,&(macRegInfo.reg_value));
        if (copy_to_user((void *)ioctl_param,(void *)(&macRegInfo),sizeof(RemoteMacRegInfo_t)))
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_to_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        break;

    case SECONDARY_PUMA_WRITE_MAC_BUFF:
        /* get the value from the user and write it to remote secondary puma reg */
        if (copy_from_user((void *)(&copyInfo), (void *)ioctl_param, sizeof(RemoteMacCopyInfo_t)))
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_from_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        secondaryPumaWriteDocsisMacBuff(copyInfo.regOffset, copyInfo.copyLen, copyInfo.data);
        break;

    case SECONDARY_PUMA_READ_MAC_BUFF:
        /* get the value from the user and write it to remote secondary puma reg */
        if (copy_from_user((void *)(&copyInfo), (void *)ioctl_param, sizeof(RemoteMacCopyInfo_t)))
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_from_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        secondaryPumaReadDocsisMacBuff(copyInfo.regOffset, copyInfo.copyLen, copyInfo.data);
        if (copy_to_user((void *)ioctl_param, (void *)(&copyInfo), sizeof(RemoteMacCopyInfo_t)))
        {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: copy_to_user:: Invalid argument for remote reg info.\n");
            up(soc_sem);
            return -EINVAL;
        }
        break;

    case SECONDARY_PUMA_DRIVER_INIT:
      if ( copy_from_user((void *)(&docsisConfParam),
                          (void *)ioctl_param,
                          sizeof(Uint32)) )
         {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: Invalid argument for docsisConfParam.\n");
            up(soc_sem);
            return -EINVAL;
         }

      if (0 != secondaryPumaDriverInit(docsisConfParam))
      {
            up(soc_sem);
            return -EINVAL;
      }
      break;

    case SECONDARY_PUMA_GET_SOC_ADDR:
      if (VLYNQ_get_link_status() == 0)
      {
          up(soc_sem);
          return -EPERM;
      }
      
      remote_addr = secondaryPumaGetSoCAddr();

      if ( copy_to_user((void *)(ioctl_param),
                        (void *)&remote_addr,
                        sizeof(Uint32)) )
         {
            printk(KERN_ERR "secondaryPumaDriverIoctl:: Invalid argument for remote_addr.\n");
            up(soc_sem);
            return -EINVAL;
         }
      break;

    case SECONDARY_PUMA_EN_RF_SPLITTER:
        secondaryRFSplitterConfig(ioctl_param);
        break;

    case SECONDARY_PUMA_MPEG_IN_ENABLE:
        secondaryMpegInEnable();
        break;

    case SECONDARY_PUMA_BUFFER_CONTROL:
        socDocsisBufferControlConfig();
        break;

    case SECONDARY_PUMA_PROXY_LOADED:
        secondaryPuma_proxy_init = 1;
        break;

    default:  /* redundant, as cmd was checked against MAXNR */
      up(soc_sem);
      return -ENOTTY;
	}

   up(soc_sem);
	return 0;
}


/**************************************************************************/
/*! \fn static void secondaryPumaDriverInit(Uint32 docsisConfParam)
 **************************************************************************
 *  \brief Perform the required initialization of the Secondary-Puma chip.
 *  \      This function is envoked at the installation of the driver.
 *  \param[in]  *soc_sem - pointer to the driver's semaphore.
 *  \return none.
 **************************************************************************/
static int secondaryPumaDriverInit(Uint32 docsisConfParam)
{
    if (secondaryPuma_driver_init)
    {
       printk(KERN_INFO "secondaryPumaDriverInit you try to reinit me !!! no can do\n");
       return 0;
    }

    printk("secondaryPumaDriverInit\n");

    /* Perform the out-of-reset sequence for the Secondary-Puma */
    printk("calling secondaryPumaOutOfResetSequence\n");
    secondaryPumaOutOfResetSequence();

    secondaryPuma_driver_init = 1;
    
    /* Configure the Secondary-Puma's Docsis SoC */
    printk("calling secondaryPumaDocsisConfig\n");
    secondaryPumaDocsisConfig(docsisConfParam);

    /* Configure I2C over VLYNQ module */
    printk("calling secondaryPumaI2CConfig\n");
    secondaryPumaI2CConfig();

    /* Configure Secondary-Puma's Proxy module (access to the remote DOCSIS PHY) */
    printk("calling secondaryPumaProxyInit\n");
    secondaryPumaProxyInit();

    /* Set dynamic window to CPPI region and RAM - for secondary CPPI init */
    secondaryPumaVlynqDynamic1WindowConfig(SECONDARY_PUMA_VLYNQ_CPPI_ADDR);
    secondaryPumaVlynqDynamic2WindowConfig(SECONDARY_PUMA_VLYNQ_RAM_LO_ADDR);

    printk("=============== calling secondaryPumaCppiInit ==================\n");
    secondaryPumaCppiInit();
    
    return 0;
}

/*****************************************/
/* LOCAL FUNCTIONS Implementation:       */
/*****************************************/

/**************************************************************************/
/*! \fn static void secondaryPumaOutOfResetSequence   (void)
 **************************************************************************
 *  \brief Perform the very first operations that need to be done for taking
 *  \      secondary-puma out of reset.
 *  \      This sequence is important in order to enable the MPEG-In interface 
 *  \      as early as possible. 
 *  \return none.
 **************************************************************************/
static void secondaryPumaOutOfResetSequence()
{
    int rc;

    secondaryPumaVlynqGlobalConfig();

    /* The following actions need to be done in sequence, without interruption. */
    printk("DEBUG[ %s ] about to perform reset\n", __FUNCTION__);
    VLYNQ_reset_secondary();

    printk("DEBUG[ %s ] reset done\n", __FUNCTION__);
    
    do 
    {
        rc = VLYNQ_poll_for_link_up(5000);
        msleep(1000);
        printk(".");
    } 
    while (0 == rc);
    
    printk("\n");
    
    msleep(1000);

    printk("DEBUG[ %s ] Link is up \n", __FUNCTION__);

    secondaryPumaVlynqConfigPhase1();
    printk("DEBUG[ %s ] phase 1 done\n", __FUNCTION__);

    msleep(100);
    
    secondaryPumaVlynqConfigPhase2();
    printk("DEBUG[ %s ] phase 2 done\n", __FUNCTION__);

    secondaryPumaVlynqLinkReset();
    printk("DEBUG[ %s ] Link reset done\n", __FUNCTION__);

}


static int secondaryPumaVlynqGlobalConfig()
{
    /* perform initial VLYNQ configurations */
    VLYNQ_setup_clock_direction(VLYNQ_SOURCE_CLOCK, VLYNQ_LOCAL);
    VLYNQ_setup_clock_divider(1, VLYNQ_LOCAL);

    return 0;
}

static int secondaryPumaVlynqLinkReset()
{
    int rc;

    VLYNQ_reset_link(1);
    msleep(100);
    VLYNQ_reset_link(0);
    msleep(100);

    do 
    {
        rc = VLYNQ_poll_for_link_up(5000);
        msleep(1000);
        printk(".");
    } 
    while (0 == rc);
    
    printk("\n");
    
    msleep(1000);

    return 0;
}



/**************************************************************************/
/*! \fn static void secondaryPumaVlynqConfigPhase2   (void)
 **************************************************************************
 *  \brief First initialization phase of the VLYNQ interface.
 *  \return none.
 **************************************************************************/
static int secondaryPumaVlynqConfigPhase1()
{
    Uint32 index;

    VLYNQ_setup_tx_region(SECONDARY_PUMA_VLYNQ_LOCAL_WIN_ADDR ,VLYNQ_LOCAL);

    /* Init Global VLYNQ Windows data for Remote RX*/
    memset(secondary_puma_remote_rx_vlynq_win,0,sizeof(secondary_puma_remote_rx_vlynq_win));
    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_SOC_INX].remoteAddr        = SECONDARY_PUMA_VLYNQ_SOC_ADDR;
    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_SOC_INX].winSize           = SECONDARY_PUMA_VLYNQ_SOC_SIZE;

    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DOCSIS_INX].remoteAddr     = SECONDARY_PUMA_VLYNQ_DOCSIS_ADDR;
    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DOCSIS_INX].winSize        = SECONDARY_PUMA_VLYNQ_DOCSIS_SIZE;


    for (index=0; index<VLYNQ_NUM_STATIC_REMOTE_RX_WIN;   index++)
    {
       if (secondary_puma_remote_rx_vlynq_win[index].winSize)
       {
          /* Allocates local addr (VLYNQ local win start addr)  */
          if (index == 0)
          {
             secondary_puma_remote_rx_vlynq_win[index].localAddr = IO_PHY2VIRT(SECONDARY_PUMA_VLYNQ_LOCAL_WIN_ADDR); 
          }
          else
          {
             secondary_puma_remote_rx_vlynq_win[index].localAddr = (secondary_puma_remote_rx_vlynq_win[index-1].localAddr + secondary_puma_remote_rx_vlynq_win[index-1].winSize);
          }
          /* Configure the VLYNQ windows */
          if (VLYNQ_setup_rx_region(secondary_puma_remote_rx_vlynq_win[index].remoteAddr,
                                    secondary_puma_remote_rx_vlynq_win[index].winSize,
                                    index,
                                    VLYNQ_REMOTE))
          {
             printk(KERN_INFO "secondaryPumaVlynqConfigPhase1:: VLYNQ_setup_rx_region failed !!!\n");
             return -1;
          }
       }
    } /* for */
    return 0;
}
/**************************************************************************/
/*! \fn static void secondaryPumaVlynqConfigPhase2   (void)
 **************************************************************************
 *  \brief Second initialization phase of the VLYNQ interface.
 *  \return none.
 **************************************************************************/
static int secondaryPumaVlynqConfigPhase2()
{
   Uint32 index;

   printk("secondaryPumaVlynqConfigPhase2\n");

   VLYNQ_setup_tx_region(SECONDARY_PUMA_VLYNQ_LOCAL_WIN_ADDR ,VLYNQ_REMOTE);

   printk("secondaryPumaVlynqConfigPhase2 Interrupts local\n");

   VLYNQ_setup_interrupt_configuration( True, True, VLYNQ_LOCAL );

   printk("secondaryPumaVlynqConfigPhase2 Interrupts remote\n");

   VLYNQ_setup_interrupt_configuration( False, False, VLYNQ_REMOTE );

   VLYNQ_enable_remote_phy_interrupt();
   VLYNQ_enable_remote_mac_interrupt();

   printk("secondaryPumaVlynqConfigPhase2 before reset \n");

   /* Init Global VLYNQ Windows data for local RX*/
   memset(secondary_puma_local_rx_vlynq_win,0,sizeof(secondary_puma_local_rx_vlynq_win));
   secondary_puma_local_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_CPPI_LOCAL_IDX].remoteAddr     = SECONDARY_PUMA_VLYNQ_CPPI_ADDR;
   secondary_puma_local_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_CPPI_LOCAL_IDX].winSize        = SECONDARY_PUMA_VLYNQ_CPPI_SIZE;

   secondary_puma_local_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_RAM_LOCAL_IDX].remoteAddr      = SECONDARY_PUMA_PRIMARY_RAM_ADDR;
   secondary_puma_local_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_RAM_LOCAL_IDX].winSize         = SECONDARY_PUMA_VLYNQ_REMOTE_RAM_SIZE;


   for (index=0; index<VLYNQ_NUM_STATIC_LOCAL_RX_WIN;   index++)
   {
     /* Configure the VLYNQ windows */
     if (VLYNQ_setup_rx_region(secondary_puma_local_rx_vlynq_win[index].remoteAddr,
                               secondary_puma_local_rx_vlynq_win[index].winSize,
                               index,
                               VLYNQ_LOCAL))
     {
        printk("secondaryPumaVlynqConfigPhase2:: VLYNQ_setup_rx_region failed !!!\n");
        return -1;
     }
   } /* for */
   return 0;
}


/**************************************************************************/
/*! \fn static int secondaryPumaVlynqDynamic1WindowConfig  (Uint32 remoteAddr)
 **************************************************************************
 *  \brief Set up the first dynamic VLYNQ window.
 *  \param[in]  remoteAddr - address at the remote device to be mapped.
 *  \return none.
 **************************************************************************/
static int secondaryPumaVlynqDynamic1WindowConfig  (Uint32 remoteAddr)
{

    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX].remoteAddr     = remoteAddr;
    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX].winSize        = SECONDARY_PUMA_VLYNQ_DYNAMIC1_SIZE;

    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX].localAddr = 
        (secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX-1].localAddr + 
         secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX-1].winSize);

    if (VLYNQ_setup_rx_region(remoteAddr,
                              SECONDARY_PUMA_VLYNQ_DYNAMIC1_SIZE,
                              SECONDARY_PUMA_VLYNQ_DYNAMIC1_INX,
                              VLYNQ_REMOTE))
    {
       printk(KERN_INFO "secondaryPumaVlynqDynamic1WindowConfig:: VLYNQ_setup_rx_region failed !!!\n");
       return -1;
    }
    return 0;
}

/**************************************************************************/
/*! \fn static int secondaryPumaVlynqDynamic2WindowConfig  (Uint32 remoteAddr)
 **************************************************************************
 *  \brief Set up the first dynamic VLYNQ window.
 *  \param[in]  remoteAddr - address at the remote device to be mapped.
 *  \return none.
 **************************************************************************/
static int secondaryPumaVlynqDynamic2WindowConfig  (Uint32 remoteAddr)
{

    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX].remoteAddr     = remoteAddr;
    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX].winSize        = SECONDARY_PUMA_VLYNQ_DYNAMIC2_SIZE;

    secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX].localAddr = 
        (secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX-1].localAddr + 
         secondary_puma_remote_rx_vlynq_win[SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX-1].winSize);

    if (VLYNQ_setup_rx_region(remoteAddr,
                              SECONDARY_PUMA_VLYNQ_DYNAMIC2_SIZE,
                              SECONDARY_PUMA_VLYNQ_DYNAMIC2_INX,
                              VLYNQ_REMOTE))
    {
       printk(KERN_INFO "secondaryPumaVlynqDynamic2WindowConfig:: VLYNQ_setup_rx_region failed !!!\n");
       return -1;
    }
    return 0;
}


/**************************************************************************/
/*! \fn static void secondaryPumaDocsisConfig      (Uint32 docsisConfParam)
 **************************************************************************
 *  \brief Perform the required initialization of the Secondary-Puma Docis section.
 *  \return none.
 **************************************************************************/
static void secondaryPumaDocsisConfig        (Uint32 docsisConfParam)
{
   Uint32 remote_soc_addr;
   printk(KERN_INFO "secondaryPumaDocsisConfig\n");

   remote_soc_addr = secondaryPumaGetSoCAddr();

#ifndef SECONDARY_PUMA_APP_MODE
   /* Configure Secondary-Puma PLL Control */
   printk(KERN_INFO "secondaryPumaDocsisConfig:: Configure Secondary-Puma PLL Control\n");
   socWriteRegister((remote_soc_addr + PUMA5_PLL_CONTROL_OFFSET),0x0,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_OSCILLATOR_DIV_OFFSET),0x1,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_WAKEUP_OFFSET),0xFFFF,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_CONTROL_CMD_OFFSET),PUMA5_GOSET_1_VAL,True);
   socWriteRegister((remote_soc_addr + PUMA5_PLL_CONTROL_OFFSET),0x9,True);
#endif // SECONDARY_PUMA_APP_MODE

   printk(KERN_INFO "secondaryPumaDocsisConfig:: initialize Secondary-Puma Docsis Global\n");
   socDocsisGlobalInit(SOC_DEVICE_REMOTE);

   printk(KERN_INFO "secondaryPumaDocsisConfig:: Configure Secondary-Puma Docsis Global\n");
   socDocsisGlobalConfig(docsisConfParam,SOC_DEVICE_REMOTE);
}  


/**************************************************************************/
/*! \fn static void secondaryPumaI2CConfig       (void)
 **************************************************************************
 *  \brief Perform the required initialization of the Secondary-Puma I2C section.
 *  \return none.
 **************************************************************************/
static void secondaryPumaI2CConfig           (void)
{
   printk(KERN_INFO "secondaryPumaI2CConfig\n");

   /* Configure Secondary-Puma I2C LPSC (out from reset) */
   socWriteRegister((secondaryPumaGetSoCAddr() + PUMA5_I2C_PSC_OFFSET),PUMA5_PSC_I2C_MODULE_EN_VAL,True);
   socWriteRegister((PUMA5_PSC_GO_REG( SOC_DEVICE_REMOTE )), PUMA5_PSC_GO_VAL  ,True); 
}


/**************************************************************************/
/*! \fn static void secondaryPumaProxyInit    (void)
 **************************************************************************
 *  \brief Initializes the proxy mechanism in the Secondary-Puma, which allows 
 *  \access to the DOCSIS PHY.
 *  \return none.
 **************************************************************************/
static void secondaryPumaProxyInit()
{
    Uint32   remote_soc_addr;

    printk(KERN_INFO "secondaryPumaProxyInit\n");

    remote_soc_addr = secondaryPumaGetSoCAddr();

    /* PROXY PSC */
    socWriteRegister((remote_soc_addr + PUMA5_PROXY_PSC_OFFSET),PUMA5_PROXY_PSC_VALUE,True);
    socWriteRegister((PUMA5_PSC_GO_REG( SOC_DEVICE_REMOTE )), PUMA5_PSC_GO_VAL  ,True);
}


/**************************************************************************/
/*! \fn static void secondaryMpegInEnable(void)
 **************************************************************************
 *  \brief Perform the required initialization of the Puma's MPEG-In interface.
 *  \return none.
 **************************************************************************/
static void secondaryMpegInEnable(void)
{
   Uint32 regValue;

   socWriteRegister((PUMA5_BOOTCFG_KICK_0_REG( SOC_DEVICE_REMOTE )), PUMA5_BOOTCFG_KICK_0_VAL, False);
   socWriteRegister((PUMA5_BOOTCFG_KICK_1_REG( SOC_DEVICE_REMOTE )), PUMA5_BOOTCFG_KICK_1_VAL, False);

   /* Bit 1 "MPEG In EN". */
   socReadRegister((PUMA5_PINMUX0_REG( SOC_DEVICE_REMOTE )),&regValue,False);
   regValue |= PUMA5_PINMUX0_MPEG_IN_EN;
   socWriteRegister((PUMA5_PINMUX0_REG( SOC_DEVICE_REMOTE )), regValue, False);

   /* The default value of MAC-PHY Interface Global Control Register (0903:A818) is: */
   /* MAC-1 connected to PHY-1 ... MAC-8 connected to PHY-8                          */
   /* So in a wonderful life we have nothing to do...                                */
}


/**************************************************************************/
/*! \fn static void secondaryRFSplitterConfig(unsigned int splitterEn)
 **************************************************************************
 *  \brief Perform RF-Splitter configuration.
 *  \      This function is envoked by a specific IOCTL.
 *  \param[in]  splitterEn - 1 - for RF splitter ON  and 0 - for OFF.
 *  \return none.
 **************************************************************************/
static void secondaryRFSplitterConfig(unsigned int splitterEnParams)
{
    Uint32 regGpioAux;

    unsigned int gpio_pin = PUMA5_RF_SPLITTER_AUX_GPIO_REV1;
    
    if (splitterEnParams & SOC_DOCSIS_RF_SPLITTER_958REV2)
    {
        /* 958 rev02 */
        gpio_pin = PUMA5_RF_SPLITTER_AUX_GPIO_REV2;
    }
    
    gpio_pin -= AVALANCHE_MAX_PRIMARY_GPIOS;

    printk("Secondary Puma: Configure RF-Splitter (AUX-GPIO %d) %d \n",(gpio_pin), (int)splitterEnParams);
    
    /* For 958 and 958W boards */

    /* Enable GPIO */
    socReadRegister(    (PUMA5_DOCSIS_AUX_GPIO_REG( SOC_DEVICE_REMOTE )),&regGpioAux, True);
    regGpioAux |=    (1 << gpio_pin);
    socWriteRegister(   (PUMA5_DOCSIS_AUX_GPIO_REG( SOC_DEVICE_REMOTE )), regGpioAux, True);
    
    /* Set GPIO to Output */
    socReadRegister(    (PUMA5_DOCSIS_AUX_GPIO_DIR_REG( SOC_DEVICE_REMOTE )),&regGpioAux, True);
    regGpioAux &=   ~(1 << gpio_pin);
    socWriteRegister(   (PUMA5_DOCSIS_AUX_GPIO_DIR_REG( SOC_DEVICE_REMOTE )), regGpioAux, True);

    if (splitterEnParams & SOC_DOCSIS_RF_SPLITTER_EN)
    {
        /* Set value to 1 */
        socReadRegister(    (PUMA5_DOCSIS_AUX_GPIO_OUT_VAL_REG( SOC_DEVICE_REMOTE )),&regGpioAux, True);
        regGpioAux |= (1 << gpio_pin);
        socWriteRegister(   (PUMA5_DOCSIS_AUX_GPIO_OUT_VAL_REG( SOC_DEVICE_REMOTE )), regGpioAux, True);
    }
    else
    {
        /* Set value to 0 */
        socReadRegister(    (PUMA5_DOCSIS_AUX_GPIO_OUT_VAL_REG( SOC_DEVICE_REMOTE )),&regGpioAux, True);
        regGpioAux &= ~(1 << gpio_pin);
        socWriteRegister(   (PUMA5_DOCSIS_AUX_GPIO_OUT_VAL_REG( SOC_DEVICE_REMOTE )), regGpioAux, True);
    }
}


/**************************************************************************/
/*! \fn static void socDocsisBufferControlConfig(void)
 **************************************************************************
 *  \brief Perform Buffer Control configuration For 958 and 958W REV2 boards.
 *  \      This function is invoked by a specific IOCTL.
 *  \return none.
 **************************************************************************/
static void socDocsisBufferControlConfig(void)
{
    Uint32 regGpioAux;

    printk(KERN_INFO "Configure Buffer Control (GPIO 13)\n");

    /* Allocate GPIO#13 to buffer control */

    /* Enable GPIO */
    socReadRegister(    (PUMA5_DOCSIS_GPIO_REG( SOC_DEVICE_REMOTE )),&regGpioAux, True);
    regGpioAux |=    (1 << PUMA5_BUFFER_CONTROL_GPIO);
    socWriteRegister(   (PUMA5_DOCSIS_GPIO_REG( SOC_DEVICE_REMOTE )), regGpioAux, True);
    
    /* Set GPIO to Output */
    socReadRegister(    (PUMA5_DOCSIS_GPIO_DIR_REG( SOC_DEVICE_REMOTE )),&regGpioAux, True);
    regGpioAux &=   ~(1 << PUMA5_BUFFER_CONTROL_GPIO);
    socWriteRegister(   (PUMA5_DOCSIS_GPIO_DIR_REG( SOC_DEVICE_REMOTE )), regGpioAux, True);
    
    /* At initialization GPIO needs to be set to 0 */
    socReadRegister(    (PUMA5_DOCSIS_GPIO_OUT_VAL_REG( SOC_DEVICE_REMOTE )),&regGpioAux, True);
    regGpioAux &=   ~(1 << PUMA5_BUFFER_CONTROL_GPIO);
    socWriteRegister(   (PUMA5_DOCSIS_GPIO_OUT_VAL_REG( SOC_DEVICE_REMOTE )), regGpioAux, True);
}

void secondaryPumaReadDocsisPhyReg(Uint32 regOffset, Uint16 *value, Bool secondaryPumaProxyLoaded)
{
   Uint32 remote_soc_addr;
   Uint32 regVal=0;
   unsigned int  lockKey = 0;
   
   remote_soc_addr  = secondaryPumaGetSoCAddr();

   /* Set proxy Config */
   regVal |= ( (PUMA5_PROXY_CONFIG_DIR_RD         << PUMA5_PROXY_CONFIG_DIR_SHIFT) |
               (2 /* 2-bytes for DOCSIS PHY access */ << PUMA5_PROXY_CONFIG_BYTE_COUNT_SHIFT));

    if (secondaryPumaProxyLoaded)
    {
        secondaryPumaGetReadyToTransfer(&lockKey);
    }
      
   socWriteRegister((remote_soc_addr + PUMA5_PROXY_CONFIG_OFFSET),regVal,IS_SECONDARY_DEBUG_ON);

   /* Set proxy address */
   socWriteRegister((remote_soc_addr + PUMA5_PROXY_ADDR_OFFSET),SECONDARY_PUMA_VLYNQ_DOCSIS_ADDR + regOffset,IS_SECONDARY_DEBUG_ON);

   /* proxy read-request */
   socReadRegister((remote_soc_addr + PUMA5_PROXY_REQ_OFFSET),&regVal,IS_SECONDARY_DEBUG_ON);

   /* get value from proxy */
   socReadRegister((remote_soc_addr + PUMA5_PROXY_RD_DATA_OFFSET),&regVal,IS_SECONDARY_DEBUG_ON);

   if (secondaryPumaProxyLoaded)
   {
       secondaryPumaFinishTransfer(lockKey);
   }

   if((regOffset%4) == 2)
   {
       *value = (Uint16)(regVal&0xFFFF);
   }
   else
   {
       *value = (Uint16)((regVal>>16)&0xFFFF);
   }
}

void secondaryPumaWriteDocsisPhyReg(Uint32 regOffset, Uint16 value, Bool secondaryPumaProxyLoaded)
{
    Uint32 remote_soc_addr;
    Uint32 regVal=0;
    unsigned int  lockKey = 0;
    
    remote_soc_addr  = secondaryPumaGetSoCAddr();

    /* Set proxy Config */
    regVal |= ( (PUMA5_PROXY_CONFIG_DIR_WR         << PUMA5_PROXY_CONFIG_DIR_SHIFT) |
                (2 /* 2-bytes for DOCSIS PHY access */ << PUMA5_PROXY_CONFIG_BYTE_COUNT_SHIFT));

    if (secondaryPumaProxyLoaded)
    {
        secondaryPumaGetReadyToTransfer(&lockKey);
    }

    socWriteRegister((remote_soc_addr + PUMA5_PROXY_CONFIG_OFFSET),regVal,True);

    /* Set proxy address */
    socWriteRegister((remote_soc_addr + PUMA5_PROXY_ADDR_OFFSET),SECONDARY_PUMA_VLYNQ_DOCSIS_ADDR + regOffset,IS_SECONDARY_DEBUG_ON);


    /* set value from proxy */
    if((regOffset%4) == 2)
    {
        regVal = (Uint32)(value&0xFFFF);
    }
    else
    {
        regVal = (Uint32)(value);
        regVal = ((regVal<<16) & 0xFFFF0000);
    }

    socWriteRegister((remote_soc_addr + PUMA5_PROXY_WR_DATA_OFFSET),regVal,IS_SECONDARY_DEBUG_ON);


    /* proxy write-request */
    socWriteRegister((remote_soc_addr + PUMA5_PROXY_REQ_OFFSET),0,IS_SECONDARY_DEBUG_ON);

    if (secondaryPumaProxyLoaded)
    {
        secondaryPumaFinishTransfer(lockKey);
    }
}


void secondaryPumaWriteDocsisMacReg(Uint32 regOffset, Uint32 value)
{
    unsigned int  lockKey;
    
    secondaryPumaGetReadyToTransfer(&lockKey);
    
    *(volatile Uint32*)(secondaryPumaGetDocsisAddr() + regOffset) = value;

    secondaryPumaFinishTransfer(lockKey);
}




void secondaryPumaReadDocsisMacReg(Uint32 regOffset, Uint32 *value)
{
    unsigned int  lockKey;
    
    secondaryPumaGetReadyToTransfer(&lockKey);
    
    *value = *(volatile Uint32*)(secondaryPumaGetDocsisAddr() + regOffset);

    secondaryPumaFinishTransfer(lockKey);
}

void secondaryPumaWriteDocsisMacBuff(Uint32 regOffset, Uint32 len, Uint8 *buff)
{
    unsigned int  lockKey;
    
    secondaryPumaGetReadyToTransfer(&lockKey);
    
    memcpy((Uint8*)(secondaryPumaGetDocsisAddr() + regOffset), buff, len);
    
    secondaryPumaFinishTransfer(lockKey);
}


void secondaryPumaReadDocsisMacBuff(Uint32 regOffset, Uint32 len, Uint8 *buff)
{
    unsigned int  lockKey;
    
    secondaryPumaGetReadyToTransfer(&lockKey);

    memcpy(buff, (Uint8*)(secondaryPumaGetDocsisAddr() + regOffset), len);
    
    secondaryPumaFinishTransfer(lockKey);
}


void secondaryPumaGetReadyToTransfer(unsigned int *lockKey)
{
    Uint32 retVal = 0;
    
    PAL_osProtectEntry(PAL_OSPROTECT_INTERRUPT, lockKey);
    
    VLYNQ_secondary_puma_set_mem_access_start();
    
    while (retVal != 1)
    {
        retVal = VLYNQ_secondary_puma_can_mem_access_start();
    }
}

void secondaryPumaFinishTransfer(unsigned int lockKey)
{
    Uint32 retVal = 0;

    VLYNQ_secondary_puma_set_mem_access_finish();
    
    while (retVal != 1)
    {
        retVal = VLYNQ_secondary_puma_is_mem_access_finish();
    }
    
    PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT, lockKey);
}


/* brief: Export the following functions to those who needs to access the Secondary-Puma registers. */
EXPORT_SYMBOL( secondaryPumaGetSoCAddr );
EXPORT_SYMBOL( secondaryPumaGetDocsisAddr );
EXPORT_SYMBOL( secondaryPumaGetDynamic1Addr );
EXPORT_SYMBOL( secondaryPumaGetDynamic2Addr );
EXPORT_SYMBOL( secondaryPumaWriteDocsisPhyReg );
EXPORT_SYMBOL( secondaryPumaReadDocsisPhyReg );
EXPORT_SYMBOL( secondaryPumaWriteDocsisMacReg );
EXPORT_SYMBOL( secondaryPumaReadDocsisMacReg );
EXPORT_SYMBOL( secondaryPumaWriteDocsisMacBuff );
EXPORT_SYMBOL( secondaryPumaReadDocsisMacBuff );
EXPORT_SYMBOL( secondaryPumaGetReadyToTransfer );
EXPORT_SYMBOL( secondaryPumaFinishTransfer );
EXPORT_SYMBOL( secondaryPumaIsInitialized );
EXPORT_SYMBOL( secondaryPumaIsProxyLoaded );
