/*
 *
 * hal_vlynq.c
 * Description:
 * Basic VLYNQ driver implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/


/**************************************************************************/
/*      INCLUDES                                                          */
/**************************************************************************/
#include <linux/kernel.h>
#include <linux/init.h> 
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>

#include <puma5/puma5.h>
#include <puma5/puma5_hardware.h>
#include <generic/pal.h>

#include "_tistdtypes.h"
#include "hal_vlynq.h"

/**************************************************************************/
/*      DEFINES                                                           */
/**************************************************************************/
                                 
#define VLYNQ_RESET_SLAVE_GPIO        8
#define VLYNQ_RESET_SECONDARY_GPIO    29

/* VLYNQ Registers */
#define VLYNQ_CONTROL_REGISTER                  (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x04 )
#define VLYNQ_STATUS_REGISTER                   (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x08 )
#define VLYNQ_INTERRUPT_STATUS_CLEAR_REGISTER   (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x10 )
#define VLYNQ_INTERRUPT_PENDING_SET_REGISTER_OFFSET 0x14
#define VLYNQ_INTERRUPT_POINTER_REGISTER        (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x18 )
#define VLYNQ_TX_ADDRESS_MAP                    (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x1C )
#define VLYNQ_INTERRUPT_VECTOR_3_0_REGISTER     (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x60 )
#define VLYNQ_INTERRUPT_VECTOR_7_4_REGISTER     (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x64 )

#define VLYNQ_RX_ADDRESS_MAP_SIZE( index ) ((AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x20 ) + ( 0x8 * index ))
#define VLYNQ_RX_ADDRESS_MAP_OFFSET( index ) ((AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x24 ) + ( 0x8 * index ))

/* Control register bits */
#define VLYNQ_CONTROL_INT_2_CFG_BIT         (1<<7)
#define VLYNQ_CONTROL_INT_LOCAL_BIT         (1<<14)
#define VLYNQ_CONTROL_CLOCK_DIRCTION_BIT    (1<<15)
#define VLYNQ_CONTROL_CLOCK_DIV_BITS( x )   (x<<16)
#define VLYNQ_CONTROL_RESET_BIT             (1<<0)
#define VLYNQ_CONTROL_RTM_ENABLE            (1<<22)
#define VLYNQ_CONTROL_RTM_VALID             (1<<23)
#define VLYNQ_CONTROL_RTM_SKEW              (0x7 << 24)

/* Status register bits */
#define VLYNQ_STATUS_LINK_BIT               (1<<0)

#define VLYNQ_INT_PHY_VECTOR                (6)
#define VLYNQ_INT_MAC_VECTOR                (2)
#define VLYNQ_MAX_INTERRUPT_NUM             (8)

/* Intterupt Vector 3-0 register bits */
#define VLYNQ_INT_EN_3                      (1<<31)
#define VLYNQ_INT_TYPE_PULSE_3              (1<<30)
#define VLYNQ_INT_POLARITY_LOW_3            (1<<29)
#define VLYNQ_INT_VEC_3(x)                  (((x) & 0x1F) << 24)
#define VLYNQ_INT_EN_2                      (1<<23)
#define VLYNQ_INT_TYPE_PULSE_2              (1<<22)
#define VLYNQ_INT_POLARITY_LOW_2            (1<<21)
#define VLYNQ_INT_VEC_2(x)                  (((x) & 0x1F) << 16)
#define VLYNQ_INT_EN_1                      (1<<15)
#define VLYNQ_INT_TYPE_PULSE_1              (1<<14)
#define VLYNQ_INT_POLARITY_LOW_1            (1<<13)
#define VLYNQ_INT_VEC_1(x)                  (((x) & 0x1F) << 8)
#define VLYNQ_INT_EN_0                      (1<<7)
#define VLYNQ_INT_TYPE_PULSE_0              (1<<6)
#define VLYNQ_INT_POLARITY_LOW_0            (1<<5)
#define VLYNQ_INT_VEC_0(x)                  (((x) & 0x1F) << 0)

/* Intterupt Vector 7-4 register bits */
#define VLYNQ_INT_EN_7                      (1<<31)
#define VLYNQ_INT_TYPE_PULSE_7              (1<<30)
#define VLYNQ_INT_POLARITY_LOW_7            (1<<29)
#define VLYNQ_INT_VEC_7(x)                  (((x) & 0x1F) << 24)
#define VLYNQ_INT_EN_6                      (1<<23)
#define VLYNQ_INT_TYPE_PULSE_6              (1<<22)
#define VLYNQ_INT_POLARITY_LOW_6            (1<<21)
#define VLYNQ_INT_VEC_6(x)                  (((x) & 0x1F) << 16)
#define VLYNQ_INT_EN_5                      (1<<15)
#define VLYNQ_INT_TYPE_PULSE_5              (1<<14)
#define VLYNQ_INT_POLARITY_LOW_5            (1<<13)
#define VLYNQ_INT_VEC_5(x)                  (((x) & 0x1F) << 8)
#define VLYNQ_INT_EN_4                      (1<<7)
#define VLYNQ_INT_TYPE_PULSE_4              (1<<6)
#define VLYNQ_INT_POLARITY_LOW_4            (1<<5)
#define VLYNQ_INT_VEC_4(x)                  (((x) & 0x1F) << 0)

static struct proc_dir_entry *vlynq_proc = NULL;
static struct proc_dir_entry *vlynq_link_proc = NULL;

static int verbose = 0;

/* Static functions */
static void writeWord(unsigned int address, unsigned int value);
static unsigned int readWord(unsigned int address);
static int vlynq_read_link_proc(char *page, char **start, off_t off, int count, int *eof, void *data);
static int vlynq_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data);
static int vlynq_write_proc(struct file *file, const char *buffer, unsigned long count, void *data);

static void VLYNQ_init(void);

/**************************************************************************/
/*      STATIC  Functions                                                 */
/**************************************************************************/

static void writeWord(unsigned int address, unsigned int value)
{
	volatile unsigned int *reg;
	reg = (unsigned int *) address;	
	if (verbose!=0)
		printk("Write: %p <-- %08x\n", reg, value);
	*reg = value;
	PAL_sysCacheFlush(PAL_OSMEM_ADDR_DAT, (void*)(reg), sizeof(unsigned int));
}

static unsigned int readWord(unsigned int address)
{
	volatile unsigned int *reg;
	unsigned int value;
	reg = (unsigned int *) address;	

	PAL_sysCacheInvalidate(PAL_OSMEM_ADDR_DAT, (void*)(reg), sizeof(unsigned int));

	if (verbose!=0)
		printk("Read: %p --> ", reg);
	value = *reg;
	if (verbose!=0)
		printk("%08x\n", value);
	return value;
}

static int vlynq_read_link_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    int len = 0;
    int link;
    int verbose_backup = verbose;
    
    verbose = 0;
    
    link = VLYNQ_get_link_status();
    
    verbose = verbose_backup;
    
    len += sprintf(page + len, "VLYNQ Link is %s.\n", link ? "UP" : "DOWN" );
    
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	
    return(len);
}

static int vlynq_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    int len = 0;
    
    len += sprintf(page + len, "Usage\n");
    len += sprintf(page + len, "\tv: verbose (is %s)\n", (verbose!=0)?"On":"Off");
    len += sprintf(page + len, "\tc:[l/r][i/o]: Setup clock direction: Local/Remote Input/Output\n");
    len += sprintf(page + len, "\td:[l/r] [0-7]: Setup clock divider: Local/Remote 0-7\n");
    len += sprintf(page + len, "\tt:[l/r] [address]: Setup TX map address: Local/Remote address\n");
    len += sprintf(page + len, "\tr:[l/r] [0-3] [address] [size]: Setup RX map address: Local/Remote index address size\n");
    len += sprintf(page + len, "\ts: Reset/Restore VLYNQ link\n");
    len += sprintf(page + len, "\tg:[l/r] Dump Local/Remote registers\n");
    len += sprintf(page + len, "\tR:[b/s] Reset slave [baby/secondary] using GPIO\n");
    len += sprintf(page + len, "\tx:[l/r] [0-7]: Set RTM skew Local/Remote \n");
    
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	
    return(len);
}


static int vlynq_write_proc(struct file *file,
                             const char *buffer,
                             unsigned long count, 
                             void *data)
{
        int len;
		char tmp_buffer[256];
		
        if(count > 256)
                len = 256;
        else
                len = count;

        if(copy_from_user(tmp_buffer, buffer, len))
                return -EFAULT;
        tmp_buffer[len-1] = '\0';

		switch (tmp_buffer[0])
		{
			case 'v':
				verbose = (verbose!=0)?0:1;
				if (verbose!=0)
					printk("verbose on\n");
				else
					printk("verbose off\n");
				break;
    		case 'c':
              {
                   vlynqType_e type;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     

                   printk("Setting %s clock to be ", type == VLYNQ_LOCAL ? "Local":"Remote");
                   if ( tmp_buffer[2]=='o' )
                   {
                        printk("source.\n");               
                        VLYNQ_setup_clock_direction( VLYNQ_SOURCE_CLOCK, type );
                   }
                   else
                   {
                        printk("sink.\n");
                        VLYNQ_setup_clock_direction( VLYNQ_SINK_CLOCK, type );
                   }
              }
              break;

    		case 'd':
              {
                   vlynqType_e type;
                   unsigned int div = 0;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     

                   sscanf( &tmp_buffer[2], "%d", &div );
                   
                   if( div > 7 )
                        div = 7;
                        
                   printk("Setting %s clock divider to be %d\n", type == VLYNQ_LOCAL ? "Local":"Remote", div);
                   VLYNQ_setup_clock_divider( (unsigned int)div, type );
              }
              break;

            case 'x':
            {
                vlynqType_e type;
                unsigned int skew = 0;
            
                if ( tmp_buffer[1]=='l' )
                {
                    type = VLYNQ_LOCAL;
                }
                else
                {    
                    type = VLYNQ_REMOTE;
                }     
            
                sscanf( &tmp_buffer[2], "%d", &skew );
            
                if( skew > 7 )
                    skew = 7;
            
                printk("Setting %s RTM to be %d\n", type == VLYNQ_LOCAL ? "Local":"Remote", skew);
                VLYNQ_setup_rtm( type, skew );
            }
            break;

            
		     case 't':
              {
                   unsigned int addr;

                   vlynqType_e type;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     
                   
				printk("Set TX map pointer of %s vlynq address\n", type == VLYNQ_LOCAL ? "Local":"Remote");
				sscanf( &tmp_buffer[3], "%x", &addr );
				printk("Address = %p\n", (unsigned int *)addr);

                   VLYNQ_setup_tx_region( addr, type );
              }
              break;

		     case 'r':
              {
                   unsigned int addr;
                   unsigned int size;
                   unsigned int index;

                   vlynqType_e type;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     
                   
				printk("Set RX map pointer of %s vlynq address\n", type == VLYNQ_LOCAL ? "Local":"Remote");
				sscanf( &tmp_buffer[3], "%d %x %x", &index, &addr, &size );
				printk("Index = %d, Address = %p, size = 0x%x\n", index, (unsigned int *)addr, size);

                   VLYNQ_setup_rx_region( addr, size, index, type );
              }
              break;
              
    		case 's':
                   printk( "Resetting link\n" );
                   VLYNQ_reset_link( 1 );
                   VLYNQ_reset_link( 0 );
                   break;
                   
    		case 'g':
              {
                   unsigned int i;
                   int last_verbose = verbose;
                   unsigned int base = AVALANCHE_LOW_VLYNQ_CONTROL_BASE;

                   if ( tmp_buffer[1]=='r' )
                        base += 0x80;
                   
                   verbose = 1;
                   for( i = 0; i < 0x20; i++ )
                        readWord( base + (i * 0x4) );
                   
                   verbose = last_verbose;     
              }
              break;

		     case 'R':

                   if ( tmp_buffer[1]=='b' )
                   {
                       printk( "Resetting slave using GPIO #%d\n", VLYNQ_RESET_SLAVE_GPIO );
                       VLYNQ_reset_slave();
                   }
                   else if ( tmp_buffer[1]=='s' )
                   {
                       printk( "Resetting secondary using GPIO #%d\n", VLYNQ_RESET_SECONDARY_GPIO );
                       VLYNQ_reset_secondary();
                   }
                   else 
                   {
                       printk( "Unknown input %s\n", &(tmp_buffer[1]) );
                   }
                   break;
         
              default:
                   printk( "Unsupported command.\n");
                   break;
		}
		
        return len;
}


static void VLYNQ_init(void)
{
    unsigned int i;
    
    /* Take out of reset GPIO PSC */
    PAL_sysPscSetModuleState( PSC_GPIO, PSC_ENABLE );		
    
    /* Change the VLYNQ related pins to the functional ones  */
    for (i=10+32; i<=18+32; i++)
    {
        PAL_sysGpioCtrl(i, FUNCTIONAL_PIN, GPIO_OUTPUT_PIN);
    }
    
    /* Take out of reset VLYNQ PSC */
    PAL_sysPscSetModuleState( PSC_VLYNQ, PSC_ENABLE );	
}

/**************************************************************************/
/*      INTERFACE  Functions                                              */
/**************************************************************************/
/* Reset slave device using GPIO */
int VLYNQ_reset_slave( void )
{

    PAL_sysGpioCtrl(VLYNQ_RESET_SLAVE_GPIO, GPIO_PIN, GPIO_OUTPUT_PIN);     
    PAL_sysGpioOutBit( VLYNQ_RESET_SLAVE_GPIO, 0 );
    msleep( 200 );    
    PAL_sysGpioOutBit( VLYNQ_RESET_SLAVE_GPIO, 1 );
    
    return 0;
}

/* Reset secondary puma using GPIO */
int VLYNQ_reset_secondary( void )
{
    PAL_sysGpioCtrl(VLYNQ_RESET_SECONDARY_GPIO, GPIO_PIN, GPIO_OUTPUT_PIN);     
    PAL_sysGpioOutBit( VLYNQ_RESET_SECONDARY_GPIO, 0 );
    msleep( 200 );    
    PAL_sysGpioOutBit( VLYNQ_RESET_SECONDARY_GPIO, 1 );
    
    return 0;
}


/* Reset VLYNQ link: reset = 1 reset, reset = 0 out of reset */
int VLYNQ_reset_link( unsigned int reset )
{
    unsigned int reg;

    reg = readWord( VLYNQ_CONTROL_REGISTER );
    
    if( reset )
        reg |= VLYNQ_CONTROL_RESET_BIT;
    else
        reg &= ~VLYNQ_CONTROL_RESET_BIT;
    
    writeWord( VLYNQ_CONTROL_REGISTER, reg );
    
    return 0;
}

/* Get VLYNQ link status: 1 Link up, 0 link down */
int VLYNQ_get_link_status( void )
{
    unsigned int reg;

    reg = readWord( VLYNQ_STATUS_REGISTER );
    
    return reg & VLYNQ_STATUS_LINK_BIT ? 1 : 0;
}

/* poll until VLYNQ link is "UP", or max retry exceeded */
/* NOTE: function does NOT release CPU!!!                        */
int VLYNQ_poll_for_link_up( unsigned int maxRetry )
{
    unsigned int reg=0;
    unsigned int repCount=0;
    int oldVerbose;

    /* For polling function that holds CPU - disable verbose prints */
    oldVerbose = verbose;
    verbose = 0;

    while( (!(reg & VLYNQ_STATUS_LINK_BIT)) &&
           (repCount < maxRetry) )
    {
        reg = readWord( VLYNQ_STATUS_REGISTER );
        repCount++;
    }

    /* restore verbose prints */
    verbose = oldVerbose;
    
    return reg & VLYNQ_STATUS_LINK_BIT ? 1 : 0;
}


/* Setup clock divider: 0 to 7 (+1) */
int VLYNQ_setup_clock_divider( unsigned int div, vlynqType_e type )
{
    unsigned int reg;

    if( div > 7 )
         return -1;

    reg = readWord( VLYNQ_CONTROL_REGISTER + type );
    
    reg &= ~VLYNQ_CONTROL_CLOCK_DIV_BITS(7);
    reg |= VLYNQ_CONTROL_CLOCK_DIV_BITS( div );
    
    writeWord( VLYNQ_CONTROL_REGISTER + type, reg );

    return 0;
}

/* Setup clock direction: Source (output) or Sink (input) */
int VLYNQ_setup_clock_direction( vlynqClock_e clock, vlynqType_e type )
{
    unsigned int reg;

    if ( type == VLYNQ_LOCAL )
    {
         if ( clock ==  VLYNQ_SOURCE_CLOCK )
         {
              /* Setup BOOTCR */
              reg = readWord( AVALANCHE_DCL_BOOTCR );
              reg |= (1<<PAL_SYS_CLKC_VLYNQ_DIR_BIT);
              writeWord( AVALANCHE_DCL_BOOTCR, reg );
         }
         else
         {
              /* Setup BOOTCR */
              reg = readWord( AVALANCHE_DCL_BOOTCR );
              reg &= ~(1<<PAL_SYS_CLKC_VLYNQ_DIR_BIT);
              writeWord( AVALANCHE_DCL_BOOTCR, reg );
         }
    }
    
    reg = readWord( VLYNQ_CONTROL_REGISTER + type );
    
    if( clock == VLYNQ_SOURCE_CLOCK )
        reg |= VLYNQ_CONTROL_CLOCK_DIRCTION_BIT;
    else
        reg &= ~VLYNQ_CONTROL_CLOCK_DIRCTION_BIT;
    
    writeWord( VLYNQ_CONTROL_REGISTER + type, reg );

    return 0;
}



/* Setup RTM */
int VLYNQ_setup_rtm(vlynqType_e type, unsigned int skew)
{
    unsigned int reg;

    reg = readWord( VLYNQ_CONTROL_REGISTER + type );
    reg |= VLYNQ_CONTROL_RTM_VALID | VLYNQ_CONTROL_RTM_ENABLE; 

    reg &= ~VLYNQ_CONTROL_RTM_SKEW;
    reg |= ((skew << 24) & VLYNQ_CONTROL_RTM_SKEW);

    writeWord( VLYNQ_CONTROL_REGISTER + type, reg );

    return 0;
}

/**************************************************************************/
/*! \fn int VLYNQ_setup_interrupt_configuration( Bool intlocal, Bool int2cfg, vlynqType_e type )
 **************************************************************************
 *  \brief Setup interrupt configuration
 *  \param[in] Bool intlocal - Interrupt local (bit14 in control register)
 *  \param[in] Bool int2cfg - Interrupt to configuration register (bit7 in control register)
 *  \param[in] vlynqType_e type - VLYNQ_LOCAL or VLYNQ_REMOTE
 *  \return 0 if successful
 **************************************************************************/
int VLYNQ_setup_interrupt_configuration( Bool intlocal, Bool int2cfg, vlynqType_e type )
{
    unsigned int reg;
    
    reg = readWord( VLYNQ_CONTROL_REGISTER + type );
    
    if( intlocal )
    {
        reg |= VLYNQ_CONTROL_INT_LOCAL_BIT;
    }
    else
    {
        reg &= ~VLYNQ_CONTROL_INT_LOCAL_BIT;
    }

    if (int2cfg)
    {
        reg |= VLYNQ_CONTROL_INT_2_CFG_BIT;
    }
    else
    {
        reg &= ~VLYNQ_CONTROL_INT_2_CFG_BIT;
    }
    
    writeWord( VLYNQ_CONTROL_REGISTER + type, reg );
    writeWord( VLYNQ_INTERRUPT_POINTER_REGISTER + type, VLYNQ_INTERRUPT_PENDING_SET_REGISTER_OFFSET );

    
    return 0;
}

/**************************************************************************/
/*! \fn int VLYNQ_enable_remote_phy_interrupt( void )
 **************************************************************************
 *  \brief Enables the remote phy interrupt
 *  \param[in] none
 *  \return 0 if successful
 **************************************************************************/
int VLYNQ_enable_remote_phy_interrupt( void )
{
    unsigned int reg;
    /* Enable PHY interrupts from remote puma using VLYNQ bit 6 */
    reg = VLYNQ_INT_EN_6;
    reg |= VLYNQ_INT_VEC_6(VLYNQ_INT_PHY_VECTOR);
    writeWord( VLYNQ_INTERRUPT_VECTOR_7_4_REGISTER + VLYNQ_REMOTE, reg );

    return 0;
    
}

/**************************************************************************/
/*! \fn int VLYNQ_enable_remote_mac_interrupt( void )
 **************************************************************************
 *  \brief Enables the remote mac interrupt
 *  \param[in] none
 *  \return 0 if successful
 **************************************************************************/
int VLYNQ_enable_remote_mac_interrupt( void )
{
    unsigned int reg;
    /* Enable MAC interrupts from remote puma using VLYNQ bit 2 */
    reg = VLYNQ_INT_EN_2;
    reg |= VLYNQ_INT_TYPE_PULSE_2;
    reg |= VLYNQ_INT_VEC_2(VLYNQ_INT_MAC_VECTOR);
    writeWord( VLYNQ_INTERRUPT_VECTOR_3_0_REGISTER + VLYNQ_REMOTE, reg );

    return 0;
    
}

int VLYNQ_secondary_puma_set_mem_access_start( void )
{
    writeWord(VLYNQ_INTERRUPT_VECTOR_3_0_REGISTER, 1);

    return 0;
}

int VLYNQ_secondary_puma_can_mem_access_start( void )
{
    unsigned int reg;

    reg = readWord(VLYNQ_INTERRUPT_VECTOR_3_0_REGISTER);
    if(reg == 2)
    {
        return 1;
    }
    return 0;
}

int VLYNQ_secondary_puma_set_mem_access_finish( void )
{
    writeWord(VLYNQ_INTERRUPT_VECTOR_3_0_REGISTER, 3);
    return 0;
}

int VLYNQ_secondary_puma_is_mem_access_finish( void )
{
    unsigned int reg;
    
    reg = readWord(VLYNQ_INTERRUPT_VECTOR_3_0_REGISTER);
    if(reg == 4)
    {
        return 1;
    }
    return 0;
}

/**************************************************************************/
/*! \fn int VLYNQ_clear_phy_interrupt( void )
 **************************************************************************
 *  \brief Clears phy interrupt (bit6)
 *  \param[in] none
 *  \return 0 if successful
 **************************************************************************/
int VLYNQ_clear_phy_interrupt( void )
{
    unsigned int reg;
    
    reg = 1 << VLYNQ_INT_PHY_VECTOR;
    
    writeWord( VLYNQ_INTERRUPT_STATUS_CLEAR_REGISTER, reg );
    
    return 0;
}

/**************************************************************************/
/*! \fn int VLYNQ_clear_mac_interrupt( void )
 **************************************************************************
 *  \brief Clears mac interrupt (bit2)
 *  \param[in] none
 *  \return 0 if successful
 **************************************************************************/
int VLYNQ_clear_mac_interrupt( void )
{
    unsigned int reg;
    
    reg = 1 << VLYNQ_INT_MAC_VECTOR;
    
    writeWord( VLYNQ_INTERRUPT_STATUS_CLEAR_REGISTER, reg );
    
    return 0;
}

/**************************************************************************/
/*! \fn int VLYNQ_is_mac_interrupt_pending( void )
 **************************************************************************
 *  \brief determines if a mac interrupt (bit2) is pending
 *  \param[in] none
 *  \return 0 if successful
 **************************************************************************/
int VLYNQ_is_mac_interrupt_pending()
{
    unsigned int reg;
    
    reg = readWord( VLYNQ_INTERRUPT_STATUS_CLEAR_REGISTER);
    if(reg & (1 << VLYNQ_INT_MAC_VECTOR))
        return 1;
    
  
    return 0;
}


/**************************************************************************/
/*! \fn int VLYNQ_is_phy_interrupt_pending( void )
 **************************************************************************
 *  \brief determines if a phy interrupt (bit6) is pending
 *  \param[in] none
 *  \return 0 if successful
 **************************************************************************/
int VLYNQ_is_phy_interrupt_pending(void)
{
    unsigned int reg;
    
    reg = readWord( VLYNQ_INTERRUPT_STATUS_CLEAR_REGISTER);
    if(reg & (1 << VLYNQ_INT_PHY_VECTOR))
        return 1;
    
  
    return 0;
}

/* TX region: The address which will be used for VLYNQ access */
int VLYNQ_setup_tx_region( unsigned int address, vlynqType_e type )
{
    writeWord( VLYNQ_TX_ADDRESS_MAP + type, address );

    return 0;
}

/* RX regions: The address windows which will be available to VLYNQ */
int VLYNQ_setup_rx_region( unsigned int address, unsigned int size, unsigned int index, vlynqType_e type )
{
    if( index > 3 )
         return -1;
         
    writeWord( VLYNQ_RX_ADDRESS_MAP_SIZE( index ) + type, size );
    writeWord( VLYNQ_RX_ADDRESS_MAP_OFFSET( index ) + type, address );
    
    return 0;
}

/* initialize module (and interrupt) */
static int __init vlynqInit (void) 
{
    vlynq_proc = create_proc_entry("vlynq", 0, NULL);
    vlynq_link_proc = create_proc_read_entry("vlynq_link", 0, NULL, vlynq_read_link_proc, NULL);    
	 vlynq_proc->read_proc = vlynq_read_proc;
	 vlynq_proc->write_proc = vlynq_write_proc;

	 VLYNQ_init();
    
	 return 0;
}

/* close and cleanup module */
static void __exit vlynqCleanup (void) 
{
	remove_proc_entry("vlynq", NULL);
	remove_proc_entry("vlynq_link", NULL);   
}

module_init(vlynqInit);
module_exit(vlynqCleanup);
MODULE_AUTHOR("Texas Instruments, Inc");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Basic VLYNQ Interface driver");

EXPORT_SYMBOL( VLYNQ_reset_link );
EXPORT_SYMBOL( VLYNQ_get_link_status );
EXPORT_SYMBOL( VLYNQ_poll_for_link_up );
EXPORT_SYMBOL( VLYNQ_setup_clock_divider );
EXPORT_SYMBOL( VLYNQ_setup_clock_direction );
EXPORT_SYMBOL( VLYNQ_setup_interrupt_configuration );
EXPORT_SYMBOL( VLYNQ_enable_remote_phy_interrupt );
EXPORT_SYMBOL( VLYNQ_enable_remote_mac_interrupt );
EXPORT_SYMBOL( VLYNQ_secondary_puma_set_mem_access_start );
EXPORT_SYMBOL( VLYNQ_secondary_puma_can_mem_access_start );
EXPORT_SYMBOL( VLYNQ_secondary_puma_is_mem_access_finish );
EXPORT_SYMBOL( VLYNQ_secondary_puma_set_mem_access_finish );
EXPORT_SYMBOL( VLYNQ_clear_phy_interrupt );
EXPORT_SYMBOL( VLYNQ_clear_mac_interrupt );
EXPORT_SYMBOL( VLYNQ_is_phy_interrupt_pending );
EXPORT_SYMBOL( VLYNQ_is_mac_interrupt_pending );
EXPORT_SYMBOL( VLYNQ_setup_tx_region );
EXPORT_SYMBOL( VLYNQ_setup_rx_region );
EXPORT_SYMBOL( VLYNQ_reset_slave );
EXPORT_SYMBOL( VLYNQ_reset_secondary );
EXPORT_SYMBOL( VLYNQ_setup_rtm );

